/*                                                                            *
 *   This file is part of the ESO X-shooter Pipeline                          *
 *   Copyright (C) 2006 European Southern Observatory                         *
 *                                                                            *
 *   This library is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as published by     *
 *   the Free Software Foundation; either version 2 of the License, or        *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU General Public License for more details.                             *
 *                                                                            *
 *   You should have received a copy of the GNU General Public License        *
 *   along with this program; if not, write to the Free Software              *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA     *
 *                                                                            */

/*
 * $Author: amodigli $
 * $Date: 2011/03/07 09:45:23 $
 * $Revision: 1.117 $
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------*/
/**
 * @defgroup xsh_mdark   xsh_mdark
 * @ingroup recipes
 *
 * This recipe calculates the master dark frame
 * See man-page for details.
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
  Includes
  ----------------------------------------------------------------------------*/


/* DRL steps */

/* Error handling */
#include <xsh_error.h>
/* Utility fonctions */
#include <xsh_utils.h>
#include <xsh_utils_image.h>
#include <xsh_msg.h>
/* DFS functions */
#include <xsh_dfs.h>
/* DRL functions */
#include <xsh_drl.h>
/* Header functions */
#include <xsh_pfits.h>
/* Library */
#include <cpl.h>
/* CRH Remove */
#include <xsh_drl_check.h>

/*-----------------------------------------------------------------------------
  Defines
  ----------------------------------------------------------------------------*/

#define RECIPE_ID "xsh_mdark"
#define RECIPE_AUTHOR "P.Goldoni, L.Guglielmi, R. Haigron, F. Royer, A. Modigliani"
#define RECIPE_CONTACT "amodigli@eso.org"

/*-----------------------------------------------------------------------------
  Functions prototypes
  ----------------------------------------------------------------------------*/

/*
 *   Plugin initalization, execute and cleanup handlers
 */

static int xsh_mdark_create(cpl_plugin *);
static int xsh_mdark_exec(cpl_plugin *);
static int xsh_mdark_destroy(cpl_plugin *);

/* The actual executor function */
static void xsh_mdark(cpl_parameterlist *, cpl_frameset *);

/*-----------------------------------------------------------------------------
  Static variables
  ----------------------------------------------------------------------------*/
static char xsh_mdark_description_short[] =
"Create the master dark frame";

static char xsh_mdark_description[] =
"This recipe creates a master dark frame\n\
Input Frames : \n\
  - A set of n RAW frames (Format=RAW, n >=3, Tag = DARK_arm)\n\
  - [UVB,VIS] A master bias frame (Format=PRE, Tag = MASTER_BIAS_arm)\n\
  - [OPTIONAL] A badpixel map (Format=QUP, Tag = MASTER_BP_MAP_arm)\n\
  - [OPTIONAL] A map of non linear pixel (Format = QUP,RAW, Tag = BP_MAP_NL_arm)\n\
Products : \n\
  - A map frame containing CRHs  (PRO.CATG = CRH_MAP_arm)\n\
  - A master dark frame (Format=PRE, PRO.CATG = MASTER_DARK_arm)\n\
  - [VIS && bp-output=yes] A badpixel map (Format=QUP, \n\
    PRO.CATG = MASTER_BP_MAP_arm)\n\
  - If an input master bad pixel map or a map of non linear pixels is provided \n\
    A master bad pixel map image, PRO.CATG = MASTER_BP_MAP_arm\n";

/*-----------------------------------------------------------------------------
  Functions code
  ----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
/**
   @brief    Build the list of available plugins, for this module.
   @param    list    the plugin list
   @return   0 if everything is ok, -1 otherwise

   Create the recipe instance and make it available to the application using the
   interface. This function is exported.
*/
/*----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist *list) {
  cpl_recipe *recipe = NULL;
  cpl_plugin *plugin = NULL;

  recipe = cpl_calloc(1, sizeof(*recipe));
  if ( recipe == NULL ){
    return -1;
  }

  plugin = &recipe->interface ;

  cpl_plugin_init(plugin,
                  CPL_PLUGIN_API,                    /* Plugin API */
                  XSH_BINARY_VERSION,             /* Plugin version */
                  CPL_PLUGIN_TYPE_RECIPE,            /* Plugin type */
                  RECIPE_ID,                         /* Plugin name */
                  xsh_mdark_description_short,       /* Short help */
                  xsh_mdark_description,             /* Detailed help */
                  RECIPE_AUTHOR,                     /* Author name */
                  RECIPE_CONTACT,                    /* Contact address */
                  xsh_get_license(),                 /* Copyright */
                  xsh_mdark_create,
                  xsh_mdark_exec,
                  xsh_mdark_destroy);

  cpl_pluginlist_append(list, plugin);

  return (cpl_error_get_code() != CPL_ERROR_NONE);
}

/*----------------------------------------------------------------------------*/
/**
   @brief    Setup the recipe options
   @param    plugin  the plugin
   @return   0 if everything is ok

   Create the recipe instance and make it available to the application using the
   interface.

*/
/*----------------------------------------------------------------------------*/

static int xsh_mdark_create(cpl_plugin *plugin){
  cpl_recipe *recipe = NULL;
  xsh_clipping_param crh_clip_param = {9.0, 3, 0.7, 0, 0.3};
  xsh_clipping_param noise_clip_param = {9.0, 5, 0.7, 0, 0.3};
  xsh_fpn_param fpn_param = {-1,-1,-1,-1,2,1000};
  xsh_ref_param ref_param = {-1,-1,-1,-1};
 xsh_ron_dark_param ron_param = {-1,-1,-1,-1,4,100};
  /* Reset library state */
  xsh_init();

  /* Check input */
  assure( plugin != NULL, CPL_ERROR_NULL_INPUT, "Null plugin");

  /* Get the recipe out of the plugin */
  assure( cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE,
          CPL_ERROR_TYPE_MISMATCH,
          "Plugin is not a recipe");

  recipe = (cpl_recipe *)plugin;

  /* Create the parameter list in the cpl_recipe object */
  recipe->parameters = cpl_parameterlist_new();
  assure( recipe->parameters != NULL,
          CPL_ERROR_ILLEGAL_OUTPUT,
          "Memory allocation failed!");

  /* Set generic parameters (common to all recipes) */
  check( xsh_parameters_generic( RECIPE_ID, recipe->parameters ) ) ;

  check( xsh_parameters_pre_overscan( RECIPE_ID, recipe->parameters ) ) ;

  /* Set bp_output param */
  check(xsh_parameters_new_boolean( recipe->parameters, RECIPE_ID, 
    "bp-output", false,
    "Flag indicating whether the noise bad pixel map is to be computed"));

  /* crh clipping params */
  check(xsh_parameters_clipping_crh_create(RECIPE_ID,recipe->parameters,
    crh_clip_param));

  /* noise clipping params */
  check(xsh_parameters_clipping_noise_create(RECIPE_ID,
                                             recipe->parameters,
					     noise_clip_param));

  check(xsh_parameters_new_double( recipe->parameters,RECIPE_ID, 
    "noise-lower-rejection", 10.,
    "Lower rejection percentile to flag bad pixels via noise"));

  check(xsh_parameters_new_double( recipe->parameters, RECIPE_ID, 
    "noise-higher-rejection", 10.,
    "Upper rejection percentile to flag bad pixels via noise"));

 
  check(xsh_parameters_ref1_create(RECIPE_ID,recipe->parameters,
                                  ref_param));


  check(xsh_parameters_ron_dark_create(RECIPE_ID,recipe->parameters,
                                       ron_param));

  check(xsh_parameters_fpn_create(RECIPE_ID,recipe->parameters,
                                  fpn_param));


 cleanup:
  if ( cpl_error_get_code() != CPL_ERROR_NONE ){
    xsh_error_dump(CPL_MSG_ERROR);
    return 1;
  }
  else {
    return 0;
  }
}

/*----------------------------------------------------------------------------*/
/**
   @brief    Execute the plugin instance given by the interface
   @param    plugin  the plugin
   @return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/

static int xsh_mdark_exec(cpl_plugin *plugin) {
  cpl_recipe *recipe = NULL;

  /* Check parameter */

  assure( plugin != NULL, CPL_ERROR_NULL_INPUT, "Null plugin" );

  /* Get the recipe out of the plugin */
  assure( cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE,
          CPL_ERROR_TYPE_MISMATCH, "Plugin is not a recipe");

  recipe = (cpl_recipe *)plugin;

  /* Check recipe */
  xsh_mdark(recipe->parameters, recipe->frames);

 cleanup:
  if ( cpl_error_get_code() != CPL_ERROR_NONE ) {
    xsh_error_dump(CPL_MSG_ERROR);
    xsh_error_reset();
    return 1;
  }
  else {
    return 0;
  }
}

/*----------------------------------------------------------------------------*/
/**
   @brief    Destroy what has been created by the 'create' function
   @param    plugin  the plugin
   @return   0 if everything is ok
*/
/*----------------------------------------------------------------------------*/
static int xsh_mdark_destroy(cpl_plugin *plugin)
{
  cpl_recipe *recipe = NULL;

  /* reset error state before detroying recipe */
  xsh_error_reset(); 
  /* Check parameter */
  assure( plugin != NULL, CPL_ERROR_NULL_INPUT, "Null plugin" );

  /* Get the recipe out of the plugin */
  assure( cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE,
	  CPL_ERROR_TYPE_MISMATCH, "Plugin is not a recipe");
  recipe = (cpl_recipe *)plugin;

  xsh_free_parameterlist(&recipe->parameters);

 cleanup:
  if (cpl_error_get_code() != CPL_ERROR_NONE)
    {
      return 1;
    }
  else
    {
      return 0;
    }
}

static cpl_error_code 
xsh_params_set_defaults(cpl_parameterlist* pars,
                        xsh_instrument* inst)
{
   cpl_parameter* p=NULL;

  if (xsh_instrument_get_arm(inst) == XSH_ARM_NIR){
     check(p=xsh_parameters_find(pars,RECIPE_ID,"crh-clip-kappa"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
        cpl_parameter_set_double(p,11.);
     }    
  }


 cleanup:
 
  return cpl_error_get_code();

}


/*----------------------------------------------------------------------------*/
/**
   @brief    Interpret the command line options and execute the data processing
   @param    parameters     the parameters list
   @param    frameset   the frames list

   In case of failure the cpl_error_code is set.
*/
/*----------------------------------------------------------------------------*/
static void xsh_mdark(cpl_parameterlist* parameters, cpl_frameset* frameset){
  
  const char* recipe_tags[1] = {XSH_DARK};
  int recipe_tags_size = 1;

  cpl_frameset* raws = NULL;
  cpl_frameset* calib = NULL;
  cpl_frame* bpmap = NULL;
  cpl_frame* master_bias = NULL;
  cpl_frame* rmbias = NULL;
  xsh_instrument* instrument = NULL;
  xsh_clipping_param* crh_clipping = NULL;
  xsh_clipping_param* noise_clipping = NULL;
  int bp_output = 0;
  cpl_imagelist * dataList = NULL;
  cpl_image* noise_image=NULL;
  cpl_parameter* p=NULL;

  /* Intermediate frames */
  cpl_frame * medFrame = NULL ;
  cpl_frame * nirFrame = NULL ;
#if defined(PICKUP_NOISE_HOT_PIXEL_MAP)
  cpl_frame * noisyFrame = NULL ;
#endif
  /* Product frames */
  cpl_frame * resFrame = NULL ; 
  cpl_frame * bpmapFrame = NULL ;

  char *prefix = NULL ;
  cpl_image* crh_ima=NULL;
  char* crh_name=NULL;
  const char* drs_name=NULL;

  cpl_frame* crh_frm=NULL;
  const char* crh_pro_catg=NULL;
  const char* pro_catg=NULL;
  char name[80];
  int binx=0;
  int biny=0;
  cpl_frame* dark_frm=NULL;
  cpl_propertylist* plist=NULL;
  char med_frame_tag[256];
  const char* bp_map_noise_pro_catg=NULL;
  char bp_map_noise_name[256];
  cpl_frame* bp_map_noise_frm=NULL;
  int llx=0;
  int lly=0;
  int urx=0;
  int ury=0;
  int iter=0;
  double low_rej=0;
  double high_rej=0;
  cpl_mask* bp_map_noise=NULL;
  int nraws=0;
  cpl_frame* frm_dark=NULL;
  cpl_frame* frm_drs=NULL;
  cpl_table* tbl_drs=NULL;
  cpl_parameterlist* drs_params=NULL;
  int pre_overscan_corr=0;
  cpl_frame* coadd_bp_map=NULL;
  cpl_image* ima=NULL;
  cpl_propertylist* head=NULL;
  /**************************************************************************/
  /* DFS management */
  /**************************************************************************/
  check( xsh_begin( frameset, parameters, &instrument, &raws, &calib,
                    recipe_tags, recipe_tags_size,
                    RECIPE_ID, XSH_BINARY_VERSION,
                    xsh_mdark_description_short ) ) ;
    

 

 
  /**************************************************************************/
  /* Recipe frames */
  /**************************************************************************/
  XSH_ASSURE_NOT_ILLEGAL((nraws=cpl_frameset_get_size(raws)) >= 1);
  //bpmap = xsh_find_bpmap(calib);
  check(bpmap=xsh_check_load_master_bpmap(calib,instrument));

  if((frm_drs=xsh_find_frame_with_tag(calib,XSH_DRS_MDARK,instrument))==NULL) {
     xsh_msg_warning("Frame %s not provided. Skip QC data reduction",
                     XSH_DRS_MDARK);
     frm_drs=NULL;
  } else {
     check(drs_name=cpl_frame_get_filename(frm_drs));
     check(tbl_drs=cpl_table_load(drs_name,1,0));
     check(drs_params=xsh_parameters_create_from_drs_table(tbl_drs,RECIPE_ID));
  }
  xsh_free_table(&tbl_drs);
  
  /* In UVB and VIS mode only */
  if ( xsh_instrument_get_arm(instrument) != XSH_ARM_NIR){
    /* RAWS frameset must have only one file */ 
    if((master_bias = xsh_find_frame_with_tag(calib,XSH_MASTER_BIAS,
					      instrument)) == NULL) {
      xsh_msg_error("%s_%s is required input for error computation",
	      XSH_MASTER_BIAS,xsh_instrument_arm_tostring (instrument));
      goto cleanup;
    }
    check(dark_frm=cpl_frameset_get_first(raws));
    check(plist=cpl_propertylist_load(cpl_frame_get_filename(dark_frm),0));
    check(binx=xsh_pfits_get_binx(plist));
    check(biny=xsh_pfits_get_biny(plist));
    xsh_free_propertylist(&plist);
    
    sprintf(med_frame_tag,"DARK_REMOVE_CRH_%s",
			       xsh_instrument_arm_tostring (instrument));



  }
  /* IN NIR mode */
  else {
    sprintf(med_frame_tag,"DARK_REMOVE_CRH_%s",
			       xsh_instrument_arm_tostring (instrument));


    /* split on and off files. 
       Amodigli: for DARKS there is not distinction ON/OFF. 
       Thus I copmment it out */
    //check(xsh_dfs_split_nir(raws,&on,&off));
    //XSH_ASSURE_NOT_ILLEGAL(cpl_frameset_get_size(on) >= 3);
    //XSH_ASSURE_NOT_ILLEGAL(cpl_frameset_get_size(off) >= 3);
  }


  /**************************************************************************/
  /* Recipe parameters */
  /**************************************************************************/
  check(xsh_params_set_defaults(parameters,instrument));
  check( crh_clipping = xsh_parameters_clipping_crh_get(RECIPE_ID,
    parameters));
  check( noise_clipping = xsh_parameters_clipping_noise_get(RECIPE_ID,
    parameters));
  check( bp_output = xsh_parameters_get_boolean( parameters,RECIPE_ID,
    "bp-output"));

  check( pre_overscan_corr = xsh_parameters_get_int( parameters, RECIPE_ID,
						     "pre-overscan-corr"));

  /**************************************************************************/
  /* Recipe code */
  /**************************************************************************/

  /* prepare RAW frames in XSH format */
  check(xsh_prepare(raws, bpmap, master_bias, XSH_DARK, instrument,pre_overscan_corr));
  
  /* Call xsh_remove_crh_multiple function */
  xsh_msg( "Calling xsh_remove_crh_multiple" ) ;
  xsh_msg_dbg_medium( "  CRH parameters: Sigma %f, Niteration %d, Fraction %f",
	   crh_clipping->sigma, crh_clipping->niter, crh_clipping->frac ) ;

  check(crh_name=cpl_sprintf("crh_map_dark_%s.fits",
                             xsh_instrument_arm_tostring( instrument))) ;
  cpl_frameset_dump(raws,stdout);
  check_msg( medFrame = xsh_remove_crh_multiple( raws,
						 med_frame_tag,
						 crh_clipping,
						 instrument,
						 &dataList,
                                                 &crh_ima ),
	     "Error in xsh_remove_crh" ) ;

  if(crh_ima!=NULL) {
    /* crh_ima is created only if nraws>=3 */
    check(crh_pro_catg=XSH_GET_TAG_FROM_ARM(XSH_CRH_MAP,instrument));
    check(crh_frm=xsh_frame_product(crh_name,crh_pro_catg,
				    CPL_FRAME_TYPE_IMAGE,
				    CPL_FRAME_GROUP_PRODUCT,
				    CPL_FRAME_LEVEL_FINAL));

    XSH_PREFIX(prefix,"CRH_MAP",instrument);
    check(cpl_image_save(crh_ima,crh_name,CPL_BPP_32_SIGNED,NULL,CPL_IO_DEFAULT));

  }


  xsh_msg_dbg_low( "MedFrame created \"%s\"", cpl_frame_get_filename( medFrame ) ) ;
  xsh_msg( "MedFrame created \"%s\"", cpl_frame_get_filename( medFrame ) ) ;


  /* NIR arm */
  if ( xsh_instrument_get_arm(instrument) == XSH_ARM_NIR ) {
    if (bp_output) {
      xsh_msg_dbg_low("Noise parameters: Sigma %lf, Niteration %d, \
        Fraction %lf, Diff %lf",
	noise_clipping->sigma, noise_clipping->niter,
	noise_clipping->frac, noise_clipping->diff ) ;

 
#if defined(PICKUP_NOISE_HOT_PIXEL_MAP)
      check( nirFrame = xsh_compute_noise_map( dataList, medFrame,
					       noise_clipping,instrument,
					       &noisyFrame ));
 
#else
      check( nirFrame = xsh_compute_noise_map( dataList, medFrame,
        noise_clipping,instrument));
#endif
    
      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ref1_llx"));
      check(llx=cpl_parameter_get_int(p));
      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ref1_lly"));
      check(lly=cpl_parameter_get_int(p));
      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ref1_urx"));
      check(urx=cpl_parameter_get_int(p));
      check(p = cpl_parameterlist_find(parameters,"xsh.xsh_mdark.ref1_ury"));
      check(ury=cpl_parameter_get_int(p));
      check(p = cpl_parameterlist_find(parameters,
                                       "xsh.xsh_mdark.noise-clip-niter"));
      check(iter=cpl_parameter_get_int(p));
      check(p = cpl_parameterlist_find(parameters,
                                       "xsh.xsh_mdark.noise-lower-rejection"));
      check(low_rej=cpl_parameter_get_double(p));
      check(p = cpl_parameterlist_find(parameters,
                                       "xsh.xsh_mdark.noise-higher-rejection"));
      check(high_rej=cpl_parameter_get_double(p));
     
      check(noise_image=xsh_image_search_bad_pixels_via_noise(dataList,
                                                              iter,
                                                              low_rej,high_rej,
                                                              llx,lly,
                                                              urx,ury));
      
      check(bp_map_noise=cpl_mask_threshold_image_create(noise_image,
							 QFLAG_HOT_PIXEL-1,
							 QFLAG_HOT_PIXEL+1));

      xsh_free_image(&noise_image);
      check(noise_image=cpl_image_new_from_mask(bp_map_noise));
      cpl_mask_not(bp_map_noise);

      xsh_bpmap_mask_bad_pixel(noise_image,bp_map_noise,QFLAG_HOT_PIXEL);
   
      //here good pixels are still flagged as '1'. To have proper code, '0',
      //we make a threshold
      cpl_image_threshold(noise_image,1.1,DBL_MAX,0,DBL_MAX);
      xsh_free_mask(&bp_map_noise);
      check(plist=cpl_propertylist_new());
      check(bp_map_noise_pro_catg=XSH_GET_TAG_FROM_ARM(XSH_BP_MAP_NP,
						       instrument));
      sprintf(bp_map_noise_name,"%s.fits",bp_map_noise_pro_catg);
      check(xsh_pfits_set_pcatg(plist,bp_map_noise_pro_catg));
      check(cpl_image_save(noise_image,bp_map_noise_name,CPL_BPP_IEEE_FLOAT,
			   plist,CPL_IO_DEFAULT));
      xsh_msg("saving noise map");
    check(bp_map_noise_frm=xsh_frame_product(bp_map_noise_name,
                                           bp_map_noise_pro_catg,
					   CPL_FRAME_TYPE_IMAGE,
					   CPL_FRAME_GROUP_PRODUCT,
					   CPL_FRAME_LEVEL_FINAL));
      xsh_free_propertylist(&plist);
    } 
    else {
      nirFrame = medFrame;
      medFrame = NULL;
    }
    xsh_msg( "Create master dark" ) ;
    check_msg(resFrame=xsh_create_master_dark(nirFrame,instrument,
                                              parameters,drs_params,
                                              crh_frm,bp_map_noise_frm ),
	       "Error in create master dark" ) ;
 
  }
  else {
    // UVB and VIS 
    //Amo: added to check results of this function also in VIS-UVB
    // Later we decided to have it only in NIR
    //check( nirFrame = xsh_compute_noise_map( dataList, medFrame,
    //	
    /*				  noise_clipping,instrument));
#if defined(PICKUP_NOISE_HOT_PIXEL_MAP)
    check( nirFrame = xsh_compute_noise_map( dataList, medFrame,
					     noise_clipping,instrument,
					     &noisyFrame ));
#else
    check( nirFrame = xsh_compute_noise_map( dataList, medFrame,
					     noise_clipping,instrument));
#endif
    */
    if(master_bias != NULL) {
      xsh_msg("Subtract bias");
      check(rmbias = xsh_subtract_bias(medFrame,master_bias,instrument,"DARK_",pre_overscan_corr));

       xsh_msg( "Generates master dark" ) ;
       check_msg(resFrame=xsh_create_master_dark(rmbias, instrument,
                                                 parameters,drs_params,
                                                 crh_frm,bp_map_noise_frm),
                 "Error in create master dark" ) ; 
    } else {
       check(resFrame=cpl_frame_duplicate(medFrame));
    }
  }

  /* saving products */   
  xsh_msg( "Save products" ) ;
  check( xsh_add_product_pre( resFrame, frameset, parameters,RECIPE_ID,
    instrument));


  if(crh_ima!=NULL) {
    /* crh_ima is created only if nraws>=3 */
    check(xsh_add_product_image(crh_frm, frameset,parameters, 
				RECIPE_ID, instrument, prefix ));
    xsh_add_temporary_file(crh_name);
  }


  check(bpmapFrame = xsh_create_master_bpmap(resFrame, instrument));

  XSH_ASSURE_NOT_NULL(bpmapFrame);

  if(bp_map_noise_frm!=NULL) {
  check( xsh_add_product_image(bp_map_noise_frm, frameset, parameters,
                               RECIPE_ID, instrument, 
                               cpl_frame_get_filename( bp_map_noise_frm)));

    xsh_msg( "Coadd bp map noise frames: %s and %s",
	     cpl_frame_get_filename(bpmapFrame),
	     cpl_frame_get_filename(bp_map_noise_frm) ) ;
    check(xsh_badpixelmap_coadd(bpmapFrame,bp_map_noise_frm));
  }
  XSH_PREFIX(prefix, XSH_MASTER_BP_MAP, instrument);
  xsh_add_temporary_file(cpl_frame_get_filename(bpmapFrame));
  check( xsh_add_product_bpmap( bpmapFrame, frameset, parameters,
  				RECIPE_ID, instrument, prefix ));
 
  frm_dark=cpl_frameset_get_frame(raws,0);
  check(pro_catg=XSH_GET_TAG_FROM_ARM(XSH_DARK,instrument));
  sprintf(name,"%s",pro_catg);
  check( xsh_add_product_image(frm_dark, frameset, parameters,
				    RECIPE_ID, instrument, name ));

  //AMO: we finally decided to have it only in NIR
  if ( xsh_instrument_get_arm(instrument) == XSH_ARM_NIR ) {
    if(bp_output) {
#if defined(PICKUP_NOISE_HOT_PIXEL_MAP)
      XSH_PREFIX(prefix, XSH_BP_MAP_PN, instrument ) ;
      check( xsh_add_product_image( noisyFrame, frameset, parameters,
				    RECIPE_ID, instrument, prefix ));
      xsh_add_temporary_file(cpl_frame_get_filename(noisyFrame));
#else
    check( xsh_add_product_pre( nirFrame, frameset, parameters, 
      RECIPE_ID, instrument));
#endif
    }
  }

  if(bpmap!=NULL) {
     if(bp_map_noise_frm!=NULL) {
        xsh_msg( "Coadd noise and input bp map frames" ) ;
        check(coadd_bp_map=cpl_frame_duplicate(bpmap));
        sprintf(name,cpl_frame_get_filename(bpmap));
        check(ima=cpl_image_load(name,CPL_TYPE_FLOAT,0,0));
        check(head=cpl_propertylist_load(name,0));
        check(pro_catg=XSH_GET_TAG_FROM_ARM(XSH_MASTER_BP_MAP,instrument));
        sprintf(name,"%s.fits",pro_catg);
        check(cpl_image_save(ima,name,CPL_BPP_IEEE_FLOAT,head,CPL_IO_DEFAULT));
        check(cpl_frame_set_filename(coadd_bp_map,name));
        check(cpl_frame_set_tag(coadd_bp_map,pro_catg));
        check(xsh_badpixelmap_coadd(coadd_bp_map,bp_map_noise_frm));
        check( xsh_add_product_bpmap(coadd_bp_map, frameset, parameters,
                                     RECIPE_ID, instrument, pro_catg ));
        xsh_free_frame(&coadd_bp_map);
        xsh_free_image(&ima);
        xsh_free_propertylist(&head);
     }
  }

 
cleanup:
  xsh_end( RECIPE_ID, frameset, parameters );
  XSH_FREE(crh_clipping);
  XSH_FREE(noise_clipping);
  xsh_instrument_free(&instrument );
  xsh_free_frameset(&raws);
  xsh_free_frameset(&calib);
  xsh_free_imagelist( &dataList );
  xsh_free_frame( &medFrame ) ;
  xsh_free_frame( &nirFrame ) ;
  xsh_free_frame( &resFrame ) ;
  xsh_free_frame(&rmbias);
  xsh_free_frame(&bpmapFrame);
  xsh_free_frame(&crh_frm);
  xsh_free_image(&crh_ima);
  xsh_free_propertylist(&plist);
  xsh_free_table(&tbl_drs);
  xsh_free_parameterlist(&drs_params);
  XSH_FREE(prefix) ;
  XSH_FREE(crh_name) ;

  return;
}

/**@}*/
