/*                                                                            *
 *   This file is part of the ESO X-shooter Pipeline                          *
 *   Copyright (C) 2006 European Southern Observatory                         *
 *                                                                            *
 *   This library is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as published by     *
 *   the Free Software Foundation; either version 2 of the License, or        *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU General Public License for more details.                             *
 *                                                                            *
 *   You should have received a copy of the GNU General Public License        *
 *   along with this program; if not, write to the Free Software              *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA     *
 *                                                                            */

/*
 * $Author: amodigli $
 * $Date: 2011/03/07 09:45:36 $
 * $Revision: 1.125 $
 * $Name: HEAD $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*--------------------------------------------------------------------------*/
/**
 * @defgroup xsh_mflat   xsh_mflat
 * @ingroup recipes
 *
 * This recipe calculates the master flat frame
 * See man-page for details.
 */
/*--------------------------------------------------------------------------*/
/**@{*/

/*---------------------------------------------------------------------------
                                Includes
 ---------------------------------------------------------------------------*/


/* DRL steps */

/* Error handling */
#include <xsh_error.h>
/* Utility fonctions */
#include <xsh_pfits.h>
#include <xsh_utils.h>
#include <xsh_utils_image.h>
#include <xsh_msg.h>
/* DFS functions */
#include <xsh_dfs.h>
/* DRL functions */
#include <xsh_data_instrument.h>
#include <xsh_data_order.h>
#include <xsh_drl.h>
#include <xsh_drl_check.h>
/* Library */
#include <cpl.h>
#include <xsh_drl_check.h>

/*---------------------------------------------------------------------------
                                Defines
 ---------------------------------------------------------------------------*/

#define RECIPE_ID "xsh_mflat"
#define RECIPE_AUTHOR "P.Goldoni, L.Guglielmi, R. Haigron, F. Royer, A. Modigliani"
#define RECIPE_CONTACT "amodigli@eso.org"

/*---------------------------------------------------------------------------
                            Functions prototypes
 ---------------------------------------------------------------------------*/

/*
 *   Plugin initalization, execute and cleanup handlers
 */

static int xsh_mflat_create(cpl_plugin *);
static int xsh_mflat_exec(cpl_plugin *);
static int xsh_mflat_destroy(cpl_plugin *);

/* The actual executor function */
static void xsh_mflat(cpl_parameterlist *, cpl_frameset *);

/*---------------------------------------------------------------------------
                            Static variables
 ---------------------------------------------------------------------------*/
static char xsh_mflat_description_short[] =
"Create the master flat and the orders edges traces table frames";

static char xsh_mflat_description[] =
"This recipe creates the master flat and the orders edges traces table frames.\n\
  Input Frames :\n\
    - [UVB] A set of n RAW frames (Format = RAW, n>=3,\
 Tag = FLAT_D2_mode_UVB, mode=SLIT/IFU)\n\
    - [UVB] A set of n RAW frames (Format = RAW, n>=3,\
 Tag = FLAT_QTH_mode_UVB)\n\
    - [VIS] A set of n RAW frames (Format = RAW, n>=3,\
 Tag = FLAT_mode_VIS)\n\
    - [NIR] A set of n x n RAW frames ((Format = RAW, n>=3,\
 Tag = FLAT_mode_NIR_ON, FLAT_mode_NIR_OFF)\n\
    - [OPTIONAL,NIR] A map of non linear pixel (Format = QUP,RAW, Tag = BP_MAP_NL_arm)\n\
    - [OPTIONAL] A master badpixel map (Format = QUP, Tag = MASTER_BP_MAP_arm)\n\
    - An order table (Format = TABLE, Tag = ORDER_TAB_CENTR_arm)\n\
    - [UVB,VIS] A master bias (Format = PRE, Tag = MASTER_BIAS_arm)\n\
    - [UVB,VIS,OPTIONAL] A master dark (Format = PRE, Tag = MASTER_DARK_arm)\n\
    - A spectral format table (Format = PRE, Tag = SPECTRAL_FORMAT_TAB_arm)\n\
  Products : \n\
    - An updated order table with edge UP and edge LOW\
 (Format = TABLE, TAG = ORDER_TAB_EDGES_mode_arm)\n\
    - A master flat (Format = PRE, PRO.CATG = MASTER_FLAT_mode_arm)\n\
    - A background subtracted flat (Format = PRE, PRO.CATG = MFLAT_SUBTRACT_BACKGROUND_mode_arm)\n\
    - The inter-order background frame (Format = PRE, PRO.CATG = BACKGROUND_mode_arm)\n\
    - The inter-order background sampling ppoints grid table\n\
      (Format = PRE, PRO.CATG = GRID_mode_arm)\n\
    - The map of dead pixels (Format = PRE, PRO.CATG = BP_MAP_DP_arm) \n\
    - The map of dead pixels (Format = PRE, PRO.CATG = BP_MAP_SP_arm) \n\
    - [if MASTER_BP_MAP_arm or BP_MAP_NL_arm are provided] The updated master bad pixel map (Format = PRE, PRO.CATG = MASTER_BP_MAP_arm) \n\
  \n\
  Prepare the flat frames.\n\
  Stack and sigma clip all the flat frames.\n\
  Subtract master bias.\n\
  Subtract master dark.\n\
  Detect order edge.\n\
  Subtract background.\n\
  Create the Master Flat.\n";

/*---------------------------------------------------------------------------
                              Functions code
 ---------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok, -1 otherwise

  Create the recipe instance and make it available to the application using 
  the interface. This function is exported.
 */
/*--------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist *list) {
  cpl_recipe *recipe = NULL;
  cpl_plugin *plugin = NULL;

  recipe = cpl_calloc(1, sizeof(*recipe));
  if ( recipe == NULL ){
    return -1;
  }

  plugin = &recipe->interface ;

  cpl_plugin_init(plugin,
                  CPL_PLUGIN_API,                   /* Plugin API */
                  XSH_BINARY_VERSION,            /* Plugin version */
                  CPL_PLUGIN_TYPE_RECIPE,           /* Plugin type */
                  RECIPE_ID,                        /* Plugin name */
                  xsh_mflat_description_short,      /* Short help */
                  xsh_mflat_description,            /* Detailed help */
                  RECIPE_AUTHOR,                    /* Author name */
                  RECIPE_CONTACT,                   /* Contact address */
                  xsh_get_license(),                /* Copyright */
                  xsh_mflat_create,
                  xsh_mflat_exec,
                  xsh_mflat_destroy);

  cpl_pluginlist_append(list, plugin);

  return (cpl_error_get_code() != CPL_ERROR_NONE);
 }

/*--------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using 
  the interface.

 */
/*--------------------------------------------------------------------------*/

static int xsh_mflat_create(cpl_plugin *plugin){
  cpl_recipe *recipe = NULL;
  xsh_clipping_param crh_clip_param = {0.3,10.0, 2, 0.7, 0};
  //xsh_hot_cold_pix_param hp_clip_param = {0, 5.0, 3, 5.0 , 3};

  /* Reset library state */
  xsh_init();

  /* Check input */
  assure( plugin != NULL, CPL_ERROR_NULL_INPUT, "Null plugin");

  /* Get the recipe out of the plugin */
  assure( cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE,
          CPL_ERROR_TYPE_MISMATCH,
          "Plugin is not a recipe");

  recipe = (cpl_recipe *)plugin;

  /* Create the parameter list in the cpl_recipe object */
  recipe->parameters = cpl_parameterlist_new();
  assure( recipe->parameters != NULL,
          CPL_ERROR_ILLEGAL_OUTPUT,
          "Memory allocation failed!");

  /* Fill the parameter list */
  /* Set generic parameters (common to all recipes) */
  check( xsh_parameters_generic( RECIPE_ID, recipe->parameters ) ) ;
  check( xsh_parameters_pre_overscan( RECIPE_ID, recipe->parameters ) ) ;

  /* crh clipping params */
  check( xsh_parameters_clipping_crh_create( RECIPE_ID, recipe->parameters,
    crh_clip_param));

  /* detect order params */
  check( xsh_parameters_detect_order_create( RECIPE_ID,
    recipe->parameters));

  /* d2 detect order params */
  check( xsh_parameters_d2_detect_order_create( RECIPE_ID,
    recipe->parameters));

  /* subtract_background_params */
  check( xsh_parameters_background_create( RECIPE_ID,
    recipe->parameters));
  /*
  check(xsh_parameters_hot_cold_pix_create(RECIPE_ID,recipe->parameters,
                                           hp_clip_param));
  */
  cleanup:
    if ( cpl_error_get_code() != CPL_ERROR_NONE ){
      xsh_error_dump(CPL_MSG_ERROR);
      return 1;
    }
    else {
      return 0;
    }
}

/*--------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*--------------------------------------------------------------------------*/

static int xsh_mflat_exec(cpl_plugin *plugin) {
  cpl_recipe *recipe = NULL;

  /* Check parameter */
  assure( plugin != NULL, CPL_ERROR_NULL_INPUT, "Null plugin" );

  /* Get the recipe out of the plugin */
  assure( cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE,
          CPL_ERROR_TYPE_MISMATCH, "Plugin is not a recipe");

  recipe = (cpl_recipe *)plugin;

  /* Check recipe */
  xsh_mflat(recipe->parameters, recipe->frames);

  cleanup:
    if ( cpl_error_get_code() != CPL_ERROR_NONE ) {
      xsh_error_dump(CPL_MSG_ERROR);
      /* RESTORE the error context */
      cpl_error_reset();
      return 1;
    }
    else {
      return 0;
    }
}

/*--------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*--------------------------------------------------------------------------*/
static int xsh_mflat_destroy(cpl_plugin *plugin)
{
    cpl_recipe *recipe = NULL;

    /* Check parameter */
    assure( plugin != NULL, CPL_ERROR_NULL_INPUT, "Null plugin" );

    /* Get the recipe out of the plugin */
    assure( cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE,
            CPL_ERROR_TYPE_MISMATCH, "Plugin is not a recipe");

    recipe = (cpl_recipe *)plugin;
 
    xsh_free_parameterlist(&recipe->parameters);

  cleanup:
    if (cpl_error_get_code() != CPL_ERROR_NONE)
        {
            return 1;
        }
    else
        {
            return 0;
        }
}




static void 
xsh_mflat_recipe_main( cpl_frameset *raw_set, 
                       cpl_frameset *on_set,
		       cpl_frameset *off_set, 
		       cpl_frame *bpmap_frame,
		       cpl_frame *master_bias_frame, 
		       cpl_frame *master_dark_frame,  
                       cpl_frame *cen_order_tab_frame,
                       cpl_frame *qc_cen_order_tab_frame,
		       xsh_clipping_param *crhclip_par,
		       xsh_detect_order_param *detectorder_par, 
		       xsh_background_param* background_par,
		       const int pre_overscan_corr,
		       cpl_frame **edges_order_tab_frame,  
		       cpl_frame **master_flat_frame,
		       cpl_frame **rmbackground,
		       cpl_frame **grid_background,
                       cpl_frame** frame_backg,
		       xsh_instrument* instrument)
{
  cpl_frameset *onoff_set = NULL;
  cpl_frame *rmcrh = NULL;
  cpl_frame *rmbias = NULL;
  cpl_frame *rmdark = NULL;
  char* fname = NULL;
  const char* ftag = NULL;
  
  if ( xsh_instrument_get_arm(instrument) != XSH_ARM_NIR){
    /* prepare RAW frames in XSH format */
    check(xsh_prepare( raw_set, bpmap_frame, master_bias_frame, 
		       XSH_FLAT, instrument,pre_overscan_corr));
    ftag=XSH_GET_TAG_FROM_ARM(XSH_FLAT_REMOVE_CRH,instrument);

    check( rmcrh = xsh_remove_crh_multiple( raw_set, ftag, crhclip_par,
      instrument, NULL,NULL));

    /* subtract bias */
    xsh_msg("Subtract bias");
    /* AMO: here one should not subtract the bias if overscan correction 
    is applied */
    check( rmbias = xsh_subtract_bias( rmcrh, master_bias_frame, 
					instrument,"FLAT_",pre_overscan_corr));

    xsh_monitor_flux(rmbias,qc_cen_order_tab_frame, instrument);
    if(  master_dark_frame != NULL) {
      xsh_msg( "Subtract dark");
      XSH_NAME_LAMP_MODE_ARM( fname, "FLAT_SUBTRACT_DARK", ".fits", 
        instrument);
      check( rmdark = xsh_subtract_dark( rmbias, master_dark_frame,
        fname, instrument));
    } 
    else {
      rmdark = cpl_frame_duplicate( rmbias);
    }
  }
  /* in NIR mode */
  else{
    check( xsh_prepare( on_set, bpmap_frame, NULL, "ON", instrument,pre_overscan_corr));
    check( xsh_prepare( off_set, bpmap_frame, NULL, "OFF", instrument,pre_overscan_corr));
    /* do ON - OFF */
    check( onoff_set = xsh_subtract_nir_on_off( on_set, off_set, instrument));
    ftag=XSH_GET_TAG_FROM_ARM(XSH_ONOFF_REMOVE_CRH,instrument);
   
    check( rmdark = xsh_remove_crh_multiple( onoff_set, ftag, 
      crhclip_par, instrument, NULL,NULL));
    xsh_monitor_flux(rmdark,qc_cen_order_tab_frame, instrument);
  }
 
  /* detect order edge */
  xsh_msg("Detect order edges");
  check( *edges_order_tab_frame = xsh_detect_order_edge( rmdark,
    cen_order_tab_frame, detectorder_par, instrument));

  /* subtract background */
  xsh_msg("Subtract inter-order background");
  check( *rmbackground = xsh_subtract_background( rmdark,
						  *edges_order_tab_frame, 
						  background_par, 
						  instrument, "MFLAT",
						  grid_background,
						  frame_backg));

  /* Create the master flat */
  check( *master_flat_frame = xsh_create_master_flat( *rmbackground, 
						      instrument));
 


  cleanup:
    XSH_FREE( fname);
    xsh_free_frameset( &onoff_set);
    xsh_free_frame( &rmcrh);
    xsh_free_frame( &rmbias);
    xsh_free_frame( &rmdark);
    return;
}


static cpl_error_code 
xsh_params_monitor(xsh_background_param* backg,
                   xsh_detect_order_param* det_order)
{


  xsh_msg_dbg_low("bkg params: sampley=%d radius_y=%d smooth_y=%d",
	  backg->sampley,backg->radius_y,backg->smooth_y);

  xsh_msg_dbg_low("bkg params: radius_x=%d smooth_x=%d",
	  backg->radius_x,backg->smooth_x);

  xsh_msg_dbg_low("detect order params: min_order_size_x=%d",
	  det_order->min_order_size_x);



  return cpl_error_get_code();

}
/*--------------------------------------------------------------------------*/
/**
  @brief    Rescale parameters for binning
  @param    raws   raw frames set
  @param    backg parameters to control inter-order background subtraction
  @param    det_order parameters to control order edges detection
  In case of failure the cpl_error_code is set.
 */
/*--------------------------------------------------------------------------*/

static cpl_error_code 
xsh_params_bin_scale(cpl_frameset* raws,
                     xsh_background_param* backg,
                     xsh_detect_order_param* det_order)
{

  cpl_frame* frame=NULL;
  const char* name=NULL;
  cpl_propertylist* plist=NULL;
  int binx=0;
  int biny=0;

  check(frame=cpl_frameset_get_first(raws));
  check(name=cpl_frame_get_filename(frame));
  check(plist=cpl_propertylist_load(name,0));
  check(binx=xsh_pfits_get_binx(plist));
  check(biny=xsh_pfits_get_biny(plist));
  xsh_free_propertylist(&plist);

  if(biny>1) {

    /* backg->sampley=backg->sampley/biny; 
       No of sampling points, not to be sampling bin dependent */
    backg->radius_y=backg->radius_y/biny;
   
    /*
    if(backg->smooth_y>0) {
      backg->smooth_y=backg->smooth_y/biny;
    }
    Radius of post smoothing: 
    for the moment not to be bin dependent,but for optimal results probably yes
    */


  }


  if(binx>1) {

    backg->radius_x=backg->radius_x/binx;
    /*
    if(backg->smooth_x>0) {
      backg->smooth_x=backg->smooth_x/binx;
    }
    Radius of post smoothing: 
    for the moment not to be bin dependent,but for optimal results probably yes
    */

    det_order->search_window_hsize=det_order->search_window_hsize/biny;
    det_order->min_order_size_x=det_order->min_order_size_x/binx;

  }
 
 cleanup:
  xsh_free_propertylist(&plist);
  return cpl_error_get_code();

}

static cpl_error_code 
xsh_params_set_defaults(cpl_parameterlist* pars,
                        xsh_instrument* inst,
                        xsh_detect_order_param* det_order,
                        xsh_background_param* backg)
{
   cpl_parameter* p=NULL;
 
  if (xsh_instrument_get_arm(inst) == XSH_ARM_NIR){

     check(p=xsh_parameters_find(pars,RECIPE_ID,"detectorder-min-order-size-x"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
        det_order->min_order_size_x=40;
     }


     check(p=xsh_parameters_find(pars,RECIPE_ID,"background-radius-x"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
       backg->radius_x=1;
     }

     check(p=xsh_parameters_find(pars,RECIPE_ID,"detectorder-min-sn"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
        if(xsh_instrument_get_mode(inst) == XSH_MODE_IFU) {
           det_order->min_sn=4;
        } else {
           det_order->min_sn=25;
        }
     }
  } else {

     check(p=xsh_parameters_find(pars,RECIPE_ID,"detectorder-min-sn"));
     if(xsh_parameter_get_default_flag(p) == 0) {
        /* if the user has not set the parameter use default */
        if(xsh_instrument_get_mode(inst) == XSH_MODE_IFU) {
           det_order->min_sn=20;
        }
     }

  }

 cleanup:
 
  return cpl_error_get_code();

}
/*--------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    parameters     the parameters list
  @param    frameset   the frames list

  In case of failure the cpl_error_code is set.
 */
/*--------------------------------------------------------------------------*/
static void xsh_mflat(cpl_parameterlist* parameters, cpl_frameset* frameset)
{
  const char* recipe_tags[1] = {XSH_FLAT};
  int recipe_tags_size = 1;

  /* ALLOCATED locally */
  cpl_frameset* raws = NULL;
  cpl_frameset* calib = NULL;
  cpl_frameset* on = NULL;
  cpl_frameset* off = NULL;
  cpl_frameset* qth = NULL;
  cpl_frameset* d2 = NULL;
  xsh_instrument* instrument = NULL;
  /* Input frames */
  cpl_frame* spectralformat_frame = NULL;
  cpl_frame* bpmap = NULL;
  cpl_frame* master_bias = NULL;
  cpl_frame* master_dark = NULL;
  cpl_frame* cen_order_tab_frame = NULL;
  /* Products */
  cpl_frame* edges_order_tab_frame = NULL;
  cpl_frame* master_flat_frame = NULL;
  cpl_frame* master_bkg_frame = NULL;

  cpl_frame *qth_edges_order_tab_frame = NULL;
  cpl_frame *qth_master_flat_frame = NULL;
  cpl_frame *d2_edges_order_tab_frame = NULL;
  cpl_frame* d2_master_flat_frame = NULL;
  /* Parameters */
  xsh_clipping_param* crh_clipping = NULL;
  xsh_detect_order_param* det_order = NULL;
  xsh_d2_detect_order_param * d2_det_order = NULL;
  xsh_background_param* backg = NULL;
  /* Others */
  cpl_frame *qth_cen_order_tab_frame = NULL;
  cpl_frame *d2_cen_order_tab_frame = NULL;
  char fname[80];
  char fname_d2[80];
  char fname_qth[80];
  char pcatg[80];
  cpl_table* grid_tab1=NULL;
  cpl_table* grid_tab2=NULL;
  cpl_propertylist* plist=NULL;
  cpl_frame* rmbackground=NULL;
  cpl_frame* qth_rmbackground=NULL;
  cpl_frame* d2_rmbackground=NULL;
  cpl_frame* grid_backg=NULL;
  cpl_frame* d2_grid_backg=NULL;
  cpl_frame* qth_grid_backg=NULL;
  cpl_frame* frame_backg=NULL;
  cpl_frame* d2_frame_backg=NULL;
  cpl_frame* qth_frame_backg=NULL;
  cpl_frame* cpix_frm=NULL;
  cpl_frame* hpix_frm=NULL;
  int pre_overscan_corr=0;
  /**************************************************************************/
  /* DFS management */
  /**************************************************************************/
  check( xsh_begin( frameset, parameters, &instrument, &raws, &calib,
		    recipe_tags, recipe_tags_size,
		    RECIPE_ID, XSH_BINARY_VERSION,
		    xsh_mflat_description_short ) ) ;
  /**************************************************************************/
  /* Recipe frames */
  /**************************************************************************/
  check(bpmap=xsh_check_load_master_bpmap(calib,instrument));
  check( cen_order_tab_frame = xsh_find_order_tab_centr( calib, instrument));

  check( spectralformat_frame = xsh_find_frame_with_tag( calib,
    XSH_SPECTRAL_FORMAT, instrument));
  //check(order_tab_centr = xsh_find_frame_with_tag(calib,XSH_ORDER_TAB_CENTR,
  //                                                 instrument));

  if (xsh_instrument_get_arm(instrument) == XSH_ARM_UVB){
    XSH_ASSURE_NOT_ILLEGAL(cpl_frameset_get_size(raws) >= 2);
    XSH_ASSURE_NOT_ILLEGAL( instrument->lamp == XSH_LAMP_QTH_D2);
  }
  else{
    XSH_ASSURE_NOT_ILLEGAL(cpl_frameset_get_size(raws) >= 1);
  }


  /* In UVB and VIS mode */
  if ( xsh_instrument_get_arm(instrument) != XSH_ARM_NIR){
    /* RAWS frameset must have only one file */
    check(master_bias = xsh_find_master_bias( calib, instrument));
    //check(master_bias = xsh_find_frame_with_tag(calib,XSH_MASTER_BIAS,
    //                                            instrument));

    if((master_dark =  xsh_find_master_dark( calib, instrument))==NULL) {
    //check(master_dark = xsh_find_frame_with_tag(calib,XSH_MASTER_DARK,
    //                                            instrument));
      xsh_msg_warning("Frame %s not provided",XSH_MASTER_DARK);
      xsh_error_reset();
    }
    if ( xsh_instrument_get_arm( instrument) == XSH_ARM_UVB){
      check( xsh_dfs_split_qth_d2( raws, &qth, &d2));
    }
  }
  /* IN NIR mode */
  else {
    /* split on and off files */
    check(xsh_dfs_split_nir(raws,&on,&off));
    /* AMO: the following check is indeeed too restrictive:
      XSH_ASSURE_NOT_ILLEGAL(cpl_frameset_get_size(on) >= 3);
      XSH_ASSURE_NOT_ILLEGAL(cpl_frameset_get_size(off) >= 3);
    */
  }


  check( xsh_instrument_update_from_spectralformat( instrument,
    spectralformat_frame));

  /**************************************************************************/
  /* Recipe parameters */
  /**************************************************************************/
  check( pre_overscan_corr = xsh_parameters_get_int( parameters, RECIPE_ID,
						     "pre-overscan-corr"));

  check( crh_clipping = xsh_parameters_clipping_crh_get( RECIPE_ID,
    parameters));
  check( det_order = xsh_parameters_detect_order_get( RECIPE_ID,
    parameters));
  check( d2_det_order = xsh_parameters_d2_detect_order_get( RECIPE_ID,
    parameters));
  check( backg =  xsh_parameters_background_get( RECIPE_ID,
    parameters));

  /* adjust relevant parameter to binning */
  check(xsh_params_monitor(backg,det_order));
  if ( xsh_instrument_get_arm(instrument) != XSH_ARM_NIR){
    check(xsh_params_bin_scale(raws,backg,det_order));
  }
  xsh_params_set_defaults(parameters,instrument,det_order,backg);

  check(xsh_params_monitor(backg,det_order));


  /**************************************************************************/
  /* Recipe code */
  /**************************************************************************/
  if ( xsh_instrument_get_arm(instrument) != XSH_ARM_UVB){
     /* VIS or NIR */
     check( xsh_mflat_recipe_main(raws, on, off,bpmap,master_bias, 
				  master_dark, cen_order_tab_frame,
                                   cen_order_tab_frame, 
				  crh_clipping, det_order, backg,
				  pre_overscan_corr,
				  &edges_order_tab_frame, &master_flat_frame, 
                                  &rmbackground, &grid_backg,&frame_backg,
                                  instrument));
     /***********************************************************************/
     /* Products */
     /***********************************************************************/
     xsh_msg("Saving products");
     check( xsh_add_product_table_multi( edges_order_tab_frame, frameset,
                                         parameters, RECIPE_ID, instrument));

    
     xsh_msg("Detect dead-saturated pixels");
     hpix_frm=xsh_image_mflat_detect_hotcold(master_flat_frame,instrument,1);
     cpix_frm=xsh_image_mflat_detect_hotcold(master_flat_frame,instrument,0);
     if(bpmap!=NULL) {
       check(xsh_badpixelmap_coadd(bpmap,hpix_frm));
       check(xsh_badpixelmap_coadd(bpmap,cpix_frm));
     }
    

     check( xsh_add_product_pre( master_flat_frame, frameset, 
                                 parameters, RECIPE_ID, instrument));
   
     check( xsh_add_product_image( rmbackground, frameset, 
                                   parameters, RECIPE_ID, instrument,
                                   cpl_frame_get_tag(rmbackground)));

     check( xsh_add_product_image( frame_backg, frameset, parameters, RECIPE_ID,
                                   instrument,
                                   cpl_frame_get_tag(frame_backg)));

     check(xsh_add_product_table_multi( grid_backg, frameset, 
                                        parameters, RECIPE_ID, instrument));

     
     check( xsh_add_product_image( cpix_frm, frameset, 
				   parameters, RECIPE_ID, instrument,
				   cpl_frame_get_tag(cpix_frm)));

     check( xsh_add_product_image( hpix_frm, frameset, 
				   parameters, RECIPE_ID, instrument,
				   cpl_frame_get_tag(hpix_frm)));
      

    

     if(bpmap!=NULL) {

        check( xsh_add_product_bpmap(bpmap, frameset, parameters,
                                     RECIPE_ID, instrument, 
                                     cpl_frame_get_tag(bpmap)));


     }
     /* END case VIS/NIR arm */
  } else {  
     /* BEGIN case UVB arm */
 
     check( xsh_order_split_qth_d2( cen_order_tab_frame, spectralformat_frame, 
                                    &qth_cen_order_tab_frame, &d2_cen_order_tab_frame, instrument));

     check( xsh_instrument_update_lamp( instrument, XSH_LAMP_QTH));
     check( xsh_mflat_recipe_main( qth, on, off,bpmap, master_bias, 
                                   master_dark, qth_cen_order_tab_frame,
                                   cen_order_tab_frame, 
                                   crh_clipping, det_order, backg, 
 				  pre_overscan_corr,
                                  &qth_edges_order_tab_frame, 
                                   &qth_master_flat_frame, 
                                   &qth_rmbackground, &qth_grid_backg,
                                   &qth_frame_backg,instrument));

     check( xsh_add_product_table_multi( qth_edges_order_tab_frame, frameset,
                                         parameters, RECIPE_ID, instrument));
     check( xsh_add_product_pre( qth_master_flat_frame, frameset,
                                 parameters, RECIPE_ID, instrument));
     check( xsh_add_product_image( qth_rmbackground, frameset,
                                   parameters, RECIPE_ID, instrument,
                                   cpl_frame_get_tag(qth_rmbackground)));
     check( xsh_add_product_image( qth_frame_backg, frameset,
                                   parameters, RECIPE_ID, instrument,
                                   cpl_frame_get_tag(qth_frame_backg)));

 
     det_order->min_sn = d2_det_order->min_sn;
     check( xsh_instrument_update_lamp( instrument, XSH_LAMP_D2));
     check( xsh_mflat_recipe_main( d2, on, off,bpmap, master_bias, 
                                   master_dark, d2_cen_order_tab_frame, 
                                   cen_order_tab_frame, 
                                   crh_clipping, det_order, backg, 
 				  pre_overscan_corr,
                                  &d2_edges_order_tab_frame, 
                                   &d2_master_flat_frame, 
                                   &d2_rmbackground,&d2_grid_backg,
                                   &d2_frame_backg,instrument));


     /* merge grid bkg tables for d2 and qth */
     sprintf(fname_d2,"MFLAT_GRID_BACK_D2_%s_%s.fits",
             xsh_instrument_mode_tostring( instrument),
             xsh_instrument_arm_tostring( instrument));

     sprintf(fname_qth,"MFLAT_GRID_BACK_QTH_%s_%s.fits",
             xsh_instrument_mode_tostring( instrument),
             xsh_instrument_arm_tostring( instrument));

     check(grid_tab1=cpl_table_load(fname_d2,1,0));
     check(grid_tab2=cpl_table_load(fname_qth,1,0));
     check(cpl_table_insert(grid_tab1,grid_tab2,cpl_table_get_nrow(grid_tab1)));
     check(plist=cpl_propertylist_load(fname_d2,0));
     sprintf(pcatg,"%s_%s_%s","MFLAT_GRID_BACK", 
             xsh_instrument_mode_tostring( instrument),
             xsh_instrument_arm_tostring( instrument));
     check( xsh_pfits_set_pcatg(plist,pcatg));


     sprintf(fname,"%s.fits",pcatg);
     check(cpl_table_save( grid_tab1, plist, NULL, fname, CPL_IO_DEFAULT));


     grid_backg=xsh_frame_product(fname,pcatg,CPL_FRAME_TYPE_TABLE, 
                                  CPL_FRAME_GROUP_CALIB,CPL_FRAME_LEVEL_FINAL);

     xsh_free_propertylist(&plist);
     xsh_free_table(&grid_tab1);
     xsh_free_table(&grid_tab2);

     check( xsh_add_product_table_multi( d2_edges_order_tab_frame, frameset,
                                         parameters, RECIPE_ID, instrument));
     check( xsh_add_product_pre( d2_master_flat_frame, frameset,
                                 parameters, RECIPE_ID, instrument));

     check( xsh_add_product_image( d2_rmbackground, frameset,
                                   parameters, RECIPE_ID, instrument,
                                   cpl_frame_get_tag(d2_rmbackground)));

     check( xsh_add_product_image( d2_frame_backg, frameset,
                                   parameters, RECIPE_ID, instrument,
                                   cpl_frame_get_tag(d2_frame_backg)));

     check(xsh_add_product_table_multi( grid_backg, frameset, 
                                        parameters, RECIPE_ID, instrument));

 

     /* merge the two master flat */
     check( xsh_instrument_update_lamp( instrument, XSH_LAMP_UNDEFINED));
     /* merge the two master flat */

     check( xsh_flat_merge_qth_d2( qth_master_flat_frame, 
                                   qth_edges_order_tab_frame,
                                   d2_master_flat_frame, 
                                   d2_edges_order_tab_frame,
                                   qth_frame_backg,d2_frame_backg,
                                   &master_flat_frame, &master_bkg_frame, 
                                   &edges_order_tab_frame, instrument));

 
     check( xsh_add_product_table_multi( edges_order_tab_frame, frameset,
                                         parameters, RECIPE_ID, instrument));

 
     xsh_msg("Detect dead-saturated pixels");
     hpix_frm=xsh_image_mflat_detect_hotcold(master_flat_frame,instrument,1);
     cpix_frm=xsh_image_mflat_detect_hotcold(master_flat_frame,instrument,0);
     if(bpmap!=NULL) {

       check(xsh_badpixelmap_coadd(bpmap,hpix_frm));
       check(xsh_badpixelmap_coadd(bpmap,cpix_frm));
     }
    

     check( xsh_add_product_pre( master_flat_frame, frameset,
                                 parameters, RECIPE_ID, instrument));
   
  
     check( xsh_add_product_image( master_bkg_frame, frameset, 
                                   parameters, RECIPE_ID, instrument,
                                   cpl_frame_get_tag(master_bkg_frame)));

  
    
     check( xsh_add_product_image( cpix_frm, frameset, 
				   parameters, RECIPE_ID, instrument,
				   cpl_frame_get_tag(cpix_frm)));

     check( xsh_add_product_image( hpix_frm, frameset, 
				   parameters, RECIPE_ID, instrument,
				   cpl_frame_get_tag(hpix_frm)));

     
     if(bpmap!=NULL) {
        check( xsh_add_product_bpmap(bpmap, frameset, parameters,
                                     RECIPE_ID, instrument, 
                                     cpl_frame_get_tag(bpmap)));

     }
     /* END case UVB arm */
  } 
  xsh_msg("xsh_mflat success!!");

  cleanup:
    xsh_end( RECIPE_ID, frameset, parameters);
    XSH_FREE( crh_clipping);
    XSH_FREE( det_order);
    XSH_FREE( d2_det_order);
    XSH_FREE( backg);
    xsh_free_frameset( &raws);
    xsh_free_frameset( &calib);
    xsh_free_frameset( &on);
    xsh_free_frameset( &off);
    xsh_free_frameset( &qth);
    xsh_free_frameset( &d2);
 
    xsh_free_frame( &qth_cen_order_tab_frame);
    xsh_free_frame( &qth_edges_order_tab_frame);
    xsh_free_frame( &qth_master_flat_frame);

 
    xsh_free_frame( &d2_cen_order_tab_frame);
    xsh_free_frame( &d2_edges_order_tab_frame);
    xsh_free_frame( &d2_master_flat_frame);
 

    xsh_free_frame( &edges_order_tab_frame);
    xsh_free_frame( &master_flat_frame);
    xsh_free_frame( &master_bkg_frame);
 
    xsh_free_frame( &rmbackground);
    xsh_free_frame( &d2_rmbackground);
    xsh_free_frame( &qth_rmbackground);
    xsh_free_frame( &grid_backg);
    xsh_free_frame( &d2_grid_backg);
    xsh_free_frame( &qth_grid_backg);
 
    xsh_free_frame( &frame_backg);
    //To free the following generates seg fault
    //xsh_free_frame( &d2_frame_backg);
    xsh_free_frame( &qth_frame_backg);
 
    xsh_instrument_free( &instrument);

    xsh_free_propertylist(&plist);
 
    xsh_free_table(&grid_tab1);
    xsh_free_table(&grid_tab2);



    return;
}


/**@}*/

