/*                                                                           *
 *   This file is part of the ESO X-shooter Pipeline                         *
 *   Copyright (C) 2006 European Southern Observatory                        *
 *                                                                           *
 *   This library is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA    *
 *                                                                           */

/*
 * $Author: amodigli $
 * $Date: 2011/03/07 09:47:06 $
 * $Revision: 1.42 $
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*---------------------------------------------------------------------------*/
/**
 * @defgroup xsh_compute_response   Compute Instrument Response (xsh_compute_response)
 * @ingroup drl_functions
 *
 */
/*---------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
  Includes
  ----------------------------------------------------------------------------*/

#include <xsh_dfs.h>
#include <xsh_error.h>
#include <xsh_msg.h>
#include <cpl.h>
#include <string.h>
#include <time.h>
#include <xsh_utils_table.h>
#include <xsh_data_star_flux.h>
#include <xsh_data_atmos_ext.h>
#include <xsh_data_spectrum.h>
#include <xsh_pfits.h>
#include <xsh_utils.h>
#include <xsh_drl.h>
#include <xsh_utils_efficiency.h>

/*----------------------------------------------------------------------------
  Function prototypes
  ----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
                              Implementation
 -----------------------------------------------------------------------------*/

/* Definition of the regions with a too high absorption; in these regions
   the response is computed by a simple linear interpolation between the
   valuse at the edges of the region
*/
typedef struct {
  double lambda_min ;
  double lambda_max ;
} HIGH_ABS_REGION ;

/*
HIGH_ABS_REGION NirHighAbsRegions[] = {
  {1090., 1170.}, // nm
  {1340., 1500.}, // nm
  {1800., 2100.}, // nm
  {0., 0.} 
} ;
*/

HIGH_ABS_REGION NirHighAbsRegions[] = {
  {1060.0, 1232.9}, // nm
  {1247.4, 1500.0}, // nm
  {1549.0, 1588.5}, // nm
  {1595.0, 1625.0}, // nm
  {1630.0, 1983.5}, // nm
  {1984.2, 1988.1}, // nm
  {1988.9, 1991.1}, // nm
  {1991.7, 2138.0}, // nm
  {2145.0, 2191.0}, // nm
  {2218.0, 2271.0}, // nm
  {2351.0, 2431.0}, // nm
  {0., 0.} 
} ;

/*
HIGH_ABS_REGION VisHighAbsRegions[] = {
  {685., 695.}, // nm
  {755., 770.}, // nm
  {0., 0.}
} ;
*/
HIGH_ABS_REGION VisHighAbsRegions[] = {
  {585.0, 604.0}, // nm
  {684.5, 746.0}, // nm
  {755.5, 775.5}, // nm
  {783.7, 856.0}, // nm
  {887.5, 993.4}, // nm
  {0., 0.}
} ;

#define  INTERPOL_WSTEP_NM 2

static int 
find_lambda_idx( double lambda, double * wave, int from, int to,
                 double step )
{
  int idx ;
  double * pwave ;

  for( idx = from, pwave = wave+from ; idx < to ; pwave++, idx++ ) {
    if ( *pwave >= (lambda-step/2) && *pwave < (lambda+step/2) ) return idx ;
  }

  return -1 ;
}

static void 
find_lambda_idx_limit( double min, double max, double * spectrum,
                       int from, int to, double step,
                       int * if0, int * if1 )
{
  int idx ;
  double *pspec ;

  xsh_msg_dbg_high( "From: %d, To: %d, step: %lf, min: %lf, max: %lf",
		    from, to, step, min, max ) ;

  for( idx = from, pspec = spectrum+from ; idx < to ; pspec++, idx++ ) {
    xsh_msg_dbg_high( "    Search Min - Spectrum_lambda: %lf (%lf)", *pspec,
		      min-(step/2.) ) ;
    if ( *pspec >= (min-(step/2.))  ) break ;
  }
  *if0 = idx ;
  for( ; idx < to ; pspec++, idx++ ) {
    xsh_msg_dbg_high( "    Search Max - Spectrum_lambda: %lf", *pspec ) ;
    if ( *pspec > (max+(step/2.)) ) break ;
  }
  *if1 = idx-1 ;

  xsh_msg_dbg_low( "Lambda Min: %lf, Max: %lf (in [%lf,%lf[",
                   spectrum[*if0], spectrum[*if1], min, max ) ;
  if ( *if1 > to ) *if1 = to ;

  return ;
}

/** 
 * Interpolate atms extinction at same sampling step as ref std star 
 * 
 * @param star_list Table of ref star flux
 * @param[in] atmos_ext_tab atmospheric extinction table
 * @param[in] instrument instrument arm and lamp setting
 * @param[out] atmos_lambda interpolated wave array
 * @param[out] atmos_K      interpolated atmospheric extinction K
 * @return cpl error code
 */
static cpl_error_code 
xsh_interpolate_atm_ext(xsh_star_flux_list * star_list,
                        cpl_table*atmos_ext_tab,
                        xsh_instrument* instrument, 
                        double** atmos_lambda, 
                        double** atmos_K)
{

   int nrow=0;
   double* pk=0;
   double* pw=0;
   int j=0 ;
   int nlambda_star=0 ;
   double * lambda_star = NULL;

   XSH_ASSURE_NOT_NULL_MSG (star_list,"null input ref std flux table!");
   XSH_ASSURE_NOT_NULL_MSG (atmos_ext_tab,"null input atm ext table!");
 
   nlambda_star = star_list->size ;
   lambda_star = star_list->lambda ;
   
   /* Fill the atmos_lambda and K with only the subset of lambdas avalable
      in star list (the std star) */
   XSH_CALLOC( *atmos_lambda, double, nlambda_star ) ;
   XSH_CALLOC( *atmos_K, double, nlambda_star ) ;
   
   cpl_table_cast_column(atmos_ext_tab,"LA_SILLA","K",CPL_TYPE_DOUBLE);
   cpl_table_cast_column(atmos_ext_tab,"LAMBDA","WAVE",CPL_TYPE_DOUBLE);
     
   nrow=cpl_table_get_nrow(atmos_ext_tab);
   pw=cpl_table_get_data_double(atmos_ext_tab,"WAVE");
   pk=cpl_table_get_data_double(atmos_ext_tab,"K");
 
   if( xsh_instrument_get_arm(instrument) != XSH_ARM_NIR){
      for( j = 0 ; j<nlambda_star ; j++) {
 
         (*atmos_lambda)[j] = lambda_star[j] ;
         (*atmos_K)[j] = xsh_data_interpolate((*atmos_lambda)[j],nrow,pw,pk);
 
         //xsh_msg("w=%g k=%g",atmos_lambda[j],atmos_K[j]);
      }
   } else {
      for( j = 0 ; j<nlambda_star ; j++) {
 
         (*atmos_lambda)[j] = lambda_star[j] ;
         (*atmos_K)[j] = 0;
 
         //xsh_msg("w=%g k=%g",atmos_lambda[j],atmos_K[j]);
      }
   }
 
   cpl_table_erase_column(atmos_ext_tab,"WAVE");
   cpl_table_erase_column(atmos_ext_tab,"K");
 

  cleanup:
   return cpl_error_get_code();
}

/** 
 * Interpolate high absorbtion regions 
 * 
 * @param[in] star_list Table of ref star flux
 * @param[out] resp_list Table of response values
 * @param[in] the_arm      interpolated atmospheric extinction K
 * @return cpl error code 
 */
static cpl_error_code 
xsh_interpolate_high_abs_regions(xsh_star_flux_list * star_list,
                                 xsh_star_flux_list * resp_list,
                                 HIGH_ABS_REGION * phigh,
                                 XSH_ARM the_arm)

{
   int kh=0;
   int min_idx=0;
   int max_idx=0;
   double min_flux=0;
   double max_flux=0;
   double delta_flux=0;
   double cur_flux=0;
   int ll=0;
   int nlambda_star=0;
   double step_star=0;
   double * lambda_star = NULL;
   double lambda_star_min=0;
   double lambda_star_max=0;
   //double* pwmin=NULL;
   //double* pwmax=NULL;
   //int nrow=0;

   XSH_ASSURE_NOT_NULL_MSG (star_list,"null input ref std flux table!");
   XSH_ASSURE_NOT_NULL_MSG (resp_list,"null input response table!");

   nlambda_star = star_list->size ;
   lambda_star = star_list->lambda ;
   lambda_star_min = *lambda_star ;
   lambda_star_max = *(lambda_star+nlambda_star-1) ;
   step_star = (lambda_star_max-lambda_star_min)/(nlambda_star-1) ;

   /*
   if(high_abs_tab!=NULL) {
      nrow=cpl_table_get_nrow(high_abs_tab);
      check(pwmin=cpl_table_get_data_double(high_abs_tab,"LAMBDA_MIN"));
      check(pwmax=cpl_table_get_data_double(high_abs_tab,"LAMBDA_MAX"));

      phigh = (HIGH_ABS_REGION *) cpl_calloc(nrow + 1, sizeof(HIGH_ABS_REGION));
      for(kh=0;kh<nrow;kh++) {
         phigh[kh].lambda_min=pwmin[kh];
         phigh[kh].lambda_max=pwmax[kh];
      }
      phigh[nrow].lambda_min=0;
      phigh[nrow].lambda_max=0;

   } else {

      // Now specific to high absorption region (linear interpolation)
      //   NIR and VIS only
      
      if ( the_arm == XSH_ARM_VIS ) {
         // Specific VIS 
         phigh = VisHighAbsRegions ;
      }
      else if ( the_arm == XSH_ARM_NIR ) {
         // Specific NIR 
         phigh = NirHighAbsRegions ;
      }

   }
   */
   if ( phigh != NULL ) {

      xsh_msg_dbg_medium( "Interpolate in High Absorption Regions" ) ;
      for( kh = 0 ; phigh->lambda_min != 0. ; kh++, phigh++ ) {

         min_idx = find_lambda_idx( phigh->lambda_min, resp_list->lambda,
                                    0, nlambda_star, step_star ) ;
         max_idx = find_lambda_idx( phigh->lambda_max, resp_list->lambda,
                                    min_idx, nlambda_star, step_star ) ;
         xsh_msg_dbg_medium( "Indexes[%d]: %d (lambda=%lf), %d (lambda=%lf)", 
                             kh,min_idx, phigh->lambda_min, max_idx, 
                             phigh->lambda_max ) ;

         if(min_idx>=resp_list->size || max_idx>=resp_list->size) {
            xsh_msg("min_idx=%d max_idx=%d size=%d",
                    min_idx,max_idx,resp_list->size);
         }

         min_flux = resp_list->flux[min_idx] ;
         max_flux = resp_list->flux[max_idx] ;
         xsh_msg_dbg_medium( "Min Flux: %le, Max Flux: %le", 
                             min_flux, max_flux ) ;
         delta_flux = (max_flux-min_flux)/(max_idx-min_idx) ;
         // Now interpolate 
         cur_flux = min_flux ;
         for ( ll = min_idx+1 ; ll < max_idx ; ll++ ) {
            cur_flux += delta_flux ;
            xsh_msg_dbg_medium( " ==> Interpolate at %lf: %le replaced by %le",
                                resp_list->lambda[ll],
                                resp_list->flux[ll], cur_flux ) ;
            resp_list->flux[ll] = cur_flux ;
         }
      }
   }

  cleanup:


   return cpl_error_get_code();
}

/** 
 * Integrate the flux on the same sampling that the star_list, taking into
 * account the bad pixels, apply atmos correction, and divide by the star_list
 * table.
 * 
 * @param resp_list Table of response values
 * @param star_list Table of ref star flux
 * @param lambda_spectrum  array with lambda values
 * @param flux_spectrum    array with flux values
 * @param flux_added       array with integrated flux values
 * 
 * @return cpl error code
 */
static cpl_error_code
xsh_response_crea_ascii(xsh_star_flux_list * resp_list,
                        xsh_star_flux_list * star_list,
                        double * lambda_spectrum,
                        double * flux_spectrum,
                        double * flux_added)
{

   FILE * fout ;
   int i=0;
   int nlambda_star=0;
   int nlambda_spectrum=0;
   double * lambda_star = NULL;
   double * flux_star = NULL ;

   XSH_ASSURE_NOT_NULL_MSG (star_list,"null input ref std flux table!");
   XSH_ASSURE_NOT_NULL_MSG (resp_list,"null input response table!");

   nlambda_star = star_list->size ;
   lambda_star = star_list->lambda ;
   flux_star = star_list->flux ;

   fout = fopen( "summed.dat", "w" ) ;
   for( i = 0 ; i < nlambda_star ; i++ ) {
      fprintf( fout, "%lf %le\n", resp_list->lambda[i],
               flux_added[i] ) ;
   }
   fclose( fout ) ;

   fout = fopen( "response.dat", "w" ) ;
   for( i = 0 ; i < nlambda_star ; i++ ) {
      fprintf( fout, "%lf %le\n", resp_list->lambda[i],
               resp_list->flux[i] ) ;
   }
   fclose( fout ) ;

   fout = fopen( "spectrum.dat", "w" ) ;
   for ( i = 0 ; i < nlambda_spectrum ; i++ )
      fprintf( fout, "%lf %lf\n", lambda_spectrum[i], flux_spectrum[i] ) ;
   fclose( fout ) ;

   fout = fopen( "star.dat", "w" ) ;
   for ( i = 0 ; i < nlambda_star ; i++ )
      fprintf( fout, "%lf %le\n", lambda_star[i], flux_star[i] ) ;
   fclose( fout ) ;

  cleanup:
   return cpl_error_get_code();
}




static cpl_error_code
xsh_correct_flux_for_badpix(int npix_in_interval,
                            double * flux_spectrum,
                            int * qual_spectrum,
                            int if0,int if1,double exptime,
                            int i,double** flux_added,int* npixels,int* nbad)

{

   double * pf0=NULL;
   double * pf1=NULL;
   double * pf=NULL;

   int * qf0=NULL;
   int * qf=NULL;

   XSH_ASSURE_NOT_NULL_MSG (flux_spectrum,"null flux_spectrum !");
   XSH_ASSURE_NOT_NULL_MSG (qual_spectrum,"null qual_spectrum !");

   if ( npix_in_interval != if1-if0 ) *nbad = npix_in_interval -(if1-if0) ;

   pf0 = flux_spectrum + if0 ;
   pf1 = flux_spectrum + if1 ;
   qf0 = qual_spectrum + if0 ;
   for( pf = pf0, qf = qf0 ; pf < pf1 ; pf++, qf++ ) {
      /* TBD: handle correctly the borders (for example with a step_spectrum
         of 0.2, and a step_star of 5, one should use only half of the
         flux on the edges
      */
      (*npixels)++ ;
      if ( *qf != QFLAG_GOOD_PIXEL ) {
         (*nbad)++ ;
      }
      else (*flux_added)[i] += *pf/exptime ;
   }

   if ( nbad != 0 )
      (*flux_added)[i] /= (double)(npix_in_interval-*nbad)/
         (double)npix_in_interval ;

  cleanup:
   return cpl_error_get_code();
}

static xsh_star_flux_list *
xsh_response_calculate(
            xsh_star_flux_list * star_list,
            xsh_spectrum * spectrum,
            cpl_table * atmos_ext_tab,
            XSH_ARM  the_arm,
            double * atmos_K,
            double airmass,
            double exptime,
            double gain)
{ 
   xsh_star_flux_list * resp_list=NULL;
   int npixels = 0;
   int nbad = 0 ;

   double lambda=0;
   double min=0;
   double max=0;
   double kvalue=0;


   double * lambda_resp = NULL ;
   double * flux_resp = NULL ;
   double * flux_added = NULL ;
   double * lambda_star = NULL;
   double  * flux_star = NULL ;

   double * lambda_spectrum = NULL;
   double * plambda=NULL ;
   double * flux_spectrum = NULL ;
   int * qual_spectrum = NULL ;

   int i=0;
   int if0=0;
   int if1=0;
   int  nlambda_star=0;
   int nlambda_spectrum=0;
   double lambda_spectrum_min=0;
   double lambda_spectrum_max=0;
   double lambda_star_min=0;
   double lambda_star_max=0;
   double step_spectrum=0 ;
   double step_star=0 ;
   int npix_in_interval=0 ;
   int binx, biny ;
   cpl_frame* resp_integrated=NULL;

   check( nlambda_spectrum = xsh_spectrum_get_size( spectrum ) ) ;
   check( flux_spectrum = xsh_spectrum_get_flux( spectrum ) ) ;
   check( qual_spectrum = xsh_spectrum_get_qual( spectrum ) ) ;
   check( lambda_spectrum_min = xsh_spectrum_get_lambda_min( spectrum ) ) ;
   check( lambda_spectrum_max = xsh_spectrum_get_lambda_max( spectrum ) ) ;
   check( step_spectrum = xsh_spectrum_get_lambda_step( spectrum ) ) ;

   /* In NIR, no binning ==> set 1x1 */
   if ( the_arm != XSH_ARM_NIR ) {
      binx = xsh_pfits_get_binx( spectrum->flux_header ) ;
      biny = xsh_pfits_get_biny( spectrum->flux_header ) ;
   }
   else {
      binx = 1 ;
      biny = 1 ;
   }
 
   flux_star = star_list->flux ;
   nlambda_star = star_list->size ;
   lambda_star = star_list->lambda ;
   lambda_star_min = *lambda_star ;
   lambda_star_max = *(lambda_star+nlambda_star-1) ;
   step_star = (lambda_star_max-lambda_star_min)/(nlambda_star-1) ;

   xsh_msg_dbg_low( "Spectrum  - Nlambda: %d, Min: %lf, Max: %lf, Step: %lf",
                    nlambda_spectrum, lambda_spectrum_min, lambda_spectrum_max,
                    step_spectrum ) ;
   xsh_msg_dbg_low( "           - BinX: %d, BinY: %d", binx, biny ) ;

   XSH_CALLOC( lambda_resp, double, nlambda_star ) ;
   XSH_CALLOC( flux_resp, double, nlambda_star ) ;
   XSH_CALLOC( flux_added, double, nlambda_star ) ;
   /* For each lambda in star_list
      if there is such a lambda in rec_list Then
      integrate (add) the flux from (lambda-delta/2) to (lambda+delta/2)
      calculate ratio rec_add_flux/star_flux
      endif
      EndFor
   */
   /* First create lambda_spectrum array */
   XSH_CALLOC( lambda_spectrum, double, nlambda_spectrum ) ;
    
   //xsh_msg("lambda min=%g lambda max=%g",lambda_spectrum_min,lambda_spectrum_max);

   for( lambda = lambda_spectrum_min, plambda = lambda_spectrum, i = 0 ;
        i<nlambda_spectrum ; 
        i++, plambda++, lambda += step_spectrum ) 
   {
      *plambda = lambda ;
   }
  
   /* Create response list (identical to star_list ?) */
   check( resp_list = xsh_star_flux_list_create( nlambda_star ) ) ;
   //xsh_msg("step_star=%g step_spectrum=%g", step_star,step_spectrum);
  
   npix_in_interval = step_star/step_spectrum ;
   xsh_msg_dbg_low( "Nb of pixels max in interval: %d", npix_in_interval ) ;
      xsh_print_rec_status(0);

   /* Loop over lamba_star */
   for ( i = 0 ; i < nlambda_star ; i++ ) {
      /* Now add spectrum flux in the star sampling interval */

      flux_added[i] = 0. ;
      min = lambda_star[i] - step_star/2 ;
      max = lambda_star[i] + step_star/2 ;

      xsh_msg_dbg_medium( "Lambda_star[%d] = %lf, Min = %lf, Max = %lf", i,
                          lambda_star[i], min, max ) ;

      find_lambda_idx_limit( min, max, lambda_spectrum, if0, nlambda_spectrum,
			     step_spectrum, &if0, &if1 ) ;

      xsh_msg_dbg_low( "    Actual nb of pixels in interval: %d/%d", if1-if0,
                       npix_in_interval ) ;

      /* correct integrated flux for bad pixels */
      xsh_correct_flux_for_badpix(npix_in_interval,flux_spectrum,qual_spectrum,
                                  if0,if1,exptime,i,&flux_added,&npixels,&nbad);

      /* initialize respons correcting also by gain (response in electrons) */
      resp_list->flux[i] = flux_star[i]/(flux_added[i]/gain) ;
      resp_list->lambda[i] = lambda_star[i] ;
      xsh_msg_dbg_low("  Flux Star = %le, Integrated = %lf (%d pixels, %d bad)",
                      flux_star[i], flux_added[i], npixels, nbad ) ;
      xsh_msg_dbg_low("   =====> Response at Lambda=%lf: %le",
                      lambda_star[i], resp_list->flux[i] ) ;

      /* Multiply by atmos ext */
      if ( atmos_ext_tab != NULL ) {
         kvalue = pow(10., -airmass*atmos_K[i]*0.4 ) ;
         resp_list->flux[i] *= kvalue ;
         //xsh_msg("resp: %g kvalue %g k %g",resp_list->flux[i],kvalue,atmos_K[i]);
      }

      /* Multiply by ratio star_binning/spectrum_binning */
      resp_list->flux[i] *= step_star/step_spectrum ;

      if0 = if1 ;
   }
   /* filter median response to smooth out local jumps */
   xsh_star_flux_list_filter_median(resp_list,3);

   resp_integrated=xsh_star_flux_list_save(resp_list,"resp_list_integrated.fits","TEST") ;
   xsh_free_frame(&resp_integrated);

   if ( xsh_debug_level_get() >= XSH_DEBUG_LEVEL_LOW ) {
      check(xsh_response_crea_ascii(resp_list,star_list,lambda_spectrum,
                                    flux_spectrum,flux_added));
   }

  cleanup:

   XSH_FREE( lambda_spectrum ) ;
   XSH_FREE( lambda_resp ) ;
   XSH_FREE( lambda_resp ) ;
   XSH_FREE( flux_resp ) ;
   XSH_FREE( flux_added ) ;

   if(cpl_error_get_code() == CPL_ERROR_NONE) {
      return resp_list;
   } else {
      return NULL;
   }
}

/*
static xsh_star_flux_list *
xsh_response_calculate2(
            xsh_star_flux_list * star_list,
            xsh_spectrum * spectrum,
            cpl_table * atmos_ext_tab,
            XSH_ARM  the_arm,
            double * atmos_K,
            double airmass,
            double exptime,
            double gain)
{ 
   xsh_star_flux_list * resp_list=NULL;

   double lambda=0;
   double min=0;
   double max=0;
   double kvalue=0;


   double * lambda_resp = NULL ;
   double * flux_resp = NULL ;
   double * flux_added = NULL ;
   double * lambda_star = NULL;
   double  * flux_star = NULL ;

   double * lambda_spectrum = NULL;
   double * plambda=NULL ;
   double * flux_spectrum = NULL ;
   int * qual_spectrum = NULL ;

   int i=0;
   int  nlambda_star=0;
   int nlambda_spectrum=0;
   double lambda_spectrum_min=0;
   double lambda_spectrum_max=0;
   double lambda_star_min=0;
   double lambda_star_max=0;
   double step_spectrum=0 ;
   double step_star=0 ;
   int npix_in_interval=0 ;
   int binx, biny ;

   check( nlambda_spectrum = xsh_spectrum_get_size( spectrum ) ) ;
   check( flux_spectrum = xsh_spectrum_get_flux( spectrum ) ) ;
   check( qual_spectrum = xsh_spectrum_get_qual( spectrum ) ) ;
   check( lambda_spectrum_min = xsh_spectrum_get_lambda_min( spectrum ) ) ;
   check( lambda_spectrum_max = xsh_spectrum_get_lambda_max( spectrum ) ) ;
   check( step_spectrum = xsh_spectrum_get_lambda_step( spectrum ) ) ;

   // In NIR, no binning ==> set 1x1 
   if ( the_arm != XSH_ARM_NIR ) {
      binx = xsh_pfits_get_binx( spectrum->flux_header ) ;
      biny = xsh_pfits_get_biny( spectrum->flux_header ) ;
   }
   else {
      binx = 1 ;
      biny = 1 ;
   }
 
   flux_star = star_list->flux ;
   nlambda_star = star_list->size ;
   lambda_star = star_list->lambda ;
   lambda_star_min = *lambda_star ;
   lambda_star_max = *(lambda_star+nlambda_star-1) ;
   step_star = (lambda_star_max-lambda_star_min)/(nlambda_star-1) ;

   xsh_msg_dbg_low( "Spectrum  - Nlambda: %d, Min: %lf, Max: %lf, Step: %lf",
                    nlambda_spectrum, lambda_spectrum_min, lambda_spectrum_max,
                    step_spectrum ) ;
   xsh_msg_dbg_low( "           - BinX: %d, BinY: %d", binx, biny ) ;

   XSH_CALLOC( lambda_resp, double, nlambda_star ) ;
   XSH_CALLOC( flux_resp, double, nlambda_star ) ;
   XSH_CALLOC( flux_added, double, nlambda_star ) ;
   // For each lambda in star_list
   //   if there is such a lambda in rec_list Then
   //     integrate (add) the flux from (lambda-delta/2) to (lambda+delta/2)
   //     calculate ratio rec_add_flux/star_flux
   //   endif
   // EndFor
   
   // First create lambda_spectrum array 
   XSH_CALLOC( lambda_spectrum, double, nlambda_spectrum ) ;
    
   //xsh_msg("lambda min=%g lambda max=%g",lambda_spectrum_min,lambda_spectrum_max);

   for( lambda = lambda_spectrum_min, plambda = lambda_spectrum, i = 0 ;
        i<nlambda_spectrum ; 
        i++, plambda++, lambda += step_spectrum ) 
   {
      *plambda = lambda ;
   }
  
   // Create response list (identical to star_list ?) 
   check( resp_list = xsh_star_flux_list_create( nlambda_star ) ) ;
   //xsh_msg("step_star=%g step_spectrum=%g", step_star,step_spectrum);
  
   npix_in_interval = step_star/step_spectrum ;
   xsh_msg_dbg_low( "Nb of pixels max in interval: %d", npix_in_interval ) ;
   xsh_print_rec_status(0);

   int i_start=0;
   int i_end=0;
   int i1_inf=0;
   int i1_sup=0;
   int i2_inf=0;
   int i2_sup=0;

   // Loop over lamba_star 
   for ( i = 0 ; i < nlambda_star ; i++ ) {
      // Now add spectrum flux in the star sampling interval 

      flux_added[i] = 0. ;
      min = lambda_star[i] - step_star/2 ;
      max = lambda_star[i] + step_star/2 ;
      i_start=0;
      i_end=nlambda_spectrum ;
     
      xsh_util_get_infsup(lambda_star,min,i_start,i_end,&i1_inf,&i1_sup);
      xsh_util_get_infsup(lambda_star,max,i_start,i_end,&i2_inf,&i2_sup);

      flux_added[i]=xsh_spectrum_integrate(flux_spectrum,lambda_spectrum,
                                           i1_inf,i1_sup,i2_inf,i2_sup,
                                           lambda_star[i],step_star);

      // initialize respons correcting also by gain (response in electrons) 
      resp_list->flux[i] = flux_star[i]/(flux_added[i]/gain) ;
      resp_list->lambda[i] = lambda_star[i] ;

      // Multiply by atmos ext 
      if ( atmos_ext_tab != NULL ) {
         kvalue = pow(10., -airmass*atmos_K[i]*0.4 ) ;
         resp_list->flux[i] *= kvalue ;
         //xsh_msg("resp: %g kvalue %g k %g",resp_list->flux[i],kvalue,atmos_K[i]);
      }

      // Multiply by ratio star_binning/spectrum_binning 
      resp_list->flux[i] *= step_star/step_spectrum ;


   }

   xsh_star_flux_list_save(resp_list,"resp_list_integrated.fits","TEST") ;


   if ( xsh_debug_level_get() >= XSH_DEBUG_LEVEL_LOW ) {
      check(xsh_response_crea_ascii(resp_list,star_list,lambda_spectrum,
                                    flux_spectrum,flux_added));
   }
   
  cleanup:

   XSH_FREE( lambda_spectrum ) ;
   XSH_FREE( lambda_resp ) ;
   XSH_FREE( lambda_resp ) ;
   XSH_FREE( flux_resp ) ;
   XSH_FREE( flux_added ) ;

   if(cpl_error_get_code() == CPL_ERROR_NONE) {
      return resp_list;
   } else {
      return NULL;
   }
}
*/

/** 
 * Integrate the flux on the same sampling that the star_list, taking into
 * account the bad pixels, apply atmos correction, and divide by the star_list
 * table.
 * 
 * @param star_list Table of ref star flux
 * @param spectrum 1./CPL/DOC/cpl_um_2.0.1.psD merged spectrum of standard star
 * @param atmos_ext_table Atmos Extinction table
 * @param airmass observed object airmass mean
 * @param exptime Exposure time
 * @param gain    detector's gain
 * @param instrument instrument arm and lamp setting
 * 
 * @return The computed response frame (table)
 */
static xsh_star_flux_list * 
do_compute( xsh_star_flux_list * star_list,
            xsh_spectrum * spectrum,
            cpl_table * atmos_ext_tab,
            HIGH_ABS_REGION * phigh,
            double airmass,
            double exptime,
            double gain,
            xsh_instrument * instrument )
{
   xsh_star_flux_list * resp_list = NULL ;

   int nlambda_star ;
   double step_spectrum ;
   double step_star ;
   double * lambda_star = NULL, * flux_star = NULL ;
   double lambda_star_min, lambda_star_max ;
   double * flux_spectrum = NULL ;
   int * qual_spectrum = NULL ;
   double lambda_spectrum_min, lambda_spectrum_max ;
   double * atmos_lambda = NULL ;
   double * atmos_K = NULL ; 
   XSH_ARM the_arm ;

   XSH_ASSURE_NOT_NULL( star_list ) ;
   XSH_ASSURE_NOT_NULL( spectrum ) ;
   /*
     get Nb of lambda, etc in spectrum.
   */
   check( flux_spectrum = xsh_spectrum_get_flux( spectrum ) ) ;
   check( qual_spectrum = xsh_spectrum_get_qual( spectrum ) ) ;
   check( lambda_spectrum_min = xsh_spectrum_get_lambda_min( spectrum ) ) ;
   check( lambda_spectrum_max = xsh_spectrum_get_lambda_max( spectrum ) ) ;
   check( step_spectrum = xsh_spectrum_get_lambda_step( spectrum ) ) ;
   the_arm=xsh_instrument_get_arm(instrument);
   /*
     get Nb of lambda in star_list.
   */
   nlambda_star = star_list->size ;
   lambda_star = star_list->lambda ;
   flux_star = star_list->flux ;
   lambda_star_min = *lambda_star ;
   lambda_star_max = *(lambda_star+nlambda_star-1) ;
   step_star = (lambda_star_max-lambda_star_min)/(nlambda_star-1) ;
   xsh_msg_dbg_low( "Star - Nlambda: %d, Min: %le, Max: %le, Step: %lf",
                    nlambda_star,
                    lambda_star_min, lambda_star_max, step_star ) ;

   check(xsh_interpolate_atm_ext(star_list,atmos_ext_tab,instrument, 
                                 &atmos_lambda,&atmos_K));

   check(resp_list=xsh_response_calculate(star_list,spectrum,atmos_ext_tab,
                                          the_arm,atmos_K,airmass,
                                          exptime,gain));

   /* linear interpolation of high absorbtion regions (NIR and VIS only) */
   check(xsh_interpolate_high_abs_regions(star_list,resp_list,phigh,the_arm));

  cleanup:
   xsh_msg_dbg_medium( "End compute_response" ) ;
   XSH_FREE( atmos_lambda ) ;
   XSH_FREE( atmos_K ) ;

   return resp_list ;

}


static HIGH_ABS_REGION *
xsh_fill_high_abs_regions(
   xsh_instrument* instrument,
   cpl_frame* high_abs_frame)
{
   HIGH_ABS_REGION * phigh=NULL;
   int nrow=0;
   double* pwmin=0;
   double* pwmax=0;
   int i=0;
   cpl_table* high_abs_tab=NULL;
   XSH_ARM the_arm;

  if(high_abs_frame !=NULL) {
     high_abs_tab=cpl_table_load(cpl_frame_get_filename(high_abs_frame),1,0);
  }
  the_arm=xsh_instrument_get_arm(instrument);



   if(high_abs_tab!=NULL) {
      nrow=cpl_table_get_nrow(high_abs_tab);
      check(pwmin=cpl_table_get_data_double(high_abs_tab,"LAMBDA_MIN"));
      check(pwmax=cpl_table_get_data_double(high_abs_tab,"LAMBDA_MAX"));

      phigh = (HIGH_ABS_REGION *) cpl_calloc(nrow + 1, sizeof(HIGH_ABS_REGION));
      for(i=0;i<nrow;i++) {
         phigh[i].lambda_min=pwmin[i];
         phigh[i].lambda_max=pwmax[i];
      }
      phigh[nrow].lambda_min=0;
      phigh[nrow].lambda_max=0;

   } else {

      /* Now specific to high absorption region (linear interpolation)
         NIR and VIS only
      */
      if ( the_arm == XSH_ARM_VIS ) {
         /* Specific VIS */
         phigh = VisHighAbsRegions ;
      }
      else if ( the_arm == XSH_ARM_NIR ) {
         /* Specific NIR */
         phigh = NirHighAbsRegions ;
      }

   }
  cleanup:
   xsh_free_table(&high_abs_tab);
   return phigh;
}


/** 
 * Compute the instrument response using the rectified frame and the
 * Standard star flux frame. The input should include also the
 * Master Response Function and the Athmospheric Extinction Curve. Also
 * should be produced the Efficiency frame.
 * 
 * @param spectrum_frame Rectified frame (1D spectrum)
 * @param flux_std_star_cat_frame Standard star flux catalog frame
 * @param atmos_ext_frame Atmospheric Extinction Frame
 * @param instrument Instrument strcture
 * @param exptime Exposure time
 * 
 * @return Frame of the computed response
 */

cpl_frame * xsh_compute_response( cpl_frame * spectrum_frame,
				  cpl_frame * flux_std_star_cat_frame,
				  cpl_frame * atmos_ext_frame,
                                  cpl_frame* high_abs_frame,
				  xsh_instrument * instrument,
				  double exptime )
{
  cpl_frame * result = NULL ;	/**< computed response frame */
  xsh_star_flux_list * star_list = NULL ; /**< Std star flux */
  xsh_spectrum * spectrum = NULL ; /**< 1d spectrum */
  cpl_frame * tmp_spectrum_frame = NULL ; /**< 1d spectrum */
  xsh_star_flux_list * resp_list = NULL ;
  xsh_atmos_ext_list * atmos_ext_list = NULL ;
  char tag[80];
  char fname[80];

  cpl_frame* flux_std_star_frame=NULL;
  double dRA=0;
  double dDEC=0;
  cpl_table* tbl_ref_std_spectrum=NULL;
  const char* tbl_ref_std_fname="ref_flux_std_star.fits";
  cpl_frame* ipol_flux_std_star_frame=NULL;
  cpl_frame* resampl_flux_std_star_frame=NULL;
  cpl_frame* smooth_flux_std_star_frame=NULL;

  double wmin=0;
  double wmax=0;
  int naxis1=0;
  double cdelt1=0;
  const char* filename=NULL;
  cpl_propertylist* plist=NULL;
  char * name=NULL;
  char * pcatg=NULL;
  cpl_table* atmos_ext_tab=NULL;
  double airmass=0;
  double gain=1.;
   HIGH_ABS_REGION * phigh=NULL;

  XSH_ASSURE_NOT_NULL( spectrum_frame ) ;
  XSH_ASSURE_NOT_NULL( flux_std_star_cat_frame ) ;
  XSH_ASSURE_NOT_NULL( instrument ) ;
 

  xsh_msg("Compute instrument response");

  /* get RA DEC from observed object frame */
  xsh_frame_sci_get_ra_dec_airmass(spectrum_frame,&dRA,&dDEC,&airmass);

  /* extract matching ref STD star spectrum */
  if(CPL_ERROR_NONE != xsh_parse_catalog_std_stars(flux_std_star_cat_frame,
                                                    dRA,dDEC,
                                                    STAR_MATCH_DEPSILON,
                                                    &tbl_ref_std_spectrum)){
     cpl_error_reset();
     xsh_free_table(&tbl_ref_std_spectrum);
     return NULL;
  }

  check(cpl_table_save(tbl_ref_std_spectrum,NULL,NULL,tbl_ref_std_fname,
                       CPL_IO_DEFAULT));

  check(flux_std_star_frame=xsh_frame_product(tbl_ref_std_fname,"FLUX_STD_STAR",
                                        CPL_FRAME_TYPE_TABLE,
                                        CPL_FRAME_GROUP_PRODUCT,
                                              CPL_FRAME_LEVEL_INTERMEDIATE));


  filename=cpl_frame_get_filename(spectrum_frame);
  plist=cpl_propertylist_load(filename,0);
  naxis1=xsh_pfits_get_naxis1(plist);
  cdelt1=xsh_pfits_get_cdelt1(plist);
  wmin=xsh_pfits_get_crval1(plist);

  if( xsh_instrument_get_arm(instrument) == XSH_ARM_NIR){
     /* we assume gain in units of ADU/e- as ESO standard */
     gain=1./2.12; 
  } else {
     check_msg( gain = xsh_pfits_get_gain(plist),
                "Could not read gain factor");
  }



  xsh_free_propertylist(&plist);
  wmax=wmin+cdelt1*naxis1;


  check(ipol_flux_std_star_frame=xsh_spectrum_interpolate_linear(flux_std_star_frame,
                                                          1,
                                                          wmin,wmax));

  /*

  check(smooth_flux_std_star_frame=xsh_spectrum_resample(flux_std_star_frame,
                                                          0.08,
                                                       wmin,wmax,instrument));


*/
  check(resampl_flux_std_star_frame=xsh_spectrum_resample(ipol_flux_std_star_frame,
                                                          INTERPOL_WSTEP_NM,
                                                       wmin,wmax,instrument));


  /* Load flux_std_star table */
  check( star_list = xsh_star_flux_list_load( resampl_flux_std_star_frame ) ) ;

 

  /* Load rect_frame */
  check( spectrum = xsh_spectrum_load( spectrum_frame, instrument ) ) ;
  pcatg=cpl_sprintf("SPECTRUM_%s",xsh_instrument_arm_tostring( instrument));
  name=cpl_sprintf("%s.fits",pcatg);
  tmp_spectrum_frame=xsh_spectrum_save( spectrum,name,pcatg);
  cpl_free(pcatg);
  cpl_free(name);

  /* Load ATMOS EXT (if not NULL ) */
  if ( atmos_ext_frame != NULL ) {
    check( atmos_ext_list = xsh_atmos_ext_list_load( atmos_ext_frame ) ) ;
    xsh_msg( "ATMOS EXT Loaded" ) ;
    //check(xsh_atmos_ext_dump_ascii(atmos_ext_list,"pippo_atm_ext_list.asc"));
    check(filename=cpl_frame_get_filename(atmos_ext_frame));
    check(atmos_ext_tab=cpl_table_load(filename,1,0));
  }


  /* Compute */
  check(phigh=xsh_fill_high_abs_regions(instrument,high_abs_frame));
  
  check( resp_list = do_compute( star_list, spectrum, atmos_ext_tab, 
                                 phigh,airmass,exptime, gain, instrument ) ) ;

  //check(xsh_star_flux_list_dump_ascii(resp_list,"pippo_resp_list.asc"));

  /* Save the response frame */
  sprintf(tag,"RESPONSE_MERGE1D_%s_%s",xsh_instrument_mode_tostring(instrument),
                          xsh_instrument_arm_tostring(instrument));
  sprintf(fname,"%s.fits",tag);
  check( result = xsh_star_flux_list_save( resp_list, fname, tag ) ) ;


 cleanup:

  if(high_abs_frame!=NULL) {
     cpl_free(phigh);
  }
  xsh_spectrum_free( &spectrum ) ;

  xsh_star_flux_list_free( &star_list ) ;
  xsh_star_flux_list_free( &resp_list ) ;
  xsh_free_table(&tbl_ref_std_spectrum);
  xsh_free_frame(&flux_std_star_frame);
  xsh_free_frame(&ipol_flux_std_star_frame);
  xsh_free_frame(&resampl_flux_std_star_frame);
  xsh_free_frame(&tmp_spectrum_frame);

  xsh_atmos_ext_list_free(&atmos_ext_list) ;
  xsh_free_propertylist(&plist);
  xsh_free_table(&atmos_ext_tab);

  return result ;
}

/** 
 * Compute the instrument response using the rectified frame and the
 * Standard star flux frame. The input should include also the
 * Master Response Function and the Athmospheric Extinction Curve. Also
 * should be produced the Efficiency frame.
 * 
 * @param spectrum_frame Rectified frame (1D spectrum)
 * @param flux_std_star_cat_frame Standard star flux catalog frame
 * @param atmos_ext_frame Atmospheric Extinction Frame
 * @param instrument Instrument strcture
 * @param exptime Exposure time
 * 
 * @return Frame of the computed response
 */

cpl_frame * xsh_compute_response_ord( cpl_frame * spectrum_frame,
				  cpl_frame * flux_std_star_cat_frame,
				  cpl_frame * atmos_ext_frame,
				  cpl_frame * high_abs_win_frame,
				  xsh_instrument * instrument,
				  double exptime )
{
  cpl_frame * result = NULL ;	/**< computed response frame */
  xsh_star_flux_list * star_list = NULL ; /**< Std star flux */
  xsh_spectrum * spectrum = NULL ; /**< 1d spectrum */
  cpl_frame * tmp_spectrum_frame = NULL ; /**< 1d spectrum */
  xsh_star_flux_list * resp_list = NULL ;
  xsh_atmos_ext_list * atmos_ext_list = NULL ;

  char tag[80];
  char fname[80];
  cpl_frame* flux_std_star_frame=NULL;
  double dRA=0;
  double dDEC=0;
  cpl_table* tbl_ref_std_spectrum=NULL;
  const char* tbl_ref_std_fname="ref_flux_std_star.fits";
  cpl_frame* ipol_flux_std_star_frame=NULL;
  double wmin=0;
  double wmax=0;
  int naxis1=0;
  double cdelt1=0;
  const char* filename=NULL;
  cpl_propertylist* plist=NULL;
  char * name=NULL;
  char * pcatg=NULL;
  cpl_table* atmos_ext_tab=NULL;
  double airmass=0;
  double gain=1.;
 
  int next=0;
  int nord=0;
  int ext=0;
  HIGH_ABS_REGION * phigh=NULL;
   XSH_ARM the_arm;


  XSH_ASSURE_NOT_NULL( spectrum_frame ) ;
  XSH_ASSURE_NOT_NULL( flux_std_star_cat_frame ) ;
  XSH_ASSURE_NOT_NULL( instrument ) ;
  XSH_ASSURE_NOT_NULL( fname ) ;

  xsh_msg("Compute instrument response");

  /* get RA DEC from observed object frame */
  xsh_frame_sci_get_ra_dec_airmass(spectrum_frame,&dRA,&dDEC,&airmass);

  /* extract matching ref STD star spectrum */
  if(CPL_ERROR_NONE != xsh_parse_catalog_std_stars(flux_std_star_cat_frame,
                                                    dRA,dDEC,
                                                    STAR_MATCH_DEPSILON,
                                                    &tbl_ref_std_spectrum)){
     cpl_error_reset();
     xsh_free_table(&tbl_ref_std_spectrum);
     return NULL;
  }

  check(cpl_table_save(tbl_ref_std_spectrum,NULL,NULL,tbl_ref_std_fname,
                       CPL_IO_DEFAULT));

  check(flux_std_star_frame=xsh_frame_product(tbl_ref_std_fname,"FLUX_STD_STAR",
                                        CPL_FRAME_TYPE_TABLE,
                                        CPL_FRAME_GROUP_PRODUCT,
                                              CPL_FRAME_LEVEL_INTERMEDIATE));

  /* Load ATMOS EXT (if not NULL ) */
  if ( atmos_ext_frame != NULL ) {
     check( atmos_ext_list = xsh_atmos_ext_list_load( atmos_ext_frame ) ) ;
     xsh_msg( "ATMOS EXT Loaded" ) ;
     //check(xsh_atmos_ext_dump_ascii(atmos_ext_list,"pippo_atm_ext_list.asc"));
     check(filename=cpl_frame_get_filename(atmos_ext_frame));
     check(atmos_ext_tab=cpl_table_load(filename,1,0));
  }
  the_arm=xsh_instrument_get_arm(instrument);



  /* Load rect_frame */
  filename=cpl_frame_get_filename(spectrum_frame);
  next=cpl_frame_get_nextensions(spectrum_frame);
  nord=next/3;

  plist=cpl_propertylist_load(filename,0);
  if( xsh_instrument_get_arm(instrument) == XSH_ARM_NIR){
    /* we assume gain in units of ADU/e- as ESO standard */
    gain=1./2.12; 
  } else {
    check_msg( gain = xsh_pfits_get_gain(plist),
	       "Could not read gain factor");
  }
  xsh_free_propertylist(&plist);

  check(phigh=xsh_fill_high_abs_regions(instrument,high_abs_win_frame));

  for(ext=0;ext<next;ext+=3) {

    xsh_free_propertylist(&plist);
    plist=cpl_propertylist_load(filename,ext);
    naxis1=xsh_pfits_get_naxis1(plist);
    cdelt1=xsh_pfits_get_cdelt1(plist);
    wmin=xsh_pfits_get_crval1(plist);
    wmax=wmin+cdelt1*naxis1;

    xsh_free_frame(&ipol_flux_std_star_frame);
    check(ipol_flux_std_star_frame=xsh_spectrum_interpolate(flux_std_star_frame,
							    INTERPOL_WSTEP_NM,
                                                            wmin,wmax));
    /* Load flux_std_star table */
    xsh_star_flux_list_free( &star_list ) ;
    check( star_list = xsh_star_flux_list_load( ipol_flux_std_star_frame ) ) ;
 
     xsh_spectrum_free( &spectrum ) ;
     check( spectrum=xsh_spectrum_load_order(spectrum_frame,instrument,ext)) ;
     pcatg=cpl_sprintf("SPECTRUM_%s",xsh_instrument_arm_tostring( instrument));
     name=cpl_sprintf("%s.fits",pcatg);
     xsh_free_frame(&tmp_spectrum_frame);
     tmp_spectrum_frame=xsh_spectrum_save_order( spectrum,name,pcatg,ext);
     cpl_free(pcatg);
     cpl_free(name);

     /* Compute */
     xsh_star_flux_list_free( &resp_list ) ;

    
     check(resp_list=do_compute( star_list, spectrum, atmos_ext_tab, 
                                 phigh,airmass, exptime, gain, instrument ) ) ;
  
     //check(xsh_star_flux_list_dump_ascii(resp_list,"pippo_resp_list.asc"));

     /* Save the response frame */
     sprintf(tag,"RESPONSE_ORDER1D_%s_%s",xsh_instrument_mode_tostring(instrument),
             xsh_instrument_arm_tostring(instrument));
     sprintf(fname,"%s.fits",tag);
     xsh_free_frame(&result);
     check(result=xsh_star_flux_list_save_order(resp_list,fname,tag,ext)) ;
  }

 cleanup:

  if(high_abs_win_frame!=NULL) {
     cpl_free(phigh);
  }
  xsh_spectrum_free( &spectrum ) ;
  xsh_star_flux_list_free( &star_list ) ;
  xsh_star_flux_list_free( &resp_list ) ;
  xsh_free_table(&tbl_ref_std_spectrum);
  xsh_free_frame(&flux_std_star_frame);
  xsh_free_frame(&ipol_flux_std_star_frame);
  xsh_free_frame(&tmp_spectrum_frame);

  xsh_atmos_ext_list_free(&atmos_ext_list) ;
  xsh_free_propertylist(&plist);
  xsh_free_table(&atmos_ext_tab);

  return result ;
}

/**@}*/
