/*                                                                           *
 *   This file is part of the ESO X-shooter Pipeline                         *
 *   Copyright (C) 2006 European Southern Observatory                        *
 *                                                                           *
 *   This library is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA    *
 *                                                                           */

/*
 * $Author: amodigli $
 * $Date: 2011/02/17 15:40:53 $
 * $Revision: 1.13 $
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*---------------------------------------------------------------------------*/
/**
 * @defgroup data_handling    Data Format Handling functions
 */
/**
 * @defgroup xsh_data_star_flux  List of Standard Star Flux
 * @ingroup data_handling
 *
 */
/*---------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
  Includes
  ----------------------------------------------------------------------------*/

#include <xsh_dfs.h>
#include <xsh_error.h>
#include <xsh_msg.h>
#include <cpl.h>
#include <string.h>
#include <time.h>
#include <xsh_utils_table.h>
#include <xsh_data_star_flux.h>

/*----------------------------------------------------------------------------
  Function prototypes
  ----------------------------------------------------------------------------*/
#define XSH_STAR_FLUX_UVB_WAV_MIN 308
/*-----------------------------------------------------------------------------
                              Implementation
 -----------------------------------------------------------------------------*/
cpl_error_code 
xsh_star_flux_list_filter_median(xsh_star_flux_list * result, 
                                 int hsize )
{
   int size=0;
   cpl_vector* vflux=NULL;
   cpl_vector* vmedian=NULL;
   double* pvmedian=NULL;
   int i=0;

   vflux=cpl_vector_wrap(result->size,result->flux);
   vmedian=cpl_vector_filter_median_create(vflux,hsize);
   pvmedian=cpl_vector_get_data(vmedian);

   for(i=0;i<result->size;i++) {
      result->flux[i]=pvmedian[i];
   }

   cpl_vector_unwrap(vflux);
   xsh_free_vector(&vmedian);
   return cpl_error_get_code();
}

xsh_star_flux_list * xsh_star_flux_list_create( int size )
{
  xsh_star_flux_list * result = NULL ;

  /* Create the list */
  XSH_CALLOC( result, xsh_star_flux_list, 1 ) ;
  result->header = NULL ;
  result->size = size ;
  XSH_CALLOC( result->lambda, double, size ) ;
  XSH_CALLOC( result->flux, double, size ) ;

 cleanup:
  return result ;
}

xsh_star_flux_list * xsh_star_flux_list_load( cpl_frame * star_frame )
{
  cpl_table *table = NULL ;
  const char * tablename = NULL ;
  xsh_star_flux_list * result = NULL ;
  int nentries, i ;
  double * plambda, * pflux ;

  /* verify input */
  XSH_ASSURE_NOT_NULL( star_frame);

  /* get table filename */
  check(tablename = cpl_frame_get_filename( star_frame ));

  /* Load the table */
  XSH_TABLE_LOAD( table, tablename ) ;
   cpl_table_and_selected_double(table,XSH_STAR_FLUX_LIST_COLNAME_WAVELENGTH,CPL_LESS_THAN,XSH_STAR_FLUX_UVB_WAV_MIN);
   cpl_table_erase_selected(table);
   check( nentries = cpl_table_get_nrow( table ) ) ;

  /* Create the list */
  check( result = xsh_star_flux_list_create( nentries ) ) ;

  plambda = result->lambda ;
  pflux = result->flux ;

  check(result->header = cpl_propertylist_load(tablename, 0));

  check(cpl_table_cast_column(table,XSH_STAR_FLUX_LIST_COLNAME_WAVELENGTH,"F_WAVELENGTH",CPL_TYPE_FLOAT));
  check(cpl_table_cast_column(table,XSH_STAR_FLUX_LIST_COLNAME_FLUX,"F_FLUX",CPL_TYPE_FLOAT))
;

  /* Fill the list */
  for( i = 0 ; i< nentries ; i++, plambda++, pflux++ ) {
    float value ;

    check( xsh_get_table_value( table, "F_WAVELENGTH",
				CPL_TYPE_FLOAT, i, &value ) ) ;
    *plambda = value ;
    check( xsh_get_table_value( table, "F_FLUX",
				CPL_TYPE_FLOAT, i, &value ) ) ;
    *pflux = value ;
  }

 cleanup:
  if (cpl_error_get_code () != CPL_ERROR_NONE) {
    xsh_error_msg("can't load frame %s",cpl_frame_get_filename(star_frame));
    xsh_star_flux_list_free(&result);
  }
  XSH_TABLE_FREE( table);
  return result ;
}

void xsh_star_flux_list_free( xsh_star_flux_list ** list )
{
  if ( list != NULL && *list != NULL ) {
     xsh_free_propertylist(&(*list)->header);
    check( cpl_free( (*list)->lambda ) ) ;
    check( cpl_free( (*list)->flux ) ) ;
    //check( cpl_free( (*list)->header ) );
    check( cpl_free( *list ) ) ;
     *list = NULL ;
  }

 cleanup:
  return ;
}

cpl_error_code 
xsh_star_flux_list_dump_ascii( xsh_star_flux_list * list, const char* filename )
{
   FILE *   fout ;
   int size=0;
   int i=0;
   double * plambda=NULL;
   double * pflux=NULL;

   XSH_ASSURE_NOT_NULL_MSG(list,"Null input std star flux list!Exit");
   size=list->size;

   plambda = list->lambda ;
   pflux = list->flux ;

   if((fout=fopen(filename,"w"))==NULL) {

      return CPL_ERROR_FILE_IO ;
   } else {
      for(i=0;i<size;i++) {
	fprintf(fout, "%f %f \n", plambda[i], pflux[i]);
      }
   }
   if ( fout ) fclose( fout ) ;

 cleanup:
   return cpl_error_get_code() ;
}



cpl_frame * xsh_star_flux_list_save( xsh_star_flux_list * list,
				     const char * filename, const char * tag )
{
  cpl_table * table = NULL ;
  cpl_frame * result = NULL ;
  int size, i ;
  double * plambda, * pflux ;

  XSH_ASSURE_NOT_NULL(list);
  XSH_ASSURE_NOT_NULL(filename);
  
  /* create a table */
  check(table = cpl_table_new( 2 ));

  /* create column names */
  check(
    cpl_table_new_column(table, XSH_STAR_FLUX_LIST_COLNAME_WAVELENGTH,
      CPL_TYPE_FLOAT));
  check(
    cpl_table_new_column(table, XSH_STAR_FLUX_LIST_COLNAME_FLUX,
      CPL_TYPE_FLOAT));

  size = list->size ;
  plambda = list->lambda ;
  pflux = list->flux ;

  check(cpl_table_set_size(table, size));

  /* insert data */
  for( i = 0 ; i<size ; i++, plambda++, pflux++ ) {
    float value ;

    value = *plambda ;
    check(cpl_table_set_float(table, XSH_STAR_FLUX_LIST_COLNAME_WAVELENGTH,
			       i, value ));
    value = *pflux ;
    check(cpl_table_set_float(table, XSH_STAR_FLUX_LIST_COLNAME_FLUX,
			       i, value ));
  }

  /* create fits file */
  check(cpl_table_save( table, list->header, NULL, filename, CPL_IO_DEFAULT));
  check( xsh_add_temporary_file( filename ));

  /* Create the frame */
  check(result=xsh_frame_product( filename, tag,
				  CPL_FRAME_TYPE_TABLE,
				  CPL_FRAME_GROUP_PRODUCT,
				  CPL_FRAME_LEVEL_TEMPORARY));

  xsh_msg_dbg_low( "Star Flux Frame Saved" ) ;

 cleanup:
  XSH_TABLE_FREE( table);
  return result ;
}




cpl_frame * 
xsh_star_flux_list_save_order( xsh_star_flux_list * list,
                               const char * filename, 
                               const char * tag,
                               const int order )
{
  cpl_table * table = NULL ;
  cpl_frame * result = NULL ;
  int size, i ;
  double * plambda, * pflux ;

  XSH_ASSURE_NOT_NULL(list);
  XSH_ASSURE_NOT_NULL(filename);
  
  /* create a table */
  check(table = cpl_table_new( 2 ));

  /* create column names */
  check(
    cpl_table_new_column(table, XSH_STAR_FLUX_LIST_COLNAME_WAVELENGTH,
      CPL_TYPE_FLOAT));
  check(
    cpl_table_new_column(table, XSH_STAR_FLUX_LIST_COLNAME_FLUX,
      CPL_TYPE_FLOAT));

  size = list->size ;
  plambda = list->lambda ;
  pflux = list->flux ;

  check(cpl_table_set_size(table, size));

  /* insert data */
  for( i = 0 ; i<size ; i++, plambda++, pflux++ ) {
    float value ;

    value = *plambda ;
    check(cpl_table_set_float(table, XSH_STAR_FLUX_LIST_COLNAME_WAVELENGTH,
			       i, value ));
    value = *pflux ;
    check(cpl_table_set_float(table, XSH_STAR_FLUX_LIST_COLNAME_FLUX,
			       i, value ));
  }

  /* create fits file */
  if(order==0) {
     check(cpl_table_save( table,list->header,NULL,filename,CPL_IO_DEFAULT));
  } else {
     check(cpl_table_save( table,list->header,NULL,filename,CPL_IO_EXTEND));
  }
  check( xsh_add_temporary_file( filename ));

  /* Create the frame */
  check(result=xsh_frame_product( filename, tag,
				  CPL_FRAME_TYPE_TABLE,
				  CPL_FRAME_GROUP_PRODUCT,
				  CPL_FRAME_LEVEL_TEMPORARY));

  xsh_msg_dbg_low( "Star Flux Frame Saved" ) ;

 cleanup:
  XSH_TABLE_FREE( table);
  return result ;
}




double * xsh_star_flux_list_get_lambda( xsh_star_flux_list * list )
{
  XSH_ASSURE_NOT_NULL( list ) ;

 cleanup:
  return list->lambda ;
}

double * xsh_star_flux_list_get_flux( xsh_star_flux_list * list )
{
  XSH_ASSURE_NOT_NULL( list ) ;

 cleanup:
  return list->flux ;
}


/**@}*/
