/*                                                                            *
 *   This file is part of the ESO X-shooter Pipeline                          *
 *   Copyright (C) 2006 European Southern Observatory                         *
 *                                                                            *
 *   This library is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as published by     *
 *   the Free Software Foundation; either version 2 of the License, or        *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU General Public License for more details.                             *
 *                                                                            *
 *   You should have received a copy of the GNU General Public License        *
 *   along with this program; if not, write to the Free Software              *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA     *
 *                                                                            */

/*
 * $Author: amodigli $
 * $Date: 2011/01/17 08:02:37 $
 * $Revision: 1.21 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------*/
/**
 * @defgroup xsh_flat_merge Some flat utilities operations
 * @ingroup drl_functions
 *
 * Function ...
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
  Includes
  -----------------------------------------------------------------------------*/

#include <math.h>
#include <xsh_drl.h>

#include <xsh_badpixelmap.h>
#include <xsh_data_pre.h>
#include <xsh_dfs.h>
#include <xsh_pfits.h>
#include <xsh_error.h>
#include <xsh_msg.h>
#include <xsh_data_instrument.h>
#include <xsh_data_spectralformat.h>
#include <cpl.h>

/*-----------------------------------------------------------------------------
  Functions prototypes
  -----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
  Implementation
  -----------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/** 
  @brief Merge two master flat fields and order tables according the spectral 
    format
  @param[in] qth_frame  QTH master flat
  @param[in] qth_order_tab_frame QTH order edges tab
  @param[in] d2_frame  D2 master flat
  @param[in] d2_order_tab_frame D2 order edges tab
  @param[in] qth_bkg_frame QTH inter-order background frame
  @param[in] d2_bkg_frame  D2 inter-order background frame
  @param[out] qth_d2_flat_frame Merged flat frame
  @param[out] qth_d2_bkg_frame  merged inter-order subtracted background
  @param[out] qth_d2_order_tab_frame merged inter-order table
  @param[in] instrument instrument arm and lamp settings

*/
/*---------------------------------------------------------------------------*/
void 
xsh_flat_merge_qth_d2( cpl_frame *qth_frame, cpl_frame *qth_order_tab_frame,  
                       cpl_frame *d2_frame,cpl_frame *d2_order_tab_frame,
                       cpl_frame *qth_bkg_frame, cpl_frame *d2_bkg_frame,
		       cpl_frame **qth_d2_flat_frame,
		       cpl_frame **qth_d2_bkg_frame,
		       cpl_frame **qth_d2_order_tab_frame,
		       xsh_instrument *instrument)

{
  xsh_pre *qth_pre = NULL;
  xsh_pre *d2_pre = NULL;
  
  float *qth_data = NULL;
  float *d2_data = NULL;
  float *qth_errs = NULL;
  float *d2_errs = NULL;
  int *qth_qual = NULL;
  int *d2_qual = NULL;

  float *d2_bkg_data = NULL;

  float *qth_bkg_data = NULL;


  xsh_order_list *qth_list = NULL;
  xsh_order_list *d2_list = NULL;
  xsh_order_list *qth_d2_list = NULL;
  int y;
  cpl_polynomial *d2_limit = NULL;
  cpl_polynomial *qth_limit = NULL;
  cpl_image* d2_bkg_ima=NULL;
  cpl_image* qth_bkg_ima=NULL;

  const char* tag=NULL;
  const char* fname=NULL;
  char* name=NULL;
  char file_name[80];
  char file_tag[25];

  double qth_flux_min=0;
  double qth_flux_max=0;
  double d2_flux_min=0;
  double d2_flux_max=0;
  double flux_min=0;
  double flux_max=0;

 
  XSH_ASSURE_NOT_NULL( qth_frame);
  XSH_ASSURE_NOT_NULL( qth_order_tab_frame);
  XSH_ASSURE_NOT_NULL( d2_frame);
  XSH_ASSURE_NOT_NULL( d2_order_tab_frame);
  XSH_ASSURE_NOT_NULL( qth_d2_flat_frame);
  XSH_ASSURE_NOT_NULL( qth_bkg_frame);
  XSH_ASSURE_NOT_NULL( d2_bkg_frame);
  XSH_ASSURE_NOT_NULL( qth_d2_bkg_frame);
  XSH_ASSURE_NOT_NULL( qth_d2_order_tab_frame);
  XSH_ASSURE_NOT_NULL( instrument);

  xsh_msg_dbg_medium( "Entering xsh_flat_merge_qth_d2") ;

  /* load data */
  check( qth_pre = xsh_pre_load( qth_frame, instrument));
  check( d2_pre = xsh_pre_load( d2_frame, instrument));
  

  check( xsh_instrument_update_lamp( instrument, XSH_LAMP_QTH)); 
  check( qth_list = xsh_order_list_load( qth_order_tab_frame, instrument));
  xsh_order_list_set_bin_x( qth_list, qth_pre->binx);
  xsh_order_list_set_bin_y( qth_list, qth_pre->biny);
  check( xsh_instrument_update_lamp( instrument, XSH_LAMP_D2));
  check( d2_list = xsh_order_list_load( d2_order_tab_frame, instrument));
  xsh_order_list_set_bin_x( d2_list, d2_pre->binx);
  xsh_order_list_set_bin_y( d2_list, d2_pre->biny);

  check( qth_data = cpl_image_get_data_float( qth_pre->data));
  check( d2_data = cpl_image_get_data_float(d2_pre->data)); 
 
  fname=cpl_frame_get_filename(qth_bkg_frame);
  qth_bkg_ima=cpl_image_load(fname,CPL_TYPE_FLOAT,0,0);
  check( qth_bkg_data = cpl_image_get_data_float(qth_bkg_ima));

  fname=cpl_frame_get_filename(d2_bkg_frame);
  d2_bkg_ima=cpl_image_load(fname,CPL_TYPE_FLOAT,0,0);
  check( d2_bkg_data = cpl_image_get_data_float(d2_bkg_ima));


  check( qth_errs = cpl_image_get_data_float( qth_pre->errs));
  check( d2_errs = cpl_image_get_data_float(d2_pre->errs));
  
  check( qth_qual = cpl_image_get_data_int( qth_pre->qual));
  check( d2_qual = cpl_image_get_data_int(d2_pre->qual));
  


  /* find limit and merge flat */
  qth_limit = qth_list->list[qth_list->size-1].edglopoly;
  d2_limit = d2_list->list[0].edguppoly;

  for( y=0; y < d2_pre->ny; y++){
    int xd2=0, xqth=0;
    int x_avg=0;
    int x;
    
    check( xd2 = xsh_order_list_eval_int( d2_list, d2_limit, y+1)-1);
    check( xqth = xsh_order_list_eval_int( qth_list, qth_limit, y+1)-1);
    x_avg = floor( (xd2+xqth)/2.0);
    xsh_msg_dbg_medium("D2  x %d y %d", xd2, y);
    xsh_msg_dbg_medium("QTH x %d y %d", xqth, y);
    xsh_msg_dbg_medium(" x_avg = %d", x_avg);
    for( x=x_avg; x< d2_pre->nx; x++){
      d2_data[x+y*d2_pre->nx] = qth_data[x+y*d2_pre->nx];
      d2_errs[x+y*d2_pre->nx] = qth_errs[x+y*d2_pre->nx];
      d2_qual[x+y*d2_pre->nx] = qth_qual[x+y*d2_pre->nx];

      d2_bkg_data[x+y*d2_pre->nx] = qth_bkg_data[x+y*d2_pre->nx];
  

    }
  }

  /* construct the new order list */
  check( qth_d2_list = xsh_order_list_merge( qth_list, d2_list));

  check( xsh_instrument_update_lamp( instrument, XSH_LAMP_UNDEFINED));

  tag=XSH_GET_TAG_FROM_LAMP( XSH_MASTER_FLAT,instrument);
  XSH_NAME_LAMP_MODE_ARM( name, "MASTER_FLAT", ".fits", instrument);


  /* adjust QC FLUX MIN/MAX for master frame */
  check(qth_flux_min=cpl_propertylist_get_double(qth_pre->data_header,
                                                 XSH_QC_FLUX_MIN));
  check(qth_flux_max=cpl_propertylist_get_double(qth_pre->data_header,
                                                 XSH_QC_FLUX_MAX));
  check(d2_flux_min=cpl_propertylist_get_double(d2_pre->data_header,
                                                XSH_QC_FLUX_MIN));
  check(d2_flux_max=cpl_propertylist_get_double(d2_pre->data_header,
                                                XSH_QC_FLUX_MAX));

  flux_min=(qth_flux_min<d2_flux_min)? qth_flux_min:d2_flux_min;
  flux_max=(qth_flux_max>d2_flux_max)? qth_flux_max:d2_flux_max;

  check(cpl_propertylist_set_double(d2_pre->data_header,XSH_QC_FLUX_MIN,
                                    flux_min));
  check(cpl_propertylist_set_double(d2_pre->data_header,XSH_QC_FLUX_MAX,
                                    flux_max));



  check( *qth_d2_flat_frame = xsh_pre_save( d2_pre, name, tag,1));
  check( cpl_frame_set_tag( *qth_d2_flat_frame, tag));
  XSH_REGDEBUG("save %s %s", fname, tag);

  sprintf(file_tag,"MFLAT_BACK_%s_%s",
          xsh_instrument_mode_tostring(instrument),
          xsh_instrument_arm_tostring(instrument));

  sprintf(file_name,"%s.fits",file_tag);

  check(xsh_pfits_set_pcatg(d2_pre->data_header,file_tag));

  check(cpl_image_save(d2_bkg_ima,file_name,CPL_BPP_IEEE_FLOAT,
		       d2_pre->data_header,CPL_IO_DEFAULT));


  check(*qth_d2_bkg_frame=xsh_frame_product(file_name,file_tag, 
					    CPL_FRAME_TYPE_IMAGE,
					    CPL_FRAME_GROUP_CALIB,
					    CPL_FRAME_LEVEL_FINAL));


  tag= XSH_GET_TAG_FROM_LAMP( XSH_ORDER_TAB_EDGES, instrument);
  XSH_NAME_LAMP_MODE_ARM( name, "ORDER_TAB_EDGES", ".fits", instrument); 

  check( *qth_d2_order_tab_frame = xsh_order_list_save( qth_d2_list,instrument,
							name, tag,
							qth_pre->ny*d2_list->bin_y));


  XSH_REGDEBUG("save %s %s",name, tag);
 

  cleanup:
    XSH_FREE( name);
    xsh_free_image(&d2_bkg_ima);
    xsh_free_image(&qth_bkg_ima);
 
    xsh_pre_free( &qth_pre);
    xsh_pre_free( &d2_pre);
    xsh_order_list_free( &qth_list);
    xsh_order_list_free( &d2_list);
    xsh_order_list_free( &qth_d2_list);
    return;

}

/*--------------------------------------------------------------------------*/

/**@}*/
