/*                                                                            *
 *   This file is part of the ESO X-shooter Pipeline                          *
 *   Copyright (C) 2006 European Southern Observatory                         *
 *                                                                            *
 *   This library is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as published by     *
 *   the Free Software Foundation; either version 2 of the License, or        *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU General Public License for more details.                             *
 *                                                                            *
 *   You should have received a copy of the GNU General Public License        *
 *   along with this program; if not, write to the Free Software              *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA     *
 *                                                                            */

/*
 * $Author: amodigli $
 * $Date: 2010/10/03 14:40:01 $
 * $Revision: 1.16 $
*/

/**
 * @defgroup xsh_extract  Extract objects (xsh_extract)
 * @ingroup drl_functions
 *
 * Function ...
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
  Includes
  -----------------------------------------------------------------------------*/

#include <math.h>
#include <xsh_drl.h>

#include <xsh_utils_table.h>
#include <xsh_badpixelmap.h>
#include <xsh_data_pre.h>
#include <xsh_dfs.h>
#include <xsh_pfits.h>
#include <xsh_error.h>
#include <xsh_msg.h>
#include <xsh_fit.h>
#include <xsh_data_instrument.h>
#include <xsh_data_localization.h>
#include <xsh_data_rec.h>
#include <xsh_ifu_defs.h>
#include <xsh_data_image_3d.h>
#include <xsh_data_pre_3d.h>

#include <cpl.h>

/*-----------------------------------------------------------------------------
  Functions prototypes
  -----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
  Implementation
  -----------------------------------------------------------------------------*/

static void fill_img( cpl_image * img, int ilambda, cpl_image * down,
		cpl_image * cen, cpl_image * up, int up_shift, int cen_shift)
{
  int nx, ny, j ;
  

  check(nx = cpl_image_get_size_x( down ));
  check(ny = cpl_image_get_size_y( down));
  xsh_msg_dbg_high( "fill image at %d (%d x %d )", ilambda, nx, ny ) ;

  for( j = 1 ; j <= ny ; j++ ) {
    int rej ;
    double flux ;

    check(flux = cpl_image_get( down, ilambda+1, j, &rej )) ;
    check(cpl_image_set( img, 1, j+cen_shift, flux )) ;
  }

  check(nx = cpl_image_get_size_x( cen));
  check(ny = cpl_image_get_size_y( cen));

  for( j = 1 ; j <= ny ; j++ ) {
    int rej ;
    double flux ;

    check(flux = cpl_image_get( cen, ilambda+1, j, &rej )) ;
    check(cpl_image_set( img, 2, j+cen_shift, flux ) );
  }

  check(nx = cpl_image_get_size_x( up ));
  check(ny = cpl_image_get_size_y( up));
  for( j = 1 ; j <= ny ; j++ ) {
    int rej ;
    double flux ;

    check(flux = cpl_image_get( up, ilambda+1, j, &rej )) ;
    check(cpl_image_set( img, 3, j+up_shift, flux ) );
  }

 cleanup:
  return ;
}

static void make_data_cube( xsh_pre_3d * pre_3d, xsh_pre * pre_down,
			    xsh_pre * pre_cen, xsh_pre * pre_up )
{
  int nx, ny, nz, i ;
  cpl_image * cur_img = NULL ;
  int up_shift=0;
  int cen_shift =0;
  int slit_up, slit_cen;

  nx = xsh_pre_3d_get_nx( pre_3d ) ;
  ny = xsh_pre_3d_get_ny( pre_3d ) ;
  nz = xsh_pre_3d_get_nz( pre_3d ) ;
  xsh_msg( "make_data_cube: %d x %d x %d", nx, ny, nz ) ;
  
  slit_up = xsh_pre_get_ny( pre_up);
  slit_cen = xsh_pre_get_ny( pre_cen);
  if ( slit_up > slit_cen){
    cen_shift = slit_up-slit_cen;
  }
  else{
    up_shift=slit_cen-slit_up;
  }
 
  xsh_msg( "make_data_cube: %d x %d x %d slit_shift %dx%dx%d", nx, ny, nz,
    cen_shift,cen_shift,up_shift);

  for( i = 0 ; i<nz ; i++ ) {
    /* Data */
    /* fill pseudo image and insert into data cube */
    check( cur_img = cpl_image_new( nx, ny, CPL_TYPE_FLOAT ) ) ;
    check(fill_img( cur_img, i, pre_down->data, pre_cen->data, pre_up->data,
      up_shift,cen_shift)) ;
    check( xsh_image_3d_insert( pre_3d->data, cur_img, i ) ) ;
    xsh_free_image( &cur_img ) ;
    /* Errs */
    check( cur_img = cpl_image_new( nx, ny, CPL_TYPE_FLOAT ) ) ;
    check(fill_img( cur_img, i, pre_down->errs, pre_cen->errs, pre_up->errs,
      up_shift,cen_shift)) ;
    check( xsh_image_3d_insert( pre_3d->errs, cur_img, i ) ) ;
    xsh_free_image( &cur_img ) ;
    /* Qual */
    check( cur_img = cpl_image_new( nx, ny, CPL_TYPE_INT ) ) ;
    check(fill_img( cur_img, i, pre_down->qual, pre_cen->qual, pre_up->qual,
      up_shift,cen_shift)) ;
    check( xsh_image_3d_insert( pre_3d->qual, cur_img, i ) ) ;
    xsh_free_image( &cur_img ) ;
  }

 cleanup:
  xsh_free_image( &cur_img ) ;
  return ;
}


cpl_frame* xsh_format( cpl_frameset *spectrum_frame_set, 
		       const char *result_name, xsh_instrument * instrument,
		       const char* rec_prefix)
{
  cpl_frame * result = NULL ;
  xsh_image_3d * flux_3d = NULL, * errs_3d = NULL, * qual_3d = NULL ;
  xsh_pre * pre_down = NULL, * pre_cen = NULL, * pre_up = NULL ;
  cpl_frame * frame_down = NULL, * frame_cen = NULL, * frame_up = NULL ;
  int nslitlet = 3, nslit, nlambda ;
  int nslit_up, nslit_lo, nslit_cen;
  char tag[80];
  char pcatg[80];
  xsh_pre_3d * pre_3d = NULL ;
  double lambda_min, lambda_max, lambda_step ;
  double crval2 = 0.0, cdelt2=0.0;

  XSH_ASSURE_NOT_NULL( spectrum_frame_set ) ;
  XSH_ASSURE_NOT_NULL( result_name ) ;
  XSH_ASSURE_NOT_NULL( instrument ) ;

  check( frame_down = cpl_frameset_get_frame( spectrum_frame_set, 0));
  check( pre_down = xsh_pre_load( frame_down, instrument));
  check( frame_cen = cpl_frameset_get_frame( spectrum_frame_set, 1));
  check( pre_cen = xsh_pre_load( frame_cen, instrument));
  check( frame_up = cpl_frameset_get_frame( spectrum_frame_set, 2));
  check( pre_up = xsh_pre_load( frame_up, instrument));

  /* Retrieve lambda_min, max and step for calculation of
     KeyWords CDELT3, CRVAL3, etc*/
  check(lambda_min=xsh_pfits_get_rectify_lambda_min( pre_down->data_header));
  check(lambda_max=xsh_pfits_get_rectify_lambda_max( pre_down->data_header));
  check(lambda_step=xsh_pfits_get_rectify_bin_lambda( pre_down->data_header));
  xsh_msg( "Rect Lambda min, max, bin: %lf, %lf, %lf", lambda_min, lambda_max,
    lambda_step) ;

  nslit_lo = xsh_pre_get_ny( pre_down);
  nslit_up = xsh_pre_get_ny( pre_up);
  nslit_cen = xsh_pre_get_ny( pre_cen);

  nslit = nslit_cen;
  check( crval2 = xsh_pfits_get_crval2(pre_cen->data_header));
  check( cdelt2 = xsh_pfits_get_cdelt2(pre_cen->data_header));

  if (nslit_up > nslit_cen){
    nslit += nslit_up-nslit_cen;
    crval2 -= (nslit_up-nslit_cen)*cdelt2;
  }
  if (nslit_lo > nslit_cen){
    nslit += nslit_lo-nslit_cen;
  }

  nlambda = xsh_pre_get_nx( pre_down);

  xsh_msg("nslit : lo %d cen %d up %d",nslit_lo,nslit_cen,nslit_up);

  check( pre_3d = xsh_pre_3d_new( nslitlet, nslit, nlambda));
  xsh_msg( "Pre_3d created: %dx%dx%d", nslitlet, nslit, nlambda);

  check( flux_3d = xsh_pre_3d_get_data( pre_3d));
  check( errs_3d = xsh_pre_3d_get_errs( pre_3d));
  check( qual_3d = xsh_pre_3d_get_qual( pre_3d));

  /* Rotate up and down by 180 degress */
  check( xsh_pre_flip( pre_down, 0));
  check( xsh_pre_flip( pre_up, 0));

  make_data_cube( pre_3d, pre_down, pre_cen, pre_up);

  /* Get headers and set PRO.CATG */
  check( cpl_propertylist_append( pre_3d->data_header,
    pre_cen->data_header));
  check( cpl_propertylist_append( pre_3d->errs_header,
    pre_cen->errs_header));
  check( cpl_propertylist_append( pre_3d->qual_header,
    pre_cen->qual_header));

  sprintf(pcatg,"%s_%s",rec_prefix,
	  XSH_GET_TAG_FROM_ARM( XSH_MERGE3D_IFU, instrument));

  xsh_msg( "   Setting PRO.CATG '%s'", pcatg ) ;
  check( xsh_pfits_set_pcatg( pre_3d->data_header, pcatg));
  check( xsh_pfits_set_pcatg( pre_3d->errs_header, pcatg));
  check( xsh_pfits_set_pcatg( pre_3d->qual_header, pcatg));




  check( xsh_pfits_set_crpix1( pre_3d->data_header, 1.0));
  check( xsh_pfits_set_crpix1( pre_3d->errs_header, 1.0));
  check( xsh_pfits_set_crpix1( pre_3d->qual_header, 1.0));

  check( xsh_pfits_set_crval1( pre_3d->data_header, -0.6));
  check( xsh_pfits_set_crval1( pre_3d->errs_header, -0.6));
  check( xsh_pfits_set_crval1( pre_3d->qual_header, -0.6));

  check( xsh_pfits_set_cdelt1( pre_3d->data_header, 0.6));
  check( xsh_pfits_set_cdelt1( pre_3d->errs_header, 0.6));
  check( xsh_pfits_set_cdelt1( pre_3d->qual_header, 0.6));


  check( xsh_pfits_set_crval2( pre_3d->data_header, crval2));
  check( xsh_pfits_set_crval2( pre_3d->errs_header, crval2));
  check( xsh_pfits_set_crval2( pre_3d->qual_header, crval2));

  check( xsh_pfits_set_crpix3( pre_3d->data_header, 0.5));
  check( xsh_pfits_set_crpix3( pre_3d->errs_header, 0.5));
  check( xsh_pfits_set_crpix3( pre_3d->qual_header, 0.5));

  check( xsh_pfits_set_crval3( pre_3d->data_header, lambda_min));
  check( xsh_pfits_set_crval3( pre_3d->errs_header, lambda_min));
  check( xsh_pfits_set_crval3( pre_3d->qual_header, lambda_min));


  check( xsh_pfits_set_cdelt3( pre_3d->data_header, lambda_step));
  check( xsh_pfits_set_cdelt3( pre_3d->errs_header, lambda_step));
  check( xsh_pfits_set_cdelt3( pre_3d->qual_header, lambda_step));

  check( xsh_pfits_set_ctype3( pre_3d->data_header, "LINEAR"));
  check( xsh_pfits_set_ctype3( pre_3d->errs_header, "LINEAR"));
  check( xsh_pfits_set_ctype3( pre_3d->qual_header, "LINEAR"));

/* the fdollowing should be done using more compact functions as: */
  check(xsh_set_cd_matrix3d(pre_3d->data_header));
  check(xsh_set_cd_matrix3d(pre_3d->errs_header));
  check(xsh_set_cd_matrix3d(pre_3d->qual_header));

  xsh_msg( "Saving %s", result_name) ;

  check( result = xsh_pre_3d_save( pre_3d, result_name, 0));

  sprintf(tag,"%s_%s",rec_prefix, 
	  XSH_GET_TAG_FROM_ARM( XSH_MERGE3D_IFU, instrument));

  xsh_msg( "   Setting TAG '%s'", tag);
  check( cpl_frame_set_tag( result, tag));
  check( cpl_frame_set_group( result, CPL_FRAME_GROUP_PRODUCT));
  check( cpl_frame_set_level( result, CPL_FRAME_LEVEL_FINAL));
  check(cpl_frame_set_type( result, CPL_FRAME_TYPE_IMAGE));

  cleanup:
    xsh_pre_free( &pre_down);
    xsh_pre_free( &pre_cen);
    xsh_pre_free( &pre_up);
    xsh_pre_3d_free( &pre_3d);
    return result;
}
