/*                                                                            *
 *   This file is part of the ESO X-shooter Pipeline                          *
 *   Copyright (C) 2006 European Southern Observatory                         *
 *                                                                            *
 *   This library is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as published by     *
 *   the Free Software Foundation; either version 2 of the License, or        *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU General Public License for more details.                             *
 *                                                                            *
 *   You should have received a copy of the GNU General Public License        *
 *   along with this program; if not, write to the Free Software              *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA     *
 *                                                                            */

/*
 * $Author: amodigli $

 * $Date: 2011/01/17 14:53:15 $
 * $Revision: 1.109 $
 * $Name: HEAD $
 *
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

/*----------------------------------------------------------------------------*/
/**
 * @defgroup xsh_remove_crh_multi  Remove Cosmic Rays Multiple (xsh_remove_crh_multiple)
 * @ingroup drl_functions
 *
 * Suppress Cosmic Rays by analysing multiple files.
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
  Includes
-----------------------------------------------------------------------------*/

#include <math.h>

#include <xsh_drl.h>
#include <xsh_data_pre.h>
#include <xsh_dfs.h>
#include <xsh_pfits.h>
#include <xsh_error.h>
#include <xsh_msg.h>
#include <cpl.h>
#include <xsh_badpixelmap.h>

/*-----------------------------------------------------------------------------
  Functions prototypes
 -----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
  Implementation
 -----------------------------------------------------------------------------*/




/** 
 * Create an error image from an image list and a "valid pixels" mask 
 *
 * @param errs [OUTPUT] The errs image result  
 * @param list error images list to collapse
 * @param mask mask given pixel to compute in error image
 * 
 * @return Pointer to error image
 */
static void xsh_collapse_errs(cpl_image * errs, cpl_imagelist * list, 
  cpl_mask* mask)
{
  int nx = 0, ny = 0, nimg = 0, i = 0;
  float **pdata = NULL;
  cpl_binary ** pbinary = NULL;
  cpl_binary * binary = NULL;  
  float* errsdata = NULL;

  check(nimg = cpl_imagelist_get_size (list));
  assure(nimg > 0,CPL_ERROR_ILLEGAL_INPUT,"you must have image to collapse");

  /* create the array of pointers to errs data */
  pdata = cpl_malloc (nimg * sizeof (float *));
  assure (pdata != NULL, cpl_error_get_code (),
	  "Cant allocate memory for data pointers");
  /* create the array of pointers to errs binary */
  pbinary = cpl_malloc (nimg * sizeof (cpl_binary *));
  assure (pbinary != NULL, cpl_error_get_code (),
          "Cant allocate memory for binary pointers");

  /* Populate images pointer array */
  for (i = 0; i < nimg; i++) {
    check( pdata[i] = cpl_image_get_data_float(cpl_imagelist_get (list, i)));
    check( pbinary[i] = cpl_mask_get_data(cpl_image_get_bpm(
      cpl_imagelist_get(list, i))));
  }

  /* get size and from first image */
  check(nx = cpl_image_get_size_x (cpl_imagelist_get (list, 0)));
  check(ny = cpl_image_get_size_y (cpl_imagelist_get (list, 0)));
  check(errsdata = cpl_image_get_data_float(errs));

  /* get binary from the mask */
  check(binary = cpl_mask_get_data(mask));

  /* Loop over all pixels */
  for (i = 0; i < nx * ny; i++){
    int count = 0;
    int k = 0;
    double errval = 0.0;
   
    for (count = 0, k = 0; k < nimg; k++) {
      /* check if good pixel */
      if ( (binary[i] == CPL_BINARY_0) && ((pbinary[k])[i] == CPL_BINARY_0)) {
	  errval += (pdata[k])[i] * (pdata[k])[i];
	  count++;
      }
    }
    /* now compute error */
    if (count > 1) {
      double eval;
      
      eval = sqrt( M_PI / 2.0 * ((double) count / ((double) count- 1.))) * 
        (1./(double)count) * sqrt (errval);
	errsdata[i] = eval;
    }
  }
  cleanup:
    cpl_free(pdata);
    cpl_free(pbinary);
}

static int xsh_remove_cr (cpl_imagelist * datalist, cpl_imagelist * errslist,
	       cpl_image * median,
	       cpl_image * errs, cpl_mask * mask,
	       xsh_clipping_param * crh_clipping)
{
  int nx, ny, nimg, i;
  cpl_image **pimg = NULL;
  cpl_image **perrs = NULL;
  int ix = 0,iy = 0;
  int crcount = 0;  
  float* medvals = NULL;
  float* mederrs = NULL;
  cpl_binary *bpmdata = NULL; 
 
 
 
  check(nimg = cpl_imagelist_get_size (datalist));
  
  /* create the array of pointers to images in list */
  pimg = cpl_malloc (nimg * sizeof (cpl_image *));
  assure (pimg != NULL, cpl_error_get_code (),
	  "Cant allocate memory for image pointers");
  perrs = cpl_malloc (nimg * sizeof (cpl_image *));
  assure (pimg != NULL, cpl_error_get_code (),
          "Cant allocate memory for image pointers");
 
  xsh_msg_dbg_low ("Nb of images: %d\n", nimg);

  /* Populate images pointer array */
  for (i = 0; i < nimg; i++){
    *(pimg + i) = cpl_imagelist_get (datalist, i);
    *(perrs + i) = cpl_imagelist_get (errslist, i);
  }
  
  /* get size from first image */
  check(nx = cpl_image_get_size_x (*pimg));
  check(ny = cpl_image_get_size_y (*pimg));

  /* Loop over all pixels */
  check(medvals = cpl_image_get_data_float(median));
  check(mederrs = cpl_image_get_data_float(errs));
  check(bpmdata = cpl_mask_get_data(mask));

  for (iy = 1; iy <= ny; iy++){
    for (ix = 1; ix <= nx; ix++) {
      int k;
      double medval;
      double error;
      int badpix = 0;
      /* crh_clipping->sigma is what is usually called 'kappa' */
      error = mederrs[ix-1+(iy-1)*nx] * crh_clipping->sigma;
      //xsh_msg("errs=%g kappa=%g",mederrs[ix-1+(iy-1)*nx],crh_clipping->sigma);
      medval = medvals[ix-1+(iy-1)*nx];
      badpix = 0;
      

      for (k = 0; k < nimg; k++) {
        /* if good pixel, check if still good */
	if (!bpmdata[ix-1+(iy-1)*nx]) {
	  double delta = 0.0, kvalue = 0.0;
          int rej = 0;
	  kvalue = cpl_image_get (*(pimg + k), ix, iy, &rej);
	  if (rej != 0) {
	    xsh_msg_dbg_high ("Pixel %d,%d already bad in image %d", ix, iy, k);
	    continue;
	  }
	  delta = fabs(kvalue - medval);
	  /*
          xsh_msg("error=%g kvalue=%g medval=%g delta=%g",
          error,kvalue,medval,delta);
	  */
	  if (delta > error) {
	    cpl_image_reject (*(pimg + k), ix, iy);
            cpl_image_reject (*(perrs + k), ix, iy);
	    	    if ( crcount < 10 )
		    xsh_msg ("CR Rejected for Pixel %d,%d [%d] (%lf vs %lf)",
		    ix, iy, k, medval, kvalue);
	    
	    crcount++;
            badpix = 1;
	  }
	}
	else {
	  xsh_msg_dbg_high("Pixel %d,%d already known as bad pixel in image %d", ix, iy,
            k+1);
	  break;
	}
      }
      /* mask good pixel */
      if (!badpix){
        /*xsh_msg("add to mask %d %d",ix,iy);*/
        bpmdata[ix-1+(iy-1)*nx] = CPL_BINARY_1;
      }
    }
  }

  cpl_free (pimg);
  cpl_free(perrs);

  cleanup:
    return crcount;
}

/** 
 * Calculates and set the QC parameters relevant for Cosmic Rays.
 * 
 * @param pre Pointer to XSH_PRE Structure
 * @param nbcrh Total umber of cosmics found
 * @param nframes Number of frames used to remove cosmics
 * @param instrument Pointer to indstrument description structure
 */
static void add_qc_crh (xsh_pre* pre, int nbcrh, int nframes,
			xsh_instrument * instrument )
{
  double qc_crrate = 0.0;
  int nbcr_avg ;		/**< Average of cosmics per frame */
  XSH_ASSURE_NOT_NULL(pre);
  XSH_ASSURE_NOT_ILLEGAL(pre->pszx >0. && pre->pszy > 0); 
  XSH_ASSURE_NOT_ILLEGAL(pre->exptime > 0); 
  /*
    QC.CRRATE = avg_nb_of_crh_per_frame/(exposure_time*size_of_detector_in_cm2)
    Nb of Cosmics per cm2 per second.
    The whole size of the image in cm2 is given by pszx and pszy. These
    values are in MICRONS.
  */
  xsh_msg_dbg_medium( "add_qc_crh - Exptime = %f", pre->exptime ) ;
  qc_crrate =
    (double) nbcrh / (((double) nframes) * (pre->exptime * 
    (pre->pszx / 10000.0) * (pre->pszy /10000.0) * pre->nx * pre->ny)) ;
  nbcr_avg = nbcrh/nframes ;

  /*
    Now set QC KW
  */
  check( xsh_pfits_set_qc_crrate( pre->data_header,qc_crrate) ) ;
  check( xsh_pfits_set_qc_ncrh( pre->data_header, nbcrh) ) ;
  check( xsh_pfits_set_qc_ncrh_mean( pre->data_header,nbcr_avg) ) ;
  /* Same thing for the qual header */
  check( xsh_pfits_set_qc_crrate( pre->qual_header,qc_crrate)) ;
  check( xsh_pfits_set_qc_ncrh( pre->qual_header,nbcrh)) ;
  check( xsh_pfits_set_qc_ncrh_mean( pre->qual_header,nbcr_avg) ) ;
  cleanup:
    return ;
}

/*---------------------------------------------------------------------------*/
/** 
 * @brief Creates median, error and bad pixel map images from a set of
 * frames.
 * @param rawFrames Pointer to the frameset (returned from xsh_prepare)
 * @param result_tag Tag of resulting file
 * @param crh_clipping Pointer to parameters for sigma clipping
 * @param instr instrument containing the arm , mode and lamp in use
 * @param list Pointer to created image list
 * @param crh_ima Pointer to created crh image map
 * @param median_mean
 * @return the median frame newly allocated where cosmic rays are removed
 */

static cpl_frame* 
xsh_remove_cosmics (cpl_frameset * rawFrames, 
		    const char *result_tag, 
		    xsh_clipping_param * crh_clipping, 
		    xsh_instrument* instr, 
		    cpl_imagelist ** list,
		    cpl_image** crh_ima,
		    int median_mean )
{
  /* MUST BE DEALLOCATED in caller */
  cpl_frame* medFrame = NULL;
  cpl_imagelist *dataList = NULL;

  /* ALLOCATED locally */
  cpl_imagelist *errsList = NULL;
  xsh_pre* pre = NULL;
  cpl_mask* badpix_mask = NULL;
  cpl_mask* crh_mask = NULL;
  /* Others */
  cpl_frame *current = NULL;
  int i=0, size = 0;
  long nbpixcosmic = -1;
  int nb_crh_total = 0 ;
  //cpl_image* crh_img=NULL;
  char name[80];
  char result_name[80];

  /* Check parameters */
  XSH_ASSURE_NOT_NULL( instr);
  XSH_ASSURE_NOT_NULL( rawFrames);
  XSH_ASSURE_NOT_NULL( result_tag);
  XSH_ASSURE_NOT_NULL( crh_clipping);
  check( size = cpl_frameset_get_size( rawFrames));

  XSH_ASSURE_NOT_ILLEGAL( size > 0);  

  /* Create the image lists */
  check( dataList = cpl_imagelist_new());
  check( errsList = cpl_imagelist_new());

  /* Populate image lists with images in frame set */
  for (i=0;i < size;i++){

    check(current = cpl_frameset_get_frame(rawFrames,i));
    /* Load pre file */
    check(pre = xsh_pre_load(current,instr));
    check_msg( cpl_imagelist_set(dataList,cpl_image_duplicate(pre->data), 
      i),"Cant add Data Image %d to imagelist", i) ;
    check_msg( cpl_imagelist_set(errsList,cpl_image_duplicate(pre->errs), 
      i),"Cant add Data Image %d to imagelist", i) ;

    /* All the input frames have the same bad pixel map, thus use the
       last one as the "final" one */
    if (i < (size-1)) {
      xsh_pre_free(&pre);
    }
  }

  /* Verify uniformity of imagelists */
  assure (cpl_imagelist_is_uniform (dataList) == 0, CPL_ERROR_ILLEGAL_INPUT,
	  "Data images are not uniform");
  assure (cpl_imagelist_is_uniform (errsList) == 0, CPL_ERROR_ILLEGAL_INPUT,
	  "Errs images are not uniform");

  /* get mask from QUAL */
  check(badpix_mask = cpl_mask_duplicate(xsh_pre_get_bpmap(pre)));
  //check(crh_mask = cpl_mask_duplicate(xsh_pre_get_bpmap(pre)));

  xsh_msg_dbg_medium ("Nb of Bad Pixels in QUAL: %d",cpl_mask_count(badpix_mask));
  
  /* in all cases */

  /* For two images always combine by mean */
  if ( size == 2){
    check( xsh_bpmap_collapse_mean( pre->data,dataList,badpix_mask));
    check( xsh_collapse_errs(pre->errs,errsList, badpix_mask));
  }

  /* loop to remove cosmics */  
  if ( size >= 3 ) {
   xsh_msg( "More than 2 input frames: remove crh (method multi)");
   check(xsh_bpmap_collapse_median(pre->data,dataList, badpix_mask));
  }

  if ( size >= 3 ) {
    long nbpixtotal = cpl_frameset_get_size(rawFrames) * pre->nx * pre->ny;
    double frac_accepted = 1;   
    int iter = 0;

    while(iter < crh_clipping->niter && frac_accepted >= crh_clipping->frac
      && (nbpixcosmic!=0) ) {

      
      /* try to remove cosmics */
      check(nbpixcosmic = xsh_remove_cr (dataList, errsList,pre->data,
        pre->errs, badpix_mask, crh_clipping));
      if(iter==0) crh_mask=cpl_mask_duplicate(badpix_mask);

      //AModigliani: here one have to use an and (or better an or) to coadd the
      //bad pixels found on each iteration. Instead the actual bad pixel
      //image saved on the QUAL part is the one coming from the last iteration
      //which usually does not find any bad pixel. This is wrong. I propose
      //to use the one which I save as crh_mask, even bettter if coming from
      //an or coaddition: cpl_mask_or(...)
      check(cpl_mask_and(crh_mask,badpix_mask));
      //check(crh_img=cpl_image_new_from_mask(badpix_mask));
      //sprintf(name,"%s_ITER_%d%s","crh_ima_",iter,".fits");
      //cpl_image_save (crh_img,name,CPL_BPP_32_SIGNED,NULL, CPL_IO_DEFAULT);
      //xsh_free_image(&crh_img);

      xsh_msg(" Iteration Nb %d/%d, new cosmic found %ld", 
	      iter+1, crh_clipping->niter, nbpixcosmic);

      /* computes fraction */
      frac_accepted -= ((double)nbpixcosmic / (double)nbpixtotal);
      xsh_msg_dbg_medium("accepted pixels / total pixels = %f)",frac_accepted);

      /* recreate median image */
      check(xsh_bpmap_collapse_median(pre->data,dataList,crh_mask));
 
      /* recreate errs image */
      check(xsh_collapse_errs(pre->errs,errsList, crh_mask));
      iter++;

      /* Total nb of cosmics */
      nb_crh_total += nbpixcosmic ;
    } 
  }
  //check( pre->data=xsh_imagelist_collapse_sigclip_iter_create(dataList,3,3,2));

  if ( size >= 3 ) {
    xsh_msg(" Nb of cosmics removed %d", nb_crh_total);
  }
  /* REGDEBUG : flag pixel only in debug to find cosmic */

  if (xsh_debug_level_get() >= XSH_DEBUG_LEVEL_MEDIUM) {
    cpl_mask* cosmic_mask = NULL;
    int imask, jmask;     

    cosmic_mask = cpl_mask_new(pre->nx, pre->ny);
    for(i=0;i< size; i++){
      cpl_image *img = cpl_imagelist_get( dataList, i);
      cpl_mask *imgmask = cpl_image_get_bpm( img);
      for(jmask=1; jmask<=pre->ny; jmask++){
        for(imask=1; imask <= pre->nx; imask++){
          if (cpl_mask_get(imgmask,imask, jmask) == CPL_BINARY_1){
            cpl_mask_set(cosmic_mask, imask, jmask,CPL_BINARY_1);
          }
        }
      }
    }
    //xsh_bpmap_mask_bad_pixel( pre->qual, cosmic_mask,
    //QFLAG_COSMIC_RAY_REMOVED);
	//AModigliani: I think here is the place one need to modify the BP map.
    //xsh_bpmap_mask_bad_pixel( pre->qual, cpl_mask_threshold_image_create(crh_ima,-0.5,0.5), 
    //  QFLAG_COSMIC_RAY_REMOVED);
    cpl_mask_delete( cosmic_mask);
  }

  /* Add PRO KW */
  check( xsh_pfits_set_datancom( pre->data_header, size));
  /* Add QC Keywords */
  if (pre->exptime > 0){
    check(add_qc_crh( pre, nb_crh_total, size, instr ));
  }

  /* remove irrelevant KW */
  check ( cpl_propertylist_erase_regexp( pre->data_header, "PRO BIAS *", 0));

  /* Now save the relevant frame of PRE
     With: median image, errs image and bad pixel map
  */
  sprintf(result_name,"%s.fits",result_tag);
  xsh_msg_dbg_high ("save frame %s tag=%s", result_name,result_tag); 
  check(medFrame = xsh_pre_save( pre, result_name, result_tag,1));
  check(cpl_frame_set_tag(medFrame, result_tag));
  
  //AMo: add crh map save for QC
    if((crh_mask!=NULL) && ((crh_ima) !=NULL)) {
    xsh_free_image(crh_ima);
    check(*crh_ima=cpl_image_new_from_mask(crh_mask));
    cpl_mask_not(crh_mask);
    xsh_bpmap_mask_bad_pixel(*crh_ima, crh_mask,QFLAG_COSMIC_RAY_REMOVED);
    //here good pixels are still flagged as '1'. To have proper code, '0',
    //we make a threshold
    cpl_image_threshold(*crh_ima,1.1,DBL_MAX,0,DBL_MAX);
    check(sprintf(name,"%s.fits",XSH_GET_TAG_FROM_ARM(XSH_CRH_MAP,instr)));
    cpl_image_save (*crh_ima,name,CPL_BPP_32_SIGNED,NULL, CPL_IO_DEFAULT);
    }

  /* At this point, the images in dataList contain the "cpl bad pixel map"
     with the locally found CRh ==> should be used in compute_noise
     How should we return this list to the main recipe ==> in list
  */

  cleanup:
    if (cpl_error_get_code () != CPL_ERROR_NONE) {
      xsh_free_imagelist(&dataList);
      xsh_free_frame(&medFrame);
      if (list != NULL){
        *list = NULL;
      }
    }
    if (list != NULL){
      *list = dataList;
    }
    else{
      xsh_free_imagelist(&dataList);
    }
    //xsh_free_image(crh_ima);
    xsh_free_mask(&crh_mask);
    xsh_free_mask(&badpix_mask);
    xsh_pre_free(&pre);
    xsh_free_imagelist(&errsList) ;
    return medFrame ;
}
/** 
 * @brief Creates master bias
 * @param rawFrames Pointer to the frameset 
 * @param instr instrument containing the arm , mode and lamp in use
 * @param result_tag product catg 
 * @return the clean mean frame newly allocated
 */

cpl_frame* 
xsh_create_master_bias2(cpl_frameset* rawFrames, 
                        xsh_instrument* instr,const char* result_tag)
{
  cpl_frame* medFrame = NULL;
  cpl_imagelist *dataList = NULL;
  cpl_imagelist *errsList = NULL;
  int i=0;
  int size=0;
  xsh_pre* pre = NULL;
  cpl_frame *current = NULL;
  cpl_mask* badpix_mask = NULL;
  char result_name[80];


  XSH_ASSURE_NOT_NULL( instr);
  XSH_ASSURE_NOT_NULL( rawFrames);
  check( size = cpl_frameset_get_size( rawFrames));

  XSH_ASSURE_NOT_ILLEGAL( size > 0);  

  check( dataList = cpl_imagelist_new());
  check( errsList = cpl_imagelist_new());

  /* Populate image lists with images in frame set */
  for (i=0;i < size;i++){

    check(current = cpl_frameset_get_frame(rawFrames,i));
    /* Load pre file */
    check(pre = xsh_pre_load(current,instr));
    check_msg( cpl_imagelist_set(dataList,cpl_image_duplicate(pre->data), 
      i),"Cant add Data Image %d to imagelist", i) ;
    check_msg( cpl_imagelist_set(errsList,cpl_image_duplicate(pre->errs), 
      i),"Cant add Data Image %d to imagelist", i) ;

    /* All the input frames have the same bad pixel map, thus use the
       last one as the "final" one */
    if (i < (size-1)) {
      xsh_pre_free(&pre);
    }
  }
  /* Verify uniformity of imagelists */
  assure (cpl_imagelist_is_uniform (dataList) == 0, CPL_ERROR_ILLEGAL_INPUT,
	  "Data images are not uniform");
  assure (cpl_imagelist_is_uniform (errsList) == 0, CPL_ERROR_ILLEGAL_INPUT,
	  "Errs images are not uniform");

  /* get mask from QUAL */
  check(badpix_mask = cpl_mask_duplicate(xsh_pre_get_bpmap(pre)));

  xsh_msg ("Nb of Bad Pixels in QUAL: %d",cpl_mask_count(badpix_mask));
  
  /* in all cases */

  /* computes the median of data and errs */
  check( xsh_bpmap_collapse_median(pre->data,dataList,badpix_mask));
  //Making a clean mean instead of a median
  check( xsh_collapse_errs(pre->errs,errsList, badpix_mask));


  //check( pre->data=xsh_imagelist_collapse_sigclip_iter_create(dataList,3,3,2));

  /* Add PRO KW */
  check( xsh_pfits_set_datancom( pre->data_header, size));

  /* remove irrelevant KW */
  check ( cpl_propertylist_erase_regexp( pre->data_header, "PRO BIAS *", 0));

  /* Now save the relevant frame of PRE
     With: median image, errs image and bad pixel map
  */
  sprintf(result_name,"%s.fits",result_tag);
  xsh_msg ("save frame %s", result_name); 
  check(medFrame = xsh_pre_save( pre, result_name, result_tag,1));

  check(cpl_frame_set_tag(medFrame, result_tag));


  cleanup:

    if (cpl_error_get_code () != CPL_ERROR_NONE) {
      xsh_free_frame(&medFrame);
    }
    xsh_free_imagelist(&dataList);
   
    //xsh_free_image(crh_ima);
    xsh_free_mask(&badpix_mask);
    xsh_pre_free(&pre);
    xsh_free_imagelist(&errsList) ;


   return medFrame;

}

cpl_frame* 
xsh_remove_crh_multiple (cpl_frameset * rawFrames, 
                         const char *result_tag, 
                         xsh_clipping_param * crh_clipping, 
			 xsh_instrument* instrument, 
                         cpl_imagelist ** list,
                         cpl_image** crh_ima)
{
  cpl_frame * result = NULL ;

  XSH_ASSURE_NOT_NULL( rawFrames ) ;
  XSH_ASSURE_NOT_NULL( result_tag ) ;
  XSH_ASSURE_NOT_NULL( crh_clipping ) ;
  XSH_ASSURE_NOT_NULL( instrument ) ;
  check( result = xsh_remove_cosmics( rawFrames, result_tag, crh_clipping,
				      instrument, list, crh_ima, 0 ) ) ;
 cleanup:
  return result ;
}

cpl_frame* 
xsh_combine_offset (cpl_frameset * rawFrames, 
		    const char * result_tag, 
		    xsh_clipping_param * crh_clipping, 
		    xsh_instrument * instrument, 
		    cpl_imagelist ** list,
		    cpl_image** crh_ima)
{
  cpl_frame * result = NULL ;

  XSH_ASSURE_NOT_NULL( rawFrames ) ;
  XSH_ASSURE_NOT_NULL( result_tag ) ;
  XSH_ASSURE_NOT_NULL( crh_clipping ) ;
  XSH_ASSURE_NOT_NULL( instrument ) ;

  check( result = xsh_remove_cosmics( rawFrames, result_tag, crh_clipping,
				      instrument, list, crh_ima, 1 ) ) ;
 cleanup:
  return result ;
}

/**@}*/
