/*                                                                           a
 *   This file is part of the ESO X-shooter Pipeline                         *
 *   Copyright (C) 2006 European Southern Observatory                        *
 *                                                                           *
 *   This library is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the Free Software             *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA    *
 *                                                                           */


/*
 * $Author: amodigli $
 * $Date: 2011/01/07 09:21:40 $
 * $Revision: 1.12 $
 * $Name: HEAD $
 */
/*----------------------------------------------------------------------------
  Includes
  ----------------------------------------------------------------------------*/
#include <math.h>
#include <string.h>
#include <xsh_error.h>
#include <xsh_utils_ifu.h>
#include <xsh_utils_image.h>
#include <xsh_utils_wrappers.h>
#include <xsh_pfits.h>
#include <xsh_model_io.h>
#include <xsh_data_rec.h>
#include <xsh_dfs.h>

#define PI_NUMB     (3.1415926535897932384626433832795)

static double
xsh_interpol(double x,double x1,double x2,double y1,double y2)
{
   return y1+(y2-y1)/(x2-x1)*(x-x1);
}


static double
xsh_sess2deg(const double sess);

static void
xsh_plist_set_coord1(cpl_propertylist** plist,
                 const double crpix1,
                 const double crval1,
                 const double cdelt1);
static void
xsh_plist_set_coord2(cpl_propertylist** plist,
                 const double crpix2,
                 const double crval2,
                 const double cdelt2);
static void
xsh_plist_set_coord3(cpl_propertylist** plist,
                 const int crpix3,
                 const double crval3,
                 const double cdelt3);


static void
xsh_plist_set_cd_matrix2(cpl_propertylist** plist,
            const double cd1_1,
            const double cd1_2,
            const double cd2_1,
            const double cd2_2);


static void
xsh_plist_set_cd_matrix3(cpl_propertylist** plist,
                     const double cd1_3,
                     const double cd2_3,
                     const double cd3_1,
                     const double cd3_2,
                     const double cd3_3);



/**
@brief set world coordinate system
@param plist input propertylist
@param crpix1 value of CRPIX1 (ref pixel axis 1 coord in pix coordinates)
@param crval1 value of CRVAL1 (ref pixel axis 1 coord in sky coordinates)
@param cdelt1 value of CDELT1 (ref pixel axis 1 size )
@return updated propertylist
*/
static void
xsh_plist_set_coord1(cpl_propertylist** plist,
                 const double crpix1,
                 const double crval1,
                 const double cdelt1)
{
   cpl_propertylist_erase_regexp(*plist, "^CTYPE1",0);
    cpl_propertylist_insert_after_string(*plist,"EXPTIME","CTYPE1","RA---TAN");
    cpl_propertylist_set_comment(*plist, "CTYPE1", "Projected Rectascension");
    cpl_propertylist_erase_regexp(*plist, "^CRPIX1",0);
    cpl_propertylist_insert_after_double(*plist,"CTYPE1","CRPIX1", crpix1) ;
    cpl_propertylist_set_comment(*plist, "CRPIX1","Reference pixel in RA" ) ;

    cpl_propertylist_erase_regexp(*plist, "^CRVAL1",0);
    cpl_propertylist_insert_after_double(*plist, "CRPIX1", "CRVAL1", crval1 ) ;
    cpl_propertylist_set_comment(*plist, "CRVAL1","Reference RA" ) ;

    cpl_propertylist_erase_regexp(*plist, "^CDELT1",0);
    cpl_propertylist_insert_after_double(*plist,"CRVAL1","CDELT1",cdelt1 ) ;
    cpl_propertylist_set_comment(*plist, "CDELT1","pixel scale" ) ;

    cpl_propertylist_erase_regexp(*plist, "^CUNIT1",0);
    cpl_propertylist_insert_after_string(*plist, "CDELT1",  "CUNIT1", "deg" ) ;
    cpl_propertylist_set_comment(*plist, "CUNIT1","RA-UNIT" ) ;

    return;
}


/**
@brief set world coordinate system
@param plist input propertylist
@param crpix2 value of CRPIX2 (ref pixel axis 2 coord in pix coordinates)
@param crval2 value of CRVAL2 (ref pixel axis 2 coord in sky coordinates)
@param cdelt2 value of CDELT2 (ref pixel axis 2 size )
@return updated propertylist
*/
static void
xsh_plist_set_coord2(cpl_propertylist** plist,
                 const double crpix2,
                 const double crval2,
                 const double cdelt2)
{
    cpl_propertylist_erase_regexp(*plist, "^CTYPE2",0);
    cpl_propertylist_insert_after_string(*plist,"CUNIT1","CTYPE2","DEC--TAN");
    cpl_propertylist_set_comment(*plist, "CTYPE2", "Projected Declination") ;

    cpl_propertylist_erase_regexp(*plist, "^CRPIX2",0);
    cpl_propertylist_insert_after_double(*plist,"CTYPE2","CRPIX2",crpix2 ) ;
    cpl_propertylist_set_comment(*plist, "CRPIX2", "Reference pixel in DEC") ;

    cpl_propertylist_erase_regexp(*plist,"^CRVAL2",0);
    cpl_propertylist_insert_after_double(*plist,"CRPIX2","CRVAL2",crval2) ;
    cpl_propertylist_set_comment(*plist,"CRVAL2","Reference DEC") ;

    cpl_propertylist_erase_regexp(*plist,"^CDELT2",0);
    cpl_propertylist_insert_after_double(*plist,"CRVAL2","CDELT2",cdelt2 ) ;
    cpl_propertylist_set_comment(*plist,"CDELT2","pixel scale") ;

    cpl_propertylist_erase_regexp(*plist,"^CUNIT2",0);
    cpl_propertylist_insert_after_string(*plist,"CDELT2","CUNIT2", "deg" ) ;
    cpl_propertylist_set_comment(*plist,"CUNIT2","DEC-UNIT") ;


}



/**
@brief set world coordinate system
@param plist input propertylist
@param crpix3 value of CRPIX3 (ref pixel axis 3 coord in pix coordinates)
@param crval3 value of CRVAL3 (ref pixel axis 3 coord in sky coordinates)
@param cdelt3 value of CDELT3 (ref pixel axis 3 size )
@return updated propertylist
*/
static void
xsh_plist_set_coord3(cpl_propertylist** plist,
                 const int crpix3,
                 const double crval3,
                 const double cdelt3)
{
    cpl_propertylist_erase_regexp(*plist, "^CTYPE3",0);
    cpl_propertylist_insert_after_string(*plist,"EXPTIME", "CTYPE3", "WAVE" ) ;
    cpl_propertylist_set_comment(*plist,"CTYPE3","wavelength axis in microns") ;

    cpl_propertylist_erase_regexp(*plist, "^CRPIX3",0);
    cpl_propertylist_insert_after_int(*plist,"CTYPE3",  "CRPIX3", crpix3 ) ;
    cpl_propertylist_set_comment(*plist, "CRPIX3", "Reference pixel in z") ;

    cpl_propertylist_erase_regexp(*plist, "^CRVAL3",0);
    cpl_propertylist_insert_after_double(*plist,"CRPIX3", "CRVAL3", crval3) ;
    cpl_propertylist_set_comment(*plist, "CRVAL3", "central wavelength") ;

    cpl_propertylist_erase_regexp(*plist, "^CDELT3",0);

    cpl_propertylist_insert_after_double(*plist,"CRVAL3","CDELT3",cdelt3) ;
    cpl_propertylist_set_comment(*plist, "CDELT3", "microns per pixel") ;

    cpl_propertylist_erase_regexp(*plist, "^CUNIT3",0);
    cpl_propertylist_insert_after_string(*plist,"CDELT3", "CUNIT3", "mum" ) ;
    cpl_propertylist_set_comment(*plist, "CUNIT3",  "spectral unit" ) ;

}


/**
@brief set world coordinate system: CD matrix
@param plist input propertylist
@param cd1_1 value of CD1_1
@param cd1_2 value of CD1_2
@param cd2_1 value of CD2_1
@param cd2_2 value of CD2_2
@return updated propertylist
*/

static void
xsh_plist_set_cd_matrix2(cpl_propertylist** plist,
            const double cd1_1,
            const double cd1_2,
            const double cd2_1,
            const double cd2_2)
{

    check(cpl_propertylist_erase_regexp(*plist, "^CD1_1",0));
    check(cpl_propertylist_insert_after_double(*plist,"EXPTIME",
						     "CD1_1", cd1_1 )) ;
    check(cpl_propertylist_set_comment(*plist, "CD1_1",
					     "CD rotation matrix" )) ;

    check(cpl_propertylist_erase_regexp(*plist, "^CD1_2",0));
    check(cpl_propertylist_insert_after_double(*plist, "CD1_1",
						     "CD1_2", cd1_2 )) ;
    check(cpl_propertylist_set_comment(*plist, "CD1_2",
					     "CD rotation matrix" )) ;

    check(cpl_propertylist_erase_regexp(*plist, "^CD2_1",0));
    check(cpl_propertylist_insert_after_double(*plist, "CD1_2",
						     "CD2_1", cd2_1 )) ;
    check(cpl_propertylist_set_comment(*plist, "CD2_1",
					     "CD rotation matrix" )) ;

    check(cpl_propertylist_erase_regexp(*plist, "^CD2_2",0));
    check(cpl_propertylist_insert_after_double(*plist, "CD2_1",
						     "CD2_2", cd2_2 )) ;
    check(cpl_propertylist_set_comment(*plist, "CD2_2",
					     "CD rotation matrix" )) ;

 cleanup:
    return;


}


/**
@brief set world coordinate system: CD matrix
@param plist input propertylist
@param cd1_3 value of CD1_3
@param cd2_3 value of CD2_3
@param cd3_1 value of CD3_1
@param cd3_2 value of CD3_2
@param cd3_3 value of CD3_3

@return updated propertylist
*/

static void
xsh_plist_set_cd_matrix3(cpl_propertylist** plist,
                     const double cd1_3,
                     const double cd2_3,
                     const double cd3_1,
                     const double cd3_2,
                     const double cd3_3)
{


    check(cpl_propertylist_erase_regexp(*plist, "^CD1_3",0));
    check(cpl_propertylist_insert_after_double(*plist,"EXPTIME",
						     "CD1_3", cd1_3 )) ;
    check(cpl_propertylist_set_comment(*plist, "CD1_3",
					     "CD rotation matrix" )) ;


    check(cpl_propertylist_erase_regexp(*plist, "^CD2_3",0));
    check(cpl_propertylist_insert_after_double(*plist,"CD1_3",
						     "CD2_3", cd2_3 )) ;
    check(cpl_propertylist_set_comment(*plist, "CD2_3",
					     "CD rotation matrix" )) ;



    check(cpl_propertylist_erase_regexp(*plist, "^CD3_1",0));
    check(cpl_propertylist_insert_after_double(*plist,"CD2_3",
						     "CD3_1", cd3_1 )) ;
    check(cpl_propertylist_set_comment(*plist, "CD3_1",
					     "CD rotation matrix" )) ;

    check(cpl_propertylist_erase_regexp(*plist, "^CD3_2",0));
    check(cpl_propertylist_insert_after_double(*plist, "CD3_1",
						     "CD3_2", cd3_2 )) ;
    check(cpl_propertylist_set_comment(*plist, "CD3_2",
					     "CD rotation matrix" )) ;

    check(cpl_propertylist_erase_regexp(*plist, "^CD3_3",0));
    check(cpl_propertylist_insert_after_double(*plist, "CD3_2",
						     "CD3_3", cd3_3 )) ;
    check(cpl_propertylist_set_comment(*plist, "CD3_3",
					     "CD rotation matrix" )) ;

 cleanup:
    return;


}

/**
  @brief    Convert a double from hour minutes seconds to deg:
  @param    hms   angle in hour-minutes-seconds
  @return 0 if suggess, else -1;
*/

static double
xsh_hms2deg(const double hms)
{
  int hrs=0;
  int min=0;
  double sec=0;
  double deg=0;
  double rest=hms;
  int sign=1;

  //xsh_msg("hms=%f",hms);

  if(hms<0) {
    sign=-1;
    rest=-hms;
  }
  //xsh_msg("rest=%f",rest);
  //xsh_msg("sign=%d",sign);

  hrs=(int)(rest/10000.);
  //xsh_msg("hrs=%d",hrs);

  rest=rest-(double)(hrs*10000.);
  min=(int)(rest/100.);
  //xsh_msg("min=%d",min);

  sec=rest-(double)(min*100.);
  //xsh_msg("sec=%f",sec);

  deg=hrs*15+(double)(min/4.)+(double)(sec/240.);
  //xsh_msg("deg=%f",deg);

  deg=sign*deg;
  //xsh_msg("deg=%f",deg);

  return deg;

}

/**
  @brief    Convert a double from ssessagesimal to deg:
            203049.197= 20:30:49.197 = 20.5136658333
  @param    sess   angle in seesagesimal units (see above)
  @return 0 if suggess, else -1;
*/

static double
xsh_sess2deg(const double sess)
{
  int grad=0;
  int min=0;
  double sec=0;
  double deg=0;
  double rest=sess;
  int sign=1;

  //xsh_msg("sess=%f",sess);

  if(sess<0) {
    sign=-1;
    rest=-sess;
  }
  //xsh_msg("rest=%f",rest);
  //xsh_msg("sign=%d",sign);

  grad=(int)(rest/10000.);
  //xsh_msg("grad=%d",grad);

  rest=rest-(double)(grad*10000.);
  min=(int)(rest/100.);
  //xsh_msg("min=%d",min);

  sec=rest-(double)(min*100.);
  //xsh_msg("sec=%f",sec);

  deg=grad+(double)(min/60.)+(double)(sec/3600.);
  //xsh_msg("deg=%f",deg);

  deg=sign*deg;
  //xsh_msg("deg=%f",deg);

  return deg;

}




/**
@brief set world coordinate system for a cube
@name xsh_change_plist_cube
@param plist input propertylist
@param name file name
@param cenLambda central wawelength
@param dispersion   dispersion
@param cenpix central pixel
@param center_x   center x
@param center_y   center y
*/
cpl_error_code
xsh_cube_set_wcs(cpl_propertylist * plist,
            float cenLambda,
            float dispersion,
            int   center_z,
            float center_x,
            float center_y )
{

    float pixelscale ;
    double ra ;
    double dec ;
    double angle ;
    float radangle ;
    double cd1_1, cd1_2, cd2_1, cd2_2 ;
    int sign_swap = -1;

 
    double cdelt1=0;
    double cdelt2=0;
    double cdelt3=dispersion;

    double crpix1=center_x;
    double crpix2=center_y;
    int crpix3=center_z;

    double crval1=0;
    double crval2=0;
    double crval3=cenLambda;


 
    pixelscale = xsh_pfits_get_pixscale(plist)/2. ;
    ra = xsh_pfits_get_ra(plist) ;
    dec = xsh_pfits_get_dec(plist) ;

    //get better coordinate values
    ra=xsh_pfits_get_tel_targ_alpha(plist);
    dec=xsh_pfits_get_tel_targ_delta(plist);
    //xsh_msg("ra=%f",ra);
    //xsh_msg("dec=%f",dec);
    ra=xsh_hms2deg(ra);
    dec=xsh_sess2deg(dec);
    //xsh_msg("ra=%f",ra);
    //xsh_msg("dec=%f",dec);

    crval1=ra;
    crval2=dec;

    angle = xsh_pfits_get_posangle(plist) ;
    /* in PUPIL data there is not posangle info: we reset the error */
    if(cpl_error_get_code() != CPL_ERROR_NONE) {
       cpl_error_reset();
    }
    cdelt1=sign_swap*pixelscale / 3600.;
    cdelt2=         +pixelscale / 3600.;


    radangle = angle * PI_NUMB / 180. ;
    cd1_1 = +cdelt1*cos(radangle);
    cd1_2 = -cdelt2*sin(radangle);
    cd2_1 = +cdelt1*sin(radangle);
    cd2_2 = +cdelt2*cos(radangle);


    xsh_plist_set_coord1(&plist,crpix1,crval1,cdelt1);
    xsh_plist_set_coord2(&plist,crpix2,crval2,cdelt2);
    xsh_plist_set_coord3(&plist,crpix3,crval3,cdelt3);
    xsh_plist_set_cd_matrix2(&plist,cd1_1,cd1_2,cd2_1,cd2_2);
    xsh_plist_set_cd_matrix3(&plist,0,0,0,0,dispersion);

    return cpl_error_get_code();
}







/**@{*/

/*---------------------------------------------------------------------------*/
/**
   @brief check edge values (llx,urx) fit in image size (1,nx)
   @param px [IN] point position
   @param nx [IN] x image size
   @param rad_x [IN] x radii
   @param llx [OUT] lower left X
   @param urx [OUT] upper right X

   @return void
*/
/*---------------------------------------------------------------------------*/


void
xsh_edge_check(const int px,const int nx,const int rad_x,
               int* llx,int* urx)
{


  *llx=px-rad_x;
  *urx=px+rad_x;
  *llx=(*llx>1) ? *llx : 1;
  *urx=(*urx<nx) ? *urx : nx;
  return;

}


/*---------------------------------------------------------------------------*/
/**
   @brief convert X-Y coordinates to Wavelengt Slit space
   @param x_centroid  centrod position on X axis 
   @param p_obj_cen [IN] pointer to central object array
   @param p_slit [IN] pointer to slit array
   @param p_wave [IN] pointer to wave array
   @param lly   [IN] Lower left X
   @param nx  [IN] number of X points (vector size) l
   @param row [IN] detector row position(Y)
   @param p_obj_cen_s [OUT] pointer to object s position
   @param p_obj_cen_w [OUT] pointer to object s position
 
   @return void
*/
/*---------------------------------------------------------------------------*/

void
xsh_convert_xy_to_ws(double x_centroid, 
		     double* p_obj_cen,
		     double* p_slit,
		     double* p_wave,
		     const int lly,
		     const int nx,
		     const int row,
		     double* p_obj_cen_s,
		     double* p_obj_cen_w)
{
  int x_ceil=0;
  int x_floor=0;
  double s_ceil=0;
  double s_floor=0;
  double w_ceil=0;
  double w_floor=0;

  p_obj_cen[row]=x_centroid;
  x_ceil=ceil(x_centroid);
  x_floor=floor(x_centroid);

  s_ceil =p_slit[lly*nx+x_ceil];
  s_floor=p_slit[lly*nx+x_floor];

  w_ceil =p_wave[lly*nx+x_ceil];
  w_floor=p_wave[lly*nx+x_floor];

  p_obj_cen_s[row]=xsh_interpol(x_centroid,x_floor,x_ceil,s_floor,s_ceil);
  p_obj_cen_w[row]=xsh_interpol(x_centroid,x_floor,x_ceil,w_floor,w_ceil);

  return;
}


/*---------------------------------------------------------------------------*/
/**
   @brief swap lower and upper edge columns
   @param tab [INOUT] pointer to input/output table
 
   @return if no error, CPL_ERROR_NONE
*/
/*---------------------------------------------------------------------------*/

cpl_error_code
xsh_table_edges_swap_low_upp(cpl_table** tab)
{
  /* swap LOW/UPP in case of VIS */
  cpl_table_duplicate_column(*tab,"OBJ_LOW_S_TMP",*tab,"OBJ_LOW_S");
  cpl_table_duplicate_column(*tab,"OBJ_UPP_S_TMP",*tab,"OBJ_UPP_S");

  cpl_table_erase_column(*tab,"OBJ_LOW_S");
  cpl_table_erase_column(*tab,"OBJ_UPP_S");

  cpl_table_duplicate_column(*tab,"OBJ_UPP_S",*tab,"OBJ_LOW_S_TMP");
  cpl_table_duplicate_column(*tab,"OBJ_LOW_S",*tab,"OBJ_UPP_S_TMP");

  cpl_table_erase_column(*tab,"OBJ_LOW_S_TMP");
  cpl_table_erase_column(*tab,"OBJ_UPP_S_TMP");

  cpl_table_duplicate_column(*tab,"OBJ_LOW_W_TMP",*tab,"OBJ_LOW_W");
  cpl_table_duplicate_column(*tab,"OBJ_UPP_W_TMP",*tab,"OBJ_UPP_W");

  cpl_table_erase_column(*tab,"OBJ_LOW_W");
  cpl_table_erase_column(*tab,"OBJ_UPP_W");

  cpl_table_duplicate_column(*tab,"OBJ_UPP_W",*tab,"OBJ_LOW_W_TMP");
  cpl_table_duplicate_column(*tab,"OBJ_LOW_W",*tab,"OBJ_UPP_W_TMP");

  cpl_table_erase_column(*tab,"OBJ_LOW_W_TMP");
  cpl_table_erase_column(*tab,"OBJ_UPP_W_TMP");


  return cpl_error_get_code();
}



/*---------------------------------------------------------------------------*/
/**
   @brief Add to edges table columns to store LOW/CEN/UPP object positions
   @param name [IN] input file name
 
   @return corrected table
*/
/*---------------------------------------------------------------------------*/

cpl_table*
xsh_table_edge_prepare(const char* name)
{
  cpl_table* tab=NULL;
  int nrow=0;

  check(tab=cpl_table_load(name,2,0));

  nrow=cpl_table_get_nrow(tab);

  cpl_table_new_column(tab,"OBJ_LOW_X",CPL_TYPE_DOUBLE);
  cpl_table_new_column(tab,"OBJ_CEN_X",CPL_TYPE_DOUBLE);
  cpl_table_new_column(tab,"OBJ_UPP_X",CPL_TYPE_DOUBLE);

  cpl_table_fill_column_window(tab,"OBJ_LOW_X",0,nrow,-1);
  cpl_table_fill_column_window(tab,"OBJ_CEN_X",0,nrow,-1);
  cpl_table_fill_column_window(tab,"OBJ_UPP_X",0,nrow,-1);

  cpl_table_new_column(tab,"OBJ_LOW_S",CPL_TYPE_DOUBLE);
  cpl_table_new_column(tab,"OBJ_LOW_W",CPL_TYPE_DOUBLE);
  cpl_table_new_column(tab,"OBJ_CEN_S",CPL_TYPE_DOUBLE);
  cpl_table_new_column(tab,"OBJ_CEN_W",CPL_TYPE_DOUBLE);
  cpl_table_new_column(tab,"OBJ_UPP_S",CPL_TYPE_DOUBLE);
  cpl_table_new_column(tab,"OBJ_UPP_W",CPL_TYPE_DOUBLE);

  cpl_table_fill_column_window(tab,"OBJ_LOW_S",0,nrow,-1);
  cpl_table_fill_column_window(tab,"OBJ_LOW_W",0,nrow,-1);
  cpl_table_fill_column_window(tab,"OBJ_CEN_S",0,nrow,-1);
  cpl_table_fill_column_window(tab,"OBJ_CEN_W",0,nrow,-1);
  cpl_table_fill_column_window(tab,"OBJ_UPP_S",0,nrow,-1);
  cpl_table_fill_column_window(tab,"OBJ_UPP_W",0,nrow,-1);

 cleanup:
  return tab;

}

/* Not used
static cpl_error_code
xsh_tab_clean_traces(cpl_table** tab,
                     const double fit_wmin,
                     const double fit_wmax,
                     const char* wave_col_name,
                     const char* trace_col_name,
                     const char* fit_col_name,
		     const char* res_col_name,
		     const char* flag_col_name,
                     const int niter,
		     const double kappa)
{

  double* pdata=NULL;
  double* pwave=NULL;
  double* pres=NULL;
  double* pfit=NULL;
  int* pflag=NULL;

  cpl_polynomial* p_fit=NULL;
  cpl_vector* vdata=NULL;
  cpl_vector* vwave=NULL;
  int fit_pows[2];
  double fit_off=0;
  double fit_slope=0;
  int row=0;
  int nrow=0;
  double mse_fit=0;
  double res_stdev;
  double res_max=1;
  int i=0;
  cpl_table* ext=NULL;
  int next=0;

  check(nrow=cpl_table_get_nrow(*tab));
  cpl_table_and_selected_double(*tab,wave_col_name,CPL_GREATER_THAN,fit_wmin);
  cpl_table_and_selected_double(*tab,wave_col_name,CPL_LESS_THAN,fit_wmax);
  ext=cpl_table_extract_selected(*tab);
  check(next=cpl_table_get_nrow(ext));


  pdata=cpl_table_get_data_double(ext,trace_col_name);
  pwave=cpl_table_get_data_double(ext,wave_col_name);
  vdata=cpl_vector_wrap(next,pdata);
  vwave=cpl_vector_wrap(next,pwave);
  p_fit=xsh_polynomial_fit_1d_create(vwave,vdata,1,&mse_fit);

  cpl_polynomial_dump(p_fit,stdout);
  fit_pows[0]=0;
  check(fit_off=cpl_polynomial_get_coeff(p_fit,fit_pows));
  fit_pows[0]=1;
  check(fit_slope=cpl_polynomial_get_coeff(p_fit,fit_pows));



  cpl_table_new_column(*tab,fit_col_name,CPL_TYPE_DOUBLE);
  cpl_table_fill_column_window(*tab,fit_col_name,0,nrow,-1);
  pfit=cpl_table_get_data_double(*tab,fit_col_name);

  cpl_table_new_column(*tab,res_col_name,CPL_TYPE_DOUBLE);
  cpl_table_fill_column_window(*tab,res_col_name,0,nrow,-1);
  pres=cpl_table_get_data_double(*tab,res_col_name);

  cpl_table_new_column(*tab,flag_col_name,CPL_TYPE_INT);
  cpl_table_fill_column_window(*tab,flag_col_name,0,nrow,-1);
  pflag=cpl_table_get_data_int(*tab,flag_col_name);



  pdata=cpl_table_get_data_double(*tab,trace_col_name);
  pwave=cpl_table_get_data_double(*tab,wave_col_name);

  for(row=0;row<nrow;row++) {
    pfit[row]=fit_off+pwave[row]*fit_slope;
    pres[row]=pfit[row]-pdata[row];
  }


  // kappa-sigma-clip upper trace
  for(i=0;i<niter;i++) {
    res_stdev=cpl_table_get_column_stdev(*tab,res_col_name);
    xsh_msg("res_stdev=%g",res_stdev);
    for(row=0;row<nrow;row++) {
      if((fabs(pres[row])>kappa*res_stdev) || 
         (fabs(pres[row])>res_max)) {
        //xsh_msg("set invalid %d res=%g",row,pres[row]);
	cpl_table_set_invalid(*tab,res_col_name,row);
      } else {
	pflag[row]=1;
      }
    }
  }

  cpl_table_erase_invalid(*tab);

 
 cleanup:
  return cpl_error_get_code();

}
*/




/*---------------------------------------------------------------------------*/
/**
   @brief Function to calibrate object traces in IFU mode
   @param ifu_object_ff_name name of input object frame (flat field corrected)
   @param order_tab_edges_ifu_name name of input edge order table
   @param slit_map_name slit map filename
   @param wave_map_name wave map filename
   @return cpl error code
*/
/*---------------------------------------------------------------------------*/

cpl_error_code
xsh_ifu_trace_object_calibrate(const char* ifu_object_ff_name,
                               const char* order_tab_edges_ifu_name,
                               const char* slit_map_name,
                               const char* wave_map_name)
{
  cpl_image* ifu_object_ff_ima=NULL;
  cpl_image* wave_map_ima=NULL;
  cpl_image* slit_map_ima=NULL;
  cpl_image* ratio_ima=NULL;
  cpl_table* tab=NULL;

  int sx=0;
  int sy=0;

  int nrow=0;

  double* p_edge_lo_x=NULL;
  double* p_edge_up_x=NULL;

  double* p_slice_lo_x=NULL;
  double* p_slice_up_x=NULL;

  double* p_center_y=NULL;

  double* p_obj_low=NULL;
  double* p_obj_cen=NULL;
  double* p_obj_upp=NULL;

  double* p_obj_low_s=NULL;
  double* p_obj_low_w=NULL;
  double* p_obj_cen_s=NULL;
  double* p_obj_cen_w=NULL;
  double* p_obj_upp_s=NULL;
  double* p_obj_upp_w=NULL;

  double* p_slit=NULL;
  double* p_wave=NULL;

  double x_centroid=0;

  int llx=0;
  int lly=0;
  int urx=0;
  int ury=0;

  int row=0;

  double range1_w_min=0;
  double range1_w_max=0;
  double range2_w_min=0;
  double range2_w_max=0;

  double range_ks_w_min=0;
  double range_ks_w_max=0;

  int nx=0;
  int ny=0;
  cpl_propertylist* plist=NULL;
  XSH_ARM arm=XSH_ARM_UNDEFINED;
  const char* pcatg=NULL;
  char tag[25];
  char name[80];
  int px=0;
  int py=0;

  int rad_x=10;
/*
  int niter=2;
  int kappa=2.0;
*/
  int binx=1;
  int biny=1;

  typedef enum {centroid, gaussian} xsh_fit_method;
  xsh_fit_method fit_method = centroid;


  check(ifu_object_ff_ima=cpl_image_load(ifu_object_ff_name,CPL_TYPE_DOUBLE,0,0));
  plist=cpl_propertylist_load(ifu_object_ff_name,0);
  pcatg=xsh_pfits_get_pcatg(plist);

  xsh_msg("pcatg=%s",pcatg);

  if(strstr(pcatg,"UVB") != NULL) {
    arm=XSH_ARM_UVB;
    range1_w_min=300;
    range1_w_max=500;
    range_ks_w_min=380;
    range_ks_w_max=550;
    binx=xsh_pfits_get_binx(plist);
    biny=xsh_pfits_get_biny(plist);


  } else if(strstr(pcatg,"VIS") != NULL) {
    arm=XSH_ARM_VIS;
    range1_w_min=600;
    range1_w_max=950;

    range_ks_w_min=600;
    range_ks_w_max=900;
    binx=xsh_pfits_get_binx(plist);
    biny=xsh_pfits_get_biny(plist);

  } else if(strstr(pcatg,"NIR") != NULL) {
    arm=XSH_ARM_NIR;
    range1_w_min=1050;
    range1_w_max=1250;
    range2_w_min=2000;
    range2_w_max=2200;

    range_ks_w_min=1100;
    range_ks_w_max=2200;


  }

  check(slit_map_ima=cpl_image_load(slit_map_name,CPL_TYPE_DOUBLE,0,0));
  check(wave_map_ima=cpl_image_load(wave_map_name,CPL_TYPE_DOUBLE,0,0));
  sx=cpl_image_get_size_x(ifu_object_ff_ima);
  sy=cpl_image_get_size_y(ifu_object_ff_ima);

  check(tab=xsh_table_edge_prepare(order_tab_edges_ifu_name));
  nrow=cpl_table_get_nrow(tab);

  p_edge_lo_x=cpl_table_get_data_double(tab,"EDG_LO_X");
  p_slice_lo_x=cpl_table_get_data_double(tab,"SLIC_LO_X");
  p_obj_low=cpl_table_get_data_double(tab,"OBJ_LOW_X");
  p_obj_low_s=cpl_table_get_data_double(tab,"OBJ_LOW_S");
  p_obj_low_w=cpl_table_get_data_double(tab,"OBJ_LOW_W");

  p_edge_up_x=cpl_table_get_data_double(tab,"EDG_UP_X");
  p_slice_up_x=cpl_table_get_data_double(tab,"SLIC_UP_X");
  p_obj_upp=cpl_table_get_data_double(tab,"OBJ_UPP_X");
  p_obj_upp_s=cpl_table_get_data_double(tab,"OBJ_UPP_S");
  p_obj_upp_w=cpl_table_get_data_double(tab,"OBJ_UPP_W");
 
  p_center_y=cpl_table_get_data_double(tab,"CENTER_Y");
  p_obj_cen=cpl_table_get_data_double(tab,"OBJ_CEN_X");
  p_obj_cen_s=cpl_table_get_data_double(tab,"OBJ_CEN_S");
  p_obj_cen_w=cpl_table_get_data_double(tab,"OBJ_CEN_W");

  p_slit=cpl_image_get_data_double(slit_map_ima);
  p_wave=cpl_image_get_data_double(wave_map_ima);
  nx=cpl_image_get_size_x(wave_map_ima);
  ny=cpl_image_get_size_y(wave_map_ima);

  //xsh_msg("binx=%d biny=%d",binx,biny);
  for(row=0;row<nrow;row++) {

     lly=(int)p_center_y[row]/biny;
     ury=(int)p_center_y[row]/biny;

     //xsh_msg("lly=%d ury=%d nx=%d ny=%d",lly,ury,nx,ny);
     if((llx<nx) && (ury<ny) ) {

        /* CEN IFU slice obj position */
        llx=floor(p_slice_lo_x[row]);
        urx=ceil(p_slice_up_x[row]);

        check(cpl_image_get_maxpos_window(ifu_object_ff_ima,
					  llx,lly,urx,ury,&px,&py));

        xsh_edge_check(px,nx,rad_x,&llx,&urx);
	if(fit_method==centroid) {
	  x_centroid=cpl_image_get_centroid_x_window(ifu_object_ff_ima,
						     llx,lly,urx,ury);

	} else {
	  check(x_centroid=xsh_image_fit_gaussian_max_pos_x_window(ifu_object_ff_ima,
								   llx,urx,lly));
	}


        p_obj_cen[row]=x_centroid;
        /* now measure position in arcsecs and wavelength */
        if(urx<nx) {
	  check(xsh_convert_xy_to_ws(x_centroid,p_obj_cen,p_slit,p_wave,
				     lly,nx,row,p_obj_cen_s,p_obj_cen_w));
        }

        /* LOW IFU slice obj position */
        llx=floor(p_edge_lo_x[row]);
        urx=ceil(p_slice_lo_x[row]);
        cpl_image_get_maxpos_window(ifu_object_ff_ima,
                                 llx,lly,urx,ury,&px,&py);

        xsh_edge_check(px,nx,rad_x,&llx,&urx);
	if(fit_method==centroid) {
	  check(x_centroid=cpl_image_get_centroid_x_window(ifu_object_ff_ima,
							   llx,lly,urx,ury));
	} else {
	  x_centroid=xsh_image_fit_gaussian_max_pos_x_window(ifu_object_ff_ima,
							     llx,urx,lly);
	}


        /* now measure position in arcsecs and wavelength */
        if(urx<nx) {
	  check(xsh_convert_xy_to_ws(x_centroid,p_obj_low,p_slit,p_wave,
				     lly,nx,row,p_obj_low_s,p_obj_low_w));
        }

        /* UPP IFU slice obj position */
        llx=floor(p_slice_up_x[row]);
        urx=ceil(p_edge_up_x[row]);
        cpl_image_get_maxpos_window(ifu_object_ff_ima,
                                 llx,lly,urx,ury,&px,&py);

        xsh_edge_check(px,nx,rad_x,&llx,&urx);
	if(fit_method==centroid) {
	  x_centroid=cpl_image_get_centroid_x_window(ifu_object_ff_ima,
						     llx,lly,urx,ury);
	} else {
	  x_centroid=xsh_image_fit_gaussian_max_pos_x_window(ifu_object_ff_ima,
							     llx,urx,lly);
	}

        p_obj_upp[row]=x_centroid;
        /* now measure position in arcsecs and wavelength */
        if(urx<nx) {
	  check(xsh_convert_xy_to_ws(x_centroid,p_obj_upp,p_slit,p_wave,
				     lly,nx,row,p_obj_upp_s,p_obj_upp_w));
        }
     }
  }
 
  if(arm==XSH_ARM_VIS) {
    xsh_table_edges_swap_low_upp(&tab);
  }
  cpl_table_duplicate_column(tab,"OBJ_CEN_PLUS_UPP_S",tab,"OBJ_CEN_S");

  cpl_table_duplicate_column(tab,"OBJ_CEN_PLUS_LOW_S",tab,"OBJ_CEN_S");

  cpl_table_add_columns(tab,"OBJ_CEN_PLUS_UPP_S","OBJ_UPP_S");
  cpl_table_add_columns(tab,"OBJ_CEN_PLUS_LOW_S","OBJ_LOW_S");



  /* we now clean results: 
     -fit linear to OBJ_CEN_PLUS_UPP/LOW_S
     -compute residuals fit-value
     -compute stdev residuals
     -apply kappa-sigma clip (2 times) flagging as null clipped values
 */

  /*
  check(xsh_tab_clean_traces(&tab,range_ks_w_min,range_ks_w_max,
			     "OBJ_CEN_W","OBJ_CEN_PLUS_UPP_S",
			     "CEN_UPP_S_FIT","CEN_UPP_S_RES",
			     "CEN_UPP_S_FLAG",niter,kappa));

  check(xsh_tab_clean_traces(&tab,range_ks_w_min,range_ks_w_max,
                             "OBJ_CEN_W","OBJ_CEN_PLUS_LOW_S",
			     "CEN_LOW_S_FIT","CEN_LOW_S_RES",
			     "CEN_LOW_S_FLAG",niter,kappa));

  */
  /* save result */
  sprintf(tag,"TRACE_OBJ_%s",xsh_arm_tostring(arm));
  sprintf(name,"%s.fits",tag);
  xsh_pfits_set_pcatg(plist,tag);
  check(cpl_table_save(tab,plist,NULL,name,CPL_IO_DEFAULT));



cleanup:
  xsh_free_image(&ifu_object_ff_ima);
  xsh_free_image(&slit_map_ima);
  xsh_free_image(&wave_map_ima);
  xsh_free_image(&ratio_ima);
  xsh_free_propertylist(&plist);
  return cpl_error_get_code();

}


/*---------------------------------------------------------------------------*/
/**
   @brief Check if a model configuration frame has been corrected for flexures
   @param model_config_frame model configuration frame
   @param sci_frame science frame
   @param instrument instrument setting
   @return cpl error code
*/
/*---------------------------------------------------------------------------*/

cpl_error_code 
xsh_frame_check_model_cfg_is_proper_for_sci(cpl_frame* model_config_frame,
                                            cpl_frame* sci_frame,
                                            xsh_instrument* instrument)
{

   cpl_propertylist* afc_plist=NULL;
   cpl_propertylist* sci_plist=NULL;
   const char* name=NULL;
   const char* sci_obs_targ_name=NULL;
   const char* afc_obs_targ_name=NULL;
   const char* afc_slit_value=NULL;
   const char* sci_slit_value=NULL;
   int sci_obs_id=0;
   int afc_obs_id=0;

   check(name=cpl_frame_get_filename(model_config_frame));
   check(afc_plist=cpl_propertylist_load(name,0));
 
   check(name=cpl_frame_get_filename(sci_frame));
   check(sci_plist=cpl_propertylist_load(name,0));

   check(afc_slit_value=xsh_pfits_get_slit_value(afc_plist,instrument ));

   check(sci_slit_value=xsh_pfits_get_slit_value(sci_plist,instrument ));

   if(strstr(afc_slit_value,"Pin_0.5") == NULL) {
      xsh_msg_error("You have used uncorrect AFC corrected model cfg frame");
      xsh_msg_error("IFU AFC corrected model CFG must have INS.OPTIi.NAME='Pin_0.5'");
      cpl_error_set(cpl_func,CPL_ERROR_ILLEGAL_INPUT);
   }


   check(sci_obs_targ_name=xsh_pfits_get_obs_targ_name(sci_plist));
   check(afc_obs_targ_name=xsh_pfits_get_obs_targ_name(afc_plist));
   if(strcmp(sci_obs_targ_name,afc_obs_targ_name) != 0) {
      xsh_msg_error("Improper AFC corrected model cfg frame to reduce sci frame");
      xsh_msg_error("Their OBS.TARG.NAME values must match");
      cpl_error_set(cpl_func,CPL_ERROR_ILLEGAL_INPUT);

   }   

   check(sci_obs_id=xsh_pfits_get_obs_id(sci_plist));
   check(afc_obs_id=xsh_pfits_get_obs_id(afc_plist));

   if(sci_obs_id != afc_obs_id) {
      xsh_msg_error("Improper AFC corrected model cfg frame to reduce sci frame");
      xsh_msg_error("Their OBS.ID values must match");
      cpl_error_set(cpl_func,CPL_ERROR_ILLEGAL_INPUT);
   }

  cleanup:

   xsh_free_propertylist(&sci_plist);
   xsh_free_propertylist(&afc_plist);

   return cpl_error_get_code();
}


/*---------------------------------------------------------------------------*/
/**
   @brief utility to check if a frame has been corrected for flexures
   @param model_config_frame  input frame
 
   @return corrected table
*/
/*---------------------------------------------------------------------------*/

cpl_error_code 
xsh_frame_check_model_cfg_is_afc_corrected(cpl_frame* model_config_frame){

   cpl_propertylist* plist=NULL;
   const char* name=NULL;
   const char* raw1_catg=NULL;

   check(name=cpl_frame_get_filename(model_config_frame));
   check(plist=cpl_propertylist_load(name,0));
   check(raw1_catg=xsh_pfits_get_raw1catg(plist));
   if(strstr(raw1_catg,"AFC_ATT") == NULL) {
      xsh_msg_error("model cfg frame seems not to be AFC corrected");
      xsh_msg_error("Their PRO.REC1.RAW1..NAME values must contain AFC_ATT");
      cpl_error_set(cpl_func,CPL_ERROR_ILLEGAL_INPUT);
   }
   
  cleanup:

   xsh_free_propertylist(&plist);

  return cpl_error_get_code();

}



/*---------------------------------------------------------------------------*/
/**
   @brief Reconstruct IFU cube
   @param[in] div_frame input frame flat field corrected
   @param[in] ifu_config_frame  IFU configuration frame
   @param[in] spectral_format_frame spectral format frame
   @param[in] model_config_frame model configuration frame
   @param[in] wavesol_frame  wavelength solution frame
   @param[in] instrument instrument (arm) setting
   @param[in] frameset input frame set
   @param[in] parameters input parameter list
   @param[in] rectify_par input parameters to specify rectification
   @param[in] recipe_id input recipe id
   @param[in] rec_prefix input recipe prefix
   @param[in] frame_is_object flag to specify if input frame is object or sky
   @return corrected table
*/
/*---------------------------------------------------------------------------*/
cpl_error_code
xsh_build_ifu_cube(cpl_frame* div_frame, 
                   cpl_frame* ifu_config_frame,
                   cpl_frame* spectral_format_frame,
                   cpl_frame* model_config_frame,
                   cpl_frame* wavesol_frame,
                   xsh_instrument* instrument,
                   cpl_frameset* frameset,
                   cpl_parameterlist* parameters,
                   xsh_rectify_param * rectify_par,
                   const char* recipe_id,
                   const char* rec_prefix,
                   const int frame_is_object)
{


  /* Parameters */

  cpl_image* data_img=NULL;
  cpl_image* errs_img=NULL;
  cpl_image* qual_img=NULL;

  cpl_image* data_tmp=NULL;
  cpl_image* errs_tmp=NULL;
  cpl_image* qual_tmp=NULL;
  cpl_image* mask_tmp=NULL;
  cpl_frame* qc_trace_frame=NULL;
  cpl_frame* qc_trace_merged_frame=NULL;
  cpl_frame* frame_data_cube=NULL;
  cpl_frame* frame_errs_cube=NULL;
  cpl_frame* frame_qual_cube=NULL;
  cpl_frame* frame_merged_cube=NULL;


  /* Intermediate frames */

  /* variables */
  XSH_ARM arm;
  xsh_xs_3 model_config ;
  xsh_rec_list * rec_list = NULL ;

  cpl_table* ifu_cfg_tab=NULL;
  cpl_table* sp_fmt_tab=NULL;
  cpl_vector* profile=NULL;
  cpl_imagelist* data_cube=NULL;
  cpl_imagelist* errs_cube=NULL;
  cpl_imagelist* qual_cube=NULL;

  cpl_imagelist* data_cube_merge=NULL;
  cpl_imagelist* errs_cube_merge=NULL;
  cpl_imagelist* qual_cube_merge=NULL;
  cpl_imagelist* mask_cube_merge=NULL;


  float* pima=NULL;
  float* perr=NULL;
  int* pqua=NULL;
  int* pmsk=NULL;
  int nord=0;
  int save_size_uvb=4;
  int save_size_vis=2;
  int save_size_nir=2;
  int save_size=0;
  const int peack_search_hsize=5;
  int method=0; /* Gaussian */



  cpl_propertylist* data_plist=NULL;
  cpl_propertylist* errs_plist=NULL;
  cpl_propertylist* qual_plist=NULL;


  const char* sci_name=NULL;
  const char* sp_fmt_name=NULL;
  const char* ifu_cfg_name=NULL;

  char name[256];
  char qualifier[10];
  int binx=1;
  int biny=1;

  double flux_upp=0;
  double flux_cen=0;
  double flux_low=0;


  double errs_upp=0;
  double errs_cen=0;
  double errs_low=0;
  double cube_wave_min=0;
  double cube_wave_max=0;


  int qual_upp=0;
  int qual_cen=0;
  int qual_low=0;

  double cube_wmin=550;
  double cube_wmax=1000;
  double cube_wstep=0;
  double cube_sstep=0;

  int is=0;

  int ord=0;
  int ord_min=0;
  int ord_max=0;

  double wave=0;
  double wave_min=0;
  double wave_max=0;
  double wave_step=0;

  double wave_min_old=0;
  double wave_max_old=0;

  int nstep_off=0;

  double s=0;
  double s_min=-2;
  double s_max=2;
  double s_step=0;

  double x=0;
  double y=0;
  double s_upp=0;
  double s_low=0;


  int radius=rectify_par->rectif_radius;

  double confidence=0;
  int naxis1=0;
  int naxis2=0;
  int naxis3=0;
  int ik=0;
  double wave_cen=0;

/*
  double stupid_offset_uvb=0.145;
  double stupid_offset_vis=0.09;
  double stupid_offset_nir=0.16;
*/
  double w_low_coeff1=0;
  double w_low_coeff2=0;

  double w_upp_coeff1=0;
  double w_upp_coeff2=0;

  double w_low_coeff1_uvb=-0.002972;     /*-0.00085*/
  double w_low_coeff2_uvb=2.26497e-6+0.2e-6;/*-2.26e-6;*/
  double w_upp_coeff1_uvb=8.3355331e-5;     /*-0.00085*/
  double w_upp_coeff2_uvb=-1.0682e-6;/*-2.26e-6;*/



  double w_low_coeff1_vis=-0.0016549569;
  double w_low_coeff2_vis=1.183805e-6+0.04e-6;
  double w_upp_coeff1_vis=-0.0016610719;
  double w_upp_coeff2_vis=1.0823013e-6+0.02e-6;

  double w_low_coeff1_nir=-0.00010;
  double w_low_coeff2_nir=-5.0e-9;

  double w_upp_coeff1_nir=0;
  double w_upp_coeff2_nir=0;


  double s_upp_off_uvb=4.0514429+0.16;/*1.91319135+stupid_offset_uvb;*/
  double s_upp_off_vis=5.2504895+0.16-0.703;/*+0.05; */
  double s_upp_off_nir=3.5452+0.31;
  double s_upp_off=0;

  double s_low_off_uvb=-3.4636662+0.16;/*-2.1704691+stupid_offset_uvb;*/

  double s_low_off_vis=-2.9428071+0.16-0.645;/*-0.07; */
  double s_low_off_nir=-4.451682+0.21+0.31;

  double s_low_off=0;

  int w_step_fct=10;

  int status=0;
  int mk=0;
  char tag[256];
  xsh_wavesol *wavesol=NULL;

  if(frame_is_object) {
     sprintf(qualifier,"OBJ");
  } else {
     sprintf(qualifier,"SKY");
  }
  /* NEW */
  /* Find STD star centroids (for each IFU slice):
     input STD star frame
     input order centre traces
     input order edges  traces
     input slit map
     output STD star centroids traces table+ QC on centroid measure
  */

  /* Find STD star centroids traces slices interdistance 
     (each measured at the same wavelength)
     Find STD star centroids distances to IFU edges traces
     (each measured at the same wavelength)
     input STD star centroid traces table
     input order edges/IFU slices traces table
     input slit map (orders are spatially over-sized)
     input wave map (orders are spatially over-sized)
     output table with STD star centroids traces slices interdistance
     output corresponding QC(statistics)
     output table with STD star distances to IFU edges traces
     output corresponding QC(statistics)
     ? output "measured" slit map
     ? output "measured" wave map
  */


  check(sci_name=cpl_frame_get_filename(div_frame));
  check(data_plist=cpl_propertylist_load(sci_name,0));
  arm=instrument->arm;


  cube_wstep=rectify_par->rectif_bin_lambda;
  cube_sstep=rectify_par->rectif_bin_space;

  if(arm == XSH_ARM_UVB) {
    cube_wmin=300;
    cube_wmax=550;

    binx=xsh_pfits_get_binx(data_plist);
    biny=xsh_pfits_get_biny(data_plist);

    s_low_off=s_low_off_uvb;
    s_upp_off=s_upp_off_uvb;
    w_low_coeff1=w_low_coeff1_uvb;
    w_low_coeff2=w_low_coeff2_uvb;

    w_upp_coeff1=w_upp_coeff1_uvb;
    w_upp_coeff2=w_upp_coeff2_uvb;
    save_size=save_size_uvb;
 

    //model_config.temper= xsh_pfits_get_temp2(data_plist);
  } else if(arm == XSH_ARM_VIS) {
     cube_wmin=550;
     cube_wmax=1000;

     //model_config.temper= xsh_pfits_get_temp5(data_plist);
     binx=xsh_pfits_get_binx(data_plist);
     biny=xsh_pfits_get_biny(data_plist);

     s_low_off=s_low_off_vis;
     s_upp_off=s_upp_off_vis;

    w_low_coeff1=w_low_coeff1_vis;
    w_low_coeff2=w_low_coeff2_vis;

    w_upp_coeff1=w_upp_coeff1_vis;
    w_upp_coeff2=w_upp_coeff2_vis;
    save_size=save_size_vis;

 
  } else if(arm == XSH_ARM_NIR) {
     cube_wmin=1000;
     cube_wmax=2500;

     //model_config.temper= xsh_pfits_get_temp82(data_plist);

     s_low_off=s_low_off_nir;
     s_upp_off=s_upp_off_nir;


    w_low_coeff1=w_low_coeff1_nir;
    w_low_coeff2=w_low_coeff2_nir;

    w_upp_coeff1=w_upp_coeff1_nir;
    w_upp_coeff2=w_upp_coeff2_nir;
    save_size=save_size_nir;

  }



  if(ifu_config_frame) {

    ifu_cfg_name=cpl_frame_get_filename(ifu_config_frame);
    ifu_cfg_tab=cpl_table_load(ifu_cfg_name,1,0);
    s_upp_off=cpl_table_get_double(ifu_cfg_tab,"S_UPP_OFF",arm,&status);
    s_low_off=cpl_table_get_double(ifu_cfg_tab,"S_LOW_OFF",arm,&status);

    w_upp_coeff1=cpl_table_get_double(ifu_cfg_tab,"W_UPP_COEF1",arm,&status);
    w_low_coeff1=cpl_table_get_double(ifu_cfg_tab,"W_LOW_COEF1",arm,&status);
    w_upp_coeff2=cpl_table_get_double(ifu_cfg_tab,"W_UPP_COEF2",arm,&status);
    w_low_coeff2=cpl_table_get_double(ifu_cfg_tab,"W_LOW_COEF2",arm,&status);

    w_step_fct=cpl_table_get_int(ifu_cfg_tab,"W_STEP_FCT",arm,&status);
    xsh_msg("s_upp_off=%10.8g",s_upp_off);
    xsh_msg("s_low_off=%10.8g",s_low_off);


    xsh_msg("w_upp_coeff1=%10.8g",w_upp_coeff1);
    xsh_msg("w_upp_coeff2=%10.8g",w_upp_coeff2);
    xsh_msg("w_low_coeff1=%10.8g",w_low_coeff1);
    xsh_msg("w_low_coeff2=%10.8g",w_low_coeff2);
 
    xsh_msg("w_step_fct=%d",w_step_fct);

  }
  cube_wstep*=w_step_fct; 
  if(model_config_frame!=NULL) {
     check( xsh_model_config_load_best( model_config_frame, &model_config));
     xsh_model_binxy(&model_config,binx,biny);
  } else {
     check( wavesol = xsh_wavesol_load(  wavesol_frame, instrument));
  }
  check( rec_list = xsh_rec_list_create( instrument));
  check( profile = cpl_vector_new( CPL_KERNEL_DEF_SAMPLES));
  check(cpl_vector_fill_kernel_profile( profile,rectify_par->kernel_type,
                                        rectify_par->rectif_radius));

  check(sci_name=cpl_frame_get_filename(div_frame));
  check(data_img=cpl_image_load(sci_name,XSH_PRE_DATA_TYPE,0,0));
  check(errs_img=cpl_image_load(sci_name,XSH_PRE_ERRS_TYPE,0,1));
  check(qual_img=cpl_image_load(sci_name,XSH_PRE_QUAL_TYPE,0,2));

  check(sp_fmt_name=cpl_frame_get_filename(spectral_format_frame));
  check(sp_fmt_tab=cpl_table_load(sp_fmt_name,1,0));
  check(ord_min=cpl_table_get_column_min(sp_fmt_tab,"ORDER"));
  check(ord_max=cpl_table_get_column_max(sp_fmt_tab,"ORDER"));
  nord=ord_max-ord_min+1;

  check(cube_wave_min=cpl_table_get(sp_fmt_tab,"WLMIN",nord-1,&status));
  check(cube_wave_max=cpl_table_get(sp_fmt_tab,"WLMAX",0,&status));

  xsh_msg_debug("cube_wave_min=%10.8g cube_wave_max=%10.8g",cube_wave_min,cube_wave_max);
 
  /* on X we have the different IFU slices */
  naxis1=3;
  xsh_msg_debug("ord_min=%d ord_max=%d",ord_min,ord_max);
  wave_step= cube_wstep;
  s_step= cube_sstep;
  naxis2=(int)((s_max-s_min)/s_step+0.5)+1;


  xsh_free_imagelist(&data_cube_merge);
  xsh_free_imagelist(&errs_cube_merge);
  xsh_free_imagelist(&qual_cube_merge);
  xsh_free_imagelist(&mask_cube_merge);

  data_cube_merge=cpl_imagelist_new();
  errs_cube_merge=cpl_imagelist_new();
  qual_cube_merge=cpl_imagelist_new();
  mask_cube_merge=cpl_imagelist_new();


  xsh_free_image(&data_tmp);
  xsh_free_image(&errs_tmp);
  xsh_free_image(&qual_tmp);
  xsh_free_image(&mask_tmp);

  mk=0;
  
    for( ord = ord_max; ord >= ord_min; ord-- ) {
//  for( ord = 24; ord >= 20; ord-- ) {
      xsh_free_imagelist(&data_cube);
      xsh_free_imagelist(&errs_cube);
      xsh_free_imagelist(&qual_cube);

      data_cube=cpl_imagelist_new();
      errs_cube=cpl_imagelist_new();
      qual_cube=cpl_imagelist_new();


      data_tmp=cpl_image_new(naxis1,naxis2,XSH_PRE_DATA_TYPE);
      errs_tmp=cpl_image_new(naxis1,naxis2,XSH_PRE_ERRS_TYPE);
      qual_tmp=cpl_image_new(naxis1,naxis2,XSH_PRE_QUAL_TYPE);
      mask_tmp=cpl_image_new(naxis1,naxis2,XSH_PRE_QUAL_TYPE);

      check(pima=cpl_image_get_data_float(data_tmp));
      check(perr=cpl_image_get_data_float(errs_tmp));
      check(pqua=cpl_image_get_data_int(qual_tmp));
      check(pmsk=cpl_image_get_data_int(mask_tmp));
  


      check(wave_min=cpl_table_get(sp_fmt_tab,"WLMIN",ord-ord_min,&status));
      check(wave_max=cpl_table_get(sp_fmt_tab,"WLMAX",ord-ord_min,&status));
      //cube_wave_min=wave_min;
      naxis3=(wave_max-wave_min)/wave_step+1;
      wave_cen=wave_min+naxis3/2*wave_step;
      xsh_msg_debug("order=%d naxis1=%d,naxis2=%d naxis3=%d, wave_min=%g wave_max=%g",
              ord,naxis1,naxis2,naxis3,wave_min,wave_max);
     
      /* in case we co from ord_min to ord_max 
      if(ord>ord_min) {
         xsh_msg("wave_min=%g wave_min_old=%g nstep_off=%d",
                 wave_min,wave_min_old,nstep_off);

         nstep_off=(int)((wave_min_old-wave_min)/wave_step+0.5);
         xsh_msg("before wave_min=%g wave_min_old=%g nstep_off=%d",
                 wave_min,wave_min_old,nstep_off);


         wave_min=wave_min_old-nstep_off*wave_step;
         xsh_msg("after wave_min=%g wave_min_old=%g nstep_off=%d",
                 wave_min,wave_min_old,nstep_off);


      }

      */
      /* in case wego from ord_max to ord_min */
      if(ord<ord_max) {
/*
         xsh_msg("wave_min=%g wave_min_old=%g nstep_off=%d",
                 wave_min,wave_min_old,nstep_off);
*/
         nstep_off=(int)((wave_min-wave_min_old)/wave_step+0.5);
/*
         xsh_msg("before wave_min=%g wave_min_old=%g nstep_off=%d",
                 wave_min,wave_min_old,nstep_off);
*/

         wave_min=wave_min_old+nstep_off*wave_step;
/*
         xsh_msg("after wave_min=%g wave_min_old=%g nstep_off=%d",
                 wave_min,wave_min_old,nstep_off);

*/

      }

     ik=0;     
//     for( wave = wave_min; wave <= wave_max; wave+=wave_step) {
     for( wave = wave_min; wave <= wave_max; wave+=wave_step) {
        is=0;
        for( s = s_min; s <= s_max; s+=s_step ) {

           /* upper slice */

           s_upp=-s+s_upp_off+w_upp_coeff1*wave+w_upp_coeff2*wave*wave;
           if(wavesol!=NULL) {
              check( x= xsh_wavesol_eval_polx( wavesol,wave, ord,s_upp));
              check( y= xsh_wavesol_eval_poly( wavesol,wave, ord,s_upp));
           } else {

              check(xsh_model_get_xy(&model_config,instrument,wave,ord,s_upp,
                                     &x,&y));
           }
           //xsh_msg("is=%d s_min=%g s_max=%g s=%g",is,s_min,s_max,s);
           check(flux_upp=cpl_image_get_interpolated(data_img,x,y,
                                                     profile,radius,
                                                     profile,radius, 
                                                     &confidence));
           check(errs_upp=cpl_image_get_interpolated(errs_img,x,y,
                                                     profile,radius,
                                                     profile,radius, 
                                                     &confidence));
           check(qual_upp=cpl_image_get_interpolated(qual_img,x,y,
                                                     profile,radius,
                                                     profile,radius, 
                                                     &confidence));

           pima[is*naxis1+0]=flux_upp;
           perr[is*naxis1+0]=errs_upp;
           pqua[is*naxis1+0]=qual_upp;
           pmsk[is*naxis1+0]=1;


  

           /* central slice */
           if(wavesol!=NULL) {
              check( x= xsh_wavesol_eval_polx( wavesol,wave, ord,s));
              check( y= xsh_wavesol_eval_poly( wavesol,wave, ord,s));
           } else {
              check(xsh_model_get_xy(&model_config,instrument,wave,ord,s,
                                     &x,&y));
           }
           check(flux_cen=cpl_image_get_interpolated(data_img,x,y,
                                                     profile,radius,
                                                     profile,radius, 
                                                     &confidence));

           check(errs_cen=cpl_image_get_interpolated(errs_img,x,y,
                                                     profile,radius,
                                                     profile,radius, 
                                                     &confidence));

           check(qual_cen=cpl_image_get_interpolated(qual_img,x,y,
                                                     profile,radius,
                                                     profile,radius, 
                                                     &confidence));

           pima[is*naxis1+1]=flux_cen;
           perr[is*naxis1+1]=errs_cen;
           pqua[is*naxis1+1]=qual_cen;
           pmsk[is*naxis1+1]=1;
/*
           xsh_msg("central flux=%g x=%g y=%g s=%g is=%d wave=%g ik=%d wave_step=%g binx=%d,biny=%d",
                   flux_cen,x,y,s,is,wave,ik,wave_step,binx,biny);

           xsh_msg("order=%d wave=%g",ord,wave);

*/

           /* lower slice */
           s_low=-s+s_low_off+w_low_coeff1*wave+w_low_coeff2*wave*wave;
           if(wavesol!=NULL) {
              check( x= xsh_wavesol_eval_polx( wavesol,wave, ord,s_low));
              check( y= xsh_wavesol_eval_poly( wavesol,wave, ord,s_low));
           } else {

              check(xsh_model_get_xy(&model_config,instrument,wave,ord,s_low,
                                     &x,&y));
           }
           //xsh_msg("x=%g y=%g",x,y);
           check(flux_low=cpl_image_get_interpolated(data_img,x,y,
                                                     profile,radius,
                                                     profile,radius, 
                                                     &confidence));
           check(errs_low=cpl_image_get_interpolated(errs_img,x,y,
                                                     profile,radius,
                                                     profile,radius, 
                                                     &confidence));
           check(qual_low=cpl_image_get_interpolated(qual_img,x,y,
                                                     profile,radius,
                                                     profile,radius, 
                                                     &confidence));

           pima[is*naxis1+2]=flux_low;
           perr[is*naxis1+2]=errs_low;
           pqua[is*naxis1+2]=qual_low;
           pmsk[is*naxis1+2]=1;


           is++;
        } /* end loop on s */
        wave_max_old=wave_max;
        wave_min_old=wave_min;

        check(cpl_imagelist_set(data_cube,cpl_image_duplicate(data_tmp),ik));
        check(cpl_imagelist_set(errs_cube,cpl_image_duplicate(errs_tmp),ik));
        check(cpl_imagelist_set(qual_cube,cpl_image_duplicate(qual_tmp),ik));
        /*
        xsh_msg("ord=%d wave=%14.8g wave_step=%14.8g cube_wave_max=%14.8g cube_wave_min=%14.8g wave_max=%14.8g wave_min=%14.8g mk=%d ik=%d",ord,wave,wave_step,cube_wave_max,cube_wave_min,wave_min,wave_max,mk,ik);
        */

        check(xsh_iml_merge_avg(&data_cube_merge,&mask_cube_merge,
				data_tmp,mask_tmp,mk));

	/*
        check(xsh_iml_merge_avg(&errs_cube_merge,&mask_cube_merge,
				data_tmp,mask_tmp,mk));

        check(xsh_iml_merge_avg(&qual_cube_merge,&mask_cube_merge,
				data_tmp,mask_tmp,mk));

	*/

/*
        check(cpl_imagelist_set(mask_cube_merge,cpl_image_duplicate(mask_tmp),mk));

        check(cpl_imagelist_set(data_cube_merge,cpl_image_duplicate(data_tmp),mk));

        check(cpl_imagelist_set(errs_cube_merge,cpl_image_duplicate(errs_tmp),mk));
        check(cpl_imagelist_set(qual_cube_merge,cpl_image_duplicate(qual_tmp),mk));
*/
        ik++;
        if(ord==ord_max) {
           mk++;
        } else {
           mk=(int)((wave-cube_wave_min)/wave_step+0.5);
        }
     } /* end loop on wave */

     sprintf(tag,"%s_ORDER3D_DATA_%s_%s",rec_prefix,qualifier,
             xsh_arm_tostring(arm));
     //sprintf(name,"%s_%2.2d.fits",tag,ord);
     sprintf(name,"%s.fits",tag);
     //xsh_pfits_set_pcatg(data_plist,tag);
     xsh_pfits_set_crval3(data_plist,wave_min);
     xsh_pfits_set_cdelt3(data_plist,wave_step);
     //xsh_cube_set_wcs(data_plist,(wave_min+wave_max)/2.,wave_step,ik/2.,1,naxis2/2.);

    if(ord==ord_max) {
     cpl_imagelist_save(data_cube,name,CPL_BPP_IEEE_FLOAT,data_plist,CPL_IO_DEFAULT);
     frame_data_cube=xsh_frame_product(name,tag,CPL_FRAME_TYPE_IMAGE,CPL_FRAME_GROUP_PRODUCT,CPL_FRAME_LEVEL_FINAL);
    } else {
     cpl_imagelist_save(data_cube,name,CPL_BPP_IEEE_FLOAT,data_plist,CPL_IO_EXTEND);
    }

     xsh_free_frame(&qc_trace_frame);
     sprintf(tag,"ORDER%2.2d",ord);
     check(qc_trace_frame=xsh_cube_qc_trace_window(frame_data_cube,
                                                   instrument,tag,rec_prefix,
                                                   save_size+1,
                                                   naxis2-save_size,
                                                   peack_search_hsize,
                                                   method,0));
   
     /*
     check( xsh_add_product_table_multi( qc_trace_frame, frameset,
                                         parameters, recipe_id, instrument));

     */
     sprintf(tag,"%s_ORDER3D_ERRS_%s_%s",rec_prefix,qualifier,
             xsh_arm_tostring(arm));
     //sprintf(name,"%s_%2.2d.fits",tag,ord);
     sprintf(name,"%s.fits",tag);

     if(ord==ord_max) {
        cpl_imagelist_save(errs_cube,name,CPL_BPP_IEEE_FLOAT,NULL,CPL_IO_DEFAULT);
        frame_errs_cube=xsh_frame_product(name,tag,CPL_FRAME_TYPE_IMAGE,CPL_FRAME_GROUP_PRODUCT,CPL_FRAME_LEVEL_FINAL);
     } else {
        cpl_imagelist_save(errs_cube,name,CPL_BPP_IEEE_FLOAT,NULL,CPL_IO_EXTEND);
     }

     sprintf(tag,"%s_ORDER3D_QUAL_%s_%s",rec_prefix,qualifier,
             xsh_arm_tostring(arm));
     //sprintf(name,"%s_%2.2d.fits",tag,ord);
     sprintf(name,"%s.fits",tag);
    
     if(ord==ord_max) {
        cpl_imagelist_save(qual_cube,name,CPL_BPP_IEEE_FLOAT,NULL,CPL_IO_DEFAULT);
        frame_qual_cube=xsh_frame_product(name,tag,CPL_FRAME_TYPE_IMAGE,CPL_FRAME_GROUP_PRODUCT,CPL_FRAME_LEVEL_FINAL);
     } else {
        cpl_imagelist_save(qual_cube,name,CPL_BPP_IEEE_FLOAT,NULL,CPL_IO_EXTEND);
     }


     sprintf(name,"QC%s_%2.2d.fits",tag,ord);

     xsh_free_image(&data_tmp);
     xsh_free_image(&errs_tmp);
     xsh_free_image(&qual_tmp);
     xsh_free_image(&mask_tmp);

    } /* end loop on ord */

    sprintf(tag,"%s_MERGE3D_DATA_%s_%s",rec_prefix,qualifier,
            xsh_arm_tostring(arm));
    sprintf(name,"%s.fits",tag);
    xsh_pfits_set_crval3(data_plist,cube_wave_min);
    xsh_pfits_set_cdelt3(data_plist,wave_step);
    //xsh_cube_set_wcs(data_plist,wave_cen,wave_step,naxis3/2.,1,naxis2/2.);

  cpl_imagelist_save(data_cube_merge,name,CPL_BPP_IEEE_FLOAT,data_plist,CPL_IO_DEFAULT);


  xsh_msg_debug("merge cube size=%d",cpl_imagelist_get_size(data_cube_merge));

  frame_merged_cube=xsh_frame_product(name,tag,CPL_FRAME_TYPE_IMAGE,
				      CPL_FRAME_GROUP_PRODUCT,CPL_FRAME_LEVEL_FINAL);

  check(xsh_add_product_imagelist(frame_merged_cube,frameset,parameters,
                                  recipe_id,instrument,tag)); 

  sprintf(tag,"%s_MERGE3D_%s",rec_prefix,qualifier);

  check(qc_trace_merged_frame=xsh_cube_qc_trace_window(frame_merged_cube,
                                                       instrument,
                                                       "MERGE3D",rec_prefix,
                                                       save_size+1,
                                                       naxis2-save_size,
                                                       peack_search_hsize,
                                                       method,1));
   
  if(frame_is_object) {
     check( xsh_add_product_table_multi( qc_trace_merged_frame, frameset,
                                         parameters, recipe_id, instrument));
  

     check(xsh_add_product_imagelist(frame_data_cube,frameset,parameters,
                                     recipe_id,
                                     instrument,
                                     cpl_frame_get_tag(frame_data_cube)));

     check(xsh_add_product_imagelist(frame_errs_cube,frameset,parameters,
                                     recipe_id,
                                     instrument, 
                                     cpl_frame_get_tag(frame_errs_cube)));

     check(xsh_add_product_imagelist(frame_qual_cube,frameset,parameters,
                                     recipe_id,
                                     instrument, 
                                     cpl_frame_get_tag(frame_qual_cube)));
  }
  /*
  cpl_imagelist_save(mask_cube_merge,name,CPL_BPP_IEEE_FLOAT,data_plist,CPL_IO_DEFAULT);

  xsh_free_frame(&qc_trace_frame);
  check(qc_trace_merged_frame=xsh_cube_qc_trace_window(frame_merged_cube,
  instrument,
  "MERGE3D",rec_prefix,
  save_size+1,
  naxis2-save_size,
  peack_search_hsize,method));
   
  check( xsh_add_product_table_multi( qc_trace_merged_frame, frameset,
  parameters, recipe_id, instrument));



  sprintf(tag,"MERGE3D_ERRS_%s",xsh_arm_tostring(arm));
  sprintf(name,"%s.fits",tag,ord);
  cpl_imagelist_save(errs_cube_merge,name,CPL_BPP_IEEE_FLOAT,data_plist,CPL_IO_DEFAULT);


  sprintf(tag,"MERGE3D_QUAL_%s",xsh_arm_tostring(arm));
  sprintf(name,"%s.fits",tag,ord);
  cpl_imagelist_save(qual_cube_merge,name,CPL_BPP_IEEE_FLOAT,data_plist,CPL_IO_DEFAULT);


  sprintf(tag,"MERGE3D_MASK_%s",xsh_arm_tostring(arm));
  sprintf(name,"%s.fits",tag,ord);
  */


 cleanup:

  xsh_free_propertylist(&data_plist);
  xsh_free_propertylist(&errs_plist);
  xsh_free_propertylist(&qual_plist);
  xsh_free_vector(&profile);

  xsh_free_image(&data_tmp);
  xsh_free_image(&errs_tmp);
  xsh_free_image(&qual_tmp);
  xsh_free_image(&mask_tmp);

  xsh_free_image(&data_img);
  xsh_free_image(&errs_img);
  xsh_free_image(&qual_img);


  xsh_free_table(&sp_fmt_tab);

  xsh_free_imagelist(&data_cube);
  xsh_free_imagelist(&errs_cube);
  xsh_free_imagelist(&qual_cube);


  xsh_free_imagelist(&data_cube_merge);
  xsh_free_imagelist(&errs_cube_merge);
  xsh_free_imagelist(&qual_cube_merge);
  xsh_free_imagelist(&mask_cube_merge);

  xsh_rec_list_free(&rec_list);
  xsh_free_frame(&qc_trace_frame);
  xsh_free_frame(&qc_trace_merged_frame);
  xsh_free_frame(&frame_data_cube);
  xsh_free_frame(&frame_errs_cube);
  xsh_free_frame(&frame_qual_cube);
  xsh_free_frame(&frame_merged_cube);
 /* 
  xsh_free_frame(&spectral_format_frame);
  xsh_free_frame(&model_config_frame);
  */
 
  return cpl_error_get_code();

}

/**@}*/
