/*                                                                            *
 *   This file is part of the ESO X-shooter Pipeline                          *
 *   Copyright (C) 2006 European Southern Observatory                         *
 *                                                                            *
 *   This library is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as published by     *
 *   the Free Software Foundation; either version 2 of the License, or        *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU General Public License for more details.                             *
 *                                                                            *
 *   You should have received a copy of the GNU General Public License        *
 *   along with this program; if not, write to the Free Software              *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA     *
 *                                                                            */

/*
 * $Author: amodigli $
 * $Date: 2011/02/10 16:08:35 $
 * $Revision: 1.12 $
 * $Name: HEAD $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
/*----------------------------------------------------------------------------
                                 Includes
 ----------------------------------------------------------------------------*/
#include <xsh_utils_table.h>
#include <xsh_utils.h>
#include <xsh_error.h>
#include <xsh_msg.h>
/*----------------------------------------------------------------------------
                                 Variables
 ----------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/**
 * @defgroup xsh_utils     Tools: Miscellaneous Utilities
 *
 * This module is for the stuff that don't fit anywhere else. It should be
 * kept as small as possible.
 *
 */
/*---------------------------------------------------------------------------*/

/**@{*/

/*---------------------------------------------------------------------------*/
/**
   @brief    Read a table value from a fits table
   @param    table table to read
   @param    colname Name of column to read
   @param    coltype Type of column
   @param    i row to read
   @param    result The value read

   @return   CPL_ERROR_NONE iff OK
   @note derived from UVES pipeline
*/
/*---------------------------------------------------------------------------*/
cpl_error_code xsh_get_table_value (const cpl_table* table,
  const char *colname, cpl_type coltype, int i, void* result)
{
  int flag = 0;
  /* Check input */
  XSH_ASSURE_NOT_NULL( table);
  XSH_ASSURE_NOT_NULL( colname);
  XSH_ASSURE_NOT_NULL( result);

  /* Read the column */
  
  switch (coltype) {
    case CPL_TYPE_INT:
      check_msg (*((int *) result) = cpl_table_get_int(table,colname,i,&flag),
        "Could not get (integer) value of %s at row %d", colname,i);
    break;
  case CPL_TYPE_FLOAT:
    check_msg (*((float *) result) = cpl_table_get_float( table, 
      colname,i,&flag),
      "Could not get (float) value of %s at row %d",colname,i);
    break;
  case CPL_TYPE_DOUBLE:
    check_msg (*((double *) result) = cpl_table_get_double (table,
      colname,i,&flag),
      "Could not get (double) value of %s at row %d",colname,i);
    break;
  case CPL_TYPE_STRING:
    check_msg (*((const char **) result) =
               cpl_table_get_string (table, colname,i),
               "Could not get (string) value of %s at row %d",colname,i);
    break;
  default:
    assure (false, CPL_ERROR_INVALID_TYPE, "Unknown type");
  }

cleanup:
  return cpl_error_get_code ();
}

XSH_TABLE_GET_ARRAY(int)
XSH_TABLE_GET_ARRAY(float)
XSH_TABLE_GET_ARRAY(double)

/*---------------------------------------------------------------------------*/
/**
   @brief    Sort a table by one column
   @param    t        Table
   @param    column   Column name
   @param    reverse  Flag indicating if column values are sorted descending
                      (CPL_TRUE) or ascending (CPL_FALSE)
   @return   CPL_ERROR_NONE iff OK
   @note derived from UVES pipeline

   This is a wrapper of @c cpl_table_sort().

*/
/*---------------------------------------------------------------------------*/
cpl_error_code
xsh_sort_table_1(cpl_table *t, const char *column, cpl_boolean reverse)
{
    cpl_propertylist *plist = NULL;
    
    assure(t != NULL, CPL_ERROR_NULL_INPUT, "Null table");
    assure(cpl_table_has_column(t, column), CPL_ERROR_ILLEGAL_INPUT, 
       "No column '%s'", column);

    check_msg(( plist = cpl_propertylist_new(),
        cpl_propertylist_append_bool(plist, column, reverse)),
       "Could not create property list for sorting");

    check_msg( cpl_table_sort(t, plist), "Could not sort table");

  cleanup:
    xsh_free_propertylist(&plist);
    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/
/**
   @brief    Sort a table by two columns
   @param    t        Table
   @param    column1  1st column name
   @param    column2  2nd column name
   @param    reverse1  Flag indicating if 1st column values are sorted
                       descending (true) or ascending (CPL_FALSE)
   @param    reverse2  Flag indicating if 2nd column values are sorted
                       descending (true) or ascending (CPL_TRUE)
   @return   CPL_ERROR_NONE iff OK
   @note derived from UVES pipeline

   This is a wrapper of @c cpl_table_sort(). @em column1 is the more
   significant column (i.e. values in @em column2 are compared, only if the
   values in @em column1 are equal).
*/
/*---------------------------------------------------------------------------*/
cpl_error_code
xsh_sort_table_2(cpl_table *t, const char *column1, const char *column2, 
          cpl_boolean reverse1, cpl_boolean reverse2)
{
    cpl_propertylist *plist = NULL;
    
    assure(t != NULL, CPL_ERROR_NULL_INPUT, "Null table");
    assure(cpl_table_has_column(t, column1), CPL_ERROR_ILLEGAL_INPUT, 
       "No column '%s'", column1);
    assure(cpl_table_has_column(t, column2), CPL_ERROR_ILLEGAL_INPUT,
       "No column '%s'", column2);

    check_msg(( plist = cpl_propertylist_new(),
        cpl_propertylist_append_bool(plist, column1, reverse1),
        cpl_propertylist_append_bool(plist, column2, reverse2)),
       "Could not create property list for sorting");
    check_msg( cpl_table_sort(t, plist), "Could not sort table");
    
  cleanup:
    xsh_free_propertylist(&plist);
    return cpl_error_get_code();
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Interpolate data points
  @param    wav       value at which is desired an interpolated value
  @param    nrow      number of data points
  @param    pw        pointer to wave array
  @param    pe        pointer to efficiency array
  @return   Interpolated data points
 */
/*---------------------------------------------------------------------------*/
double
xsh_data_interpolate(
		     double wav,
		     int nrow,
		     double* pw,
		     double* pe
		     )
{
  double y = 0;
  double w1=pw[0];
  double w2=pw[nrow-1];
  double y1_=pe[0]; /*was changed from y1 to y1_ due a warning from compiler - shadowed variable*/
  double y2=pe[nrow-1];
  if(wav < pw[0])
    {
      w1=pw[0];
      w2=pw[1];
      y1_=pe[0];
      y2=pe[1];
    }
  else if (wav > pw[nrow - 1])
    {
      w1=pw[nrow - 2];
      w2=pw[nrow - 1];
      y1_=pe[nrow - 2];
      y2=pe[nrow - 1];
    }
  else
    {
      int l = 0;
      int h = nrow - 1;
      int curr_row = 0;
      curr_row = (h-l) / 2;
      while( h-l >1 )
	{
	  if(wav < pw[curr_row])
	    {
	      h = curr_row;
	    }
	  else
	    {
	      l = curr_row;
	    }
	  curr_row = (h-l) / 2 + l;
	}
      w1=pw[curr_row];
      w2=pw[curr_row + 1];
      y1_=pe[curr_row];
      y2=pe[curr_row + 1];
    }
  y=y1_+(y2-y1_)/(w2-w1)*(wav-w1);
  return y;
}

/*---------------------------------------------------------------------------*/
/**
   @brief    Interpolate table columns
   @param    tbl        Table
   @param    wav    value at which is desired to get an interpolated value
   @param    colx   1st column name
   @param    coly   2nd column name

   @return   interpolated value if ok

*/
/*---------------------------------------------------------------------------*/
double
xsh_table_interpolate(cpl_table* tbl,
		  double wav,
		  const char* colx,
		  const char* coly)
{

  double y=0;
  double* pe=NULL;
  double* pw=NULL;
  int nrow=0;

  check(pw=cpl_table_get_data_double(tbl,colx));
  check(pe=cpl_table_get_data_double(tbl,coly));
  check(nrow=cpl_table_get_nrow(tbl));

  y = xsh_data_interpolate(wav, nrow, pw, pe);
 cleanup:
  return y;

}

/**
   @brief Computes statistics on spectrum for QC
   @param[in] table input table
   @param[in] ws    wavelength start value
   @param[in] we    wavelength end value
   @param[in] prefix prefix of QC parameter
   @param[in] index index to specify QC parameter name
   @param[out] header header containing computed QC key
   @return  CPL error code
*/
static cpl_error_code
xsh_table_monitor_flux_qc(cpl_table* table,
			  const double ws,
			  const double we,
			  const char* prefix,
			  const int index,
			  cpl_propertylist** header)
{

  double flux=0;
  cpl_table* ext=NULL;
  char comment[40];
  char qc_key[20];
  int next=0;
  
  check(cpl_table_and_selected_double(table,"W",CPL_GREATER_THAN,ws));
  check(next=cpl_table_and_selected_double(table,"W",CPL_LESS_THAN,we));
  if(next>0) {
    check(ext=cpl_table_extract_selected(table));
    check(cpl_table_select_all(table));
    if(cpl_table_has_valid(ext,"F")) {
       check(flux=cpl_table_get_column_mean(ext,"F"));
    } else {
       flux=-999;
    }
    sprintf(qc_key,"ESO QC %s%d",prefix,index);
    sprintf(comment,"Mean value of %s in %4.0f-%4.0f nm",prefix,ws,we);
    check(cpl_propertylist_append_double(*header,qc_key,flux));
    check(cpl_propertylist_set_comment(*header,qc_key,comment));
  }


 cleanup:
  xsh_free_table(&ext);

  return cpl_error_get_code();

}

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/**
   @brief Computes statistics on spectrum for QC
   @param[in] table input table describing spectrum 
   @param[in] colw  wavelength column name
   @param[in] colf  flux column name
   @param[in] prefix prefix of QC parameter
   @param[in] instrument instrument arm setting
   @return  CPL error code
*/
static cpl_propertylist*
xsh_frame_table_monitor_flux_qc_ext(cpl_table* table,
				    const char* colw,
				    const char* colf,
				    const char* prefix,
				    xsh_instrument* instrument)
{
  cpl_propertylist* header=NULL;

  XSH_ASSURE_NOT_NULL_MSG(table,"Null input table");
  header=cpl_propertylist_new();
  check(cpl_table_cast_column(table,colw,"W",CPL_TYPE_DOUBLE));
  check(cpl_table_cast_column(table,colf,"F",CPL_TYPE_DOUBLE));

  if ( xsh_instrument_get_arm(instrument) == XSH_ARM_UVB){

    check( xsh_table_monitor_flux_qc(table,450,470,prefix,1,&header));
    check( xsh_table_monitor_flux_qc(table,510,530,prefix,2,&header));

  }
  else if ( xsh_instrument_get_arm(instrument) == XSH_ARM_VIS){

    check( xsh_table_monitor_flux_qc(table,672,680,prefix,1,&header));
    check( xsh_table_monitor_flux_qc(table,745,756,prefix,2,&header));
    check( xsh_table_monitor_flux_qc(table,992,999,prefix,3,&header));

  }
  else if ( xsh_instrument_get_arm(instrument) == XSH_ARM_NIR){

    check( xsh_table_monitor_flux_qc(table,1514,1548,prefix,1,&header));
    check( xsh_table_monitor_flux_qc(table,2214,2243,prefix,2,&header));

  }

  cpl_table_erase_column(table,"W");
  cpl_table_erase_column(table,"F");

cleanup:
  if(cpl_error_get_code() != CPL_ERROR_NONE) {
    return NULL;
  } else {
    return header;
  }
}


/*---------------------------------------------------------------------------*/
/**
   @brief Computes statistics on spectrum for QC
   @param[in] frm   spectrum frame
   @param[in] colw  wavelength column name
   @param[in] colf  flux column name
   @param[in] prefix prefix of QC parameter
   @param[in] instrument instrument arm setting
   @return  CPL error code
*/
/*---------------------------------------------------------------------------*/
cpl_error_code
xsh_frame_table_monitor_flux_qc(cpl_frame* frm,
				const char* colw,
				const char* colf,
				const char* prefix,
				xsh_instrument* instrument)
{

  const char* name=NULL;
  cpl_table* table=NULL;
  cpl_propertylist* header=NULL;
  cpl_propertylist* qc_header=NULL;
 
  int next=0;
  int ext=0;

  XSH_ASSURE_NOT_NULL_MSG(frm,"Null input spectrum frame");
  next=cpl_frame_get_nextensions(frm);
  name=cpl_frame_get_filename(frm);
 
  /* compute QC */
 
  if (next) {
    for(ext=1;ext<=next;ext++) {
    
      table=cpl_table_load(name,1,ext);
      check(header=cpl_propertylist_load(name,0));
      check(qc_header=xsh_frame_table_monitor_flux_qc_ext(table,colw,colf,
							  prefix,instrument));
      cpl_propertylist_append(header,qc_header);
      if(next==1) {
	check(cpl_table_save(table,header,NULL,name,CPL_IO_DEFAULT));
      } else {
	check(cpl_table_save(table,header,NULL,name,CPL_IO_EXTEND));
      }
      xsh_free_table(&table);
      xsh_free_propertylist(&qc_header);
      xsh_free_propertylist(&header);
    }
  } else {
    table=cpl_table_load(name,1,0);
    check(header=cpl_propertylist_load(name,0));
    check(qc_header=xsh_frame_table_monitor_flux_qc_ext(table,colw,colf,
							prefix,instrument));
    cpl_propertylist_append(header,qc_header);
    check(cpl_table_save(table,header,NULL,name,CPL_IO_DEFAULT));
    xsh_free_propertylist(&qc_header);
    xsh_free_propertylist(&header);
  }

 cleanup:
  xsh_free_table(&table);
  xsh_free_propertylist(&qc_header);
  xsh_free_propertylist(&header);

  return cpl_error_get_code();

}
/**@}*/

